# include "stdafx.h"
# include "Resource.h"
# include "UndoRedo.h"
# include "Map.h"
# include "MapView.h"
# include "Scale.h"
# include "Position.h"
# include "AffTransform.h"
# include "MathAlgo.h"
# include "Settings.h"
# include "Globals.h"

const int c_AB_SIZE = 150;

static size_t g_cPointIndex;
static bool g_bPairDefined [3];
static point_t g_ptFrom [3], g_ptTo [3];

static aff_transform_t g_atNet;

//////////////////////////////////

class CAffineTransformationAction : public CEditAction {
  public:
	CAffineTransformationAction (const aff_transform_t & _at) : m_at (_at) {
		m_at.Reverse (m_atReverse);

		for (size_t cPoint = 0; cPoint < 3; ++ cPoint) {
			m_bPairDefined [cPoint] = g_bPairDefined [cPoint];
			m_ptFrom       [cPoint] = g_ptFrom       [cPoint];
			m_ptTo         [cPoint] = g_ptTo         [cPoint];
		}

		m_atNet = g_atNet;
	}

	virtual void GetName (std::string & _str) const;
	virtual bool CanUndo () const;
	virtual bool Do ();
	virtual void Undo ();
	virtual void Redo ();

  private:
	const aff_transform_t m_at;
	aff_transform_t m_atReverse;

	aff_transform_t m_atNet;

	bool m_bPairDefined [3];
	point_t m_ptFrom [3], m_ptTo [3];
};

void CAffineTransformationAction::GetName (std::string & _str) const {
	_str.append ("affine transformation");
}

bool CAffineTransformationAction::CanUndo () const {return true;}

bool CAffineTransformationAction::Do () {
	CCritSec::CLock lock (g_cs);

	g_map.AffTransform (m_at);

	m_at.Multiply (m_atNet, g_atNet);

	for (size_t cPoint = 0; cPoint < 3; ++ cPoint) {
		if (g_bPairDefined [cPoint]) {
			m_at.ApplyTo (g_ptFrom [cPoint]);
			assert (
				::fabs (g_ptFrom [cPoint].x - g_ptTo [cPoint].x) < 1e-5 &&
				::fabs (g_ptFrom [cPoint].y - g_ptTo [cPoint].y) < 1e-5
			);
		}
	}

	return true;
}

void CAffineTransformationAction::Undo () {
	CCritSec::CLock lock (g_cs);

	g_map.AffTransform (m_atReverse);

	g_atNet = m_atNet;

	for (size_t cPoint = 0; cPoint < 3; ++ cPoint) {
		g_bPairDefined [cPoint] = m_bPairDefined [cPoint];
		g_ptFrom       [cPoint] = m_ptFrom       [cPoint];
		g_ptTo         [cPoint] = m_ptTo         [cPoint];
	}
}

void CAffineTransformationAction::Redo () {
	CAffineTransformationAction::Do ();
}

static inline
void AffTransformMap (const aff_transform_t & _at) {
	PerformEditAction (new CAffineTransformationAction (_at));
}

/////////////////////////////////

static HCURSOR g_hcurAT [3];
static HCURSOR g_hcurATDrag;

static HPEN g_hRedThinPen;
static HPEN g_hDkRedThinPen;

static HWND g_hwndAffineTransformation;

int CALLBACK ATProc (HWND _hWnd, UINT _uMsg, WPARAM _wParam, LPARAM);
void RefreshATDialog ();

bool CAffTransformMode::Init () {
	g_hcurAT [0] = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorModeAT));
	g_hcurAT [1] = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorModeAT2));
	g_hcurAT [2] = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorModeAT3));
	g_hcurATDrag = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorModeATDrag));

	g_hRedThinPen   = ::CreatePen (PS_SOLID, 1, c_colRed);
	g_hDkRedThinPen = ::CreatePen (PS_SOLID, 1, c_colDkRed);

	assert (g_hMapWnd);
	g_hwndAffineTransformation = ::CreateDialog (g_hInst, MAKEINTRESOURCE (IDD_AFFINE_TRANSFORMATION), g_hMapWnd, ATProc);
	if (g_hwndAffineTransformation == NULL)
		return false;
	::SetWindowPos (g_hwndAffineTransformation, NULL, GetSetting ("AffTransX", (DWORD) 0), GetSetting ("AffTransY", (DWORD) 0), 0, 0, SWP_NOZORDER | SWP_NOSIZE);

	return true;
}

static
void RefreshATDialog () {
	assert (g_hwndAffineTransformation);

	switch (g_cPointIndex) {
		default:
		case 0: ::CheckRadioButton (g_hwndAffineTransformation, IDC_POINT1, IDC_POINT3, IDC_POINT1); break;
		case 1: ::CheckRadioButton (g_hwndAffineTransformation, IDC_POINT1, IDC_POINT3, IDC_POINT2); break;
		case 2: ::CheckRadioButton (g_hwndAffineTransformation, IDC_POINT1, IDC_POINT3, IDC_POINT3); break;
	}

	::EnableWindow (::GetDlgItem (g_hwndAffineTransformation, IDC_LOCATE_POINT1), g_bPairDefined [0]);
	::EnableWindow (::GetDlgItem (g_hwndAffineTransformation, IDC_DELETE_POINT1), g_bPairDefined [0]);

	::EnableWindow (::GetDlgItem (g_hwndAffineTransformation, IDC_LOCATE_POINT2), g_bPairDefined [1]);
	::EnableWindow (::GetDlgItem (g_hwndAffineTransformation, IDC_DELETE_POINT2), g_bPairDefined [1]);

	::EnableWindow (::GetDlgItem (g_hwndAffineTransformation, IDC_LOCATE_POINT3), g_bPairDefined [2]);
	::EnableWindow (::GetDlgItem (g_hwndAffineTransformation, IDC_DELETE_POINT3), g_bPairDefined [2]);

	const size_t cPairsDefined = g_bPairDefined [0] + g_bPairDefined [1] + g_bPairDefined [2];
	::EnableWindow (::GetDlgItem (g_hwndAffineTransformation, IDC_APPLY), cPairsDefined == 3 || cPairsDefined == 1);
}

static void CalculateAT () {
	size_t cPairsDefined = 0;
	point_t ptFrom [3], ptTo [3];
	for (size_t c = 0; c < 3; ++ c)
		if (g_bPairDefined [c]) {
			ptFrom [cPairsDefined] = g_ptFrom [c];
			ptTo   [cPairsDefined] = g_ptTo   [c];
			++ cPairsDefined;
		}
	assert (cPairsDefined < 3);

	aff_transform_t at;
	switch (cPairsDefined) {
		case 3: {
			float fX_To [3], fY_To [3];
			fX_To [0] = ptTo [0].x;
			fY_To [0] = ptTo [0].y;
			fX_To [1] = ptTo [1].x;
			fY_To [1] = ptTo [1].y;
			fX_To [2] = ptTo [2].x;
			fY_To [2] = ptTo [2].y;
			if (
				! Solve3x3 (ptFrom, fX_To, at.fAxx, at.fAxy, at.fBx) ||
				! Solve3x3 (ptFrom, fY_To, at.fAyx, at.fAyy, at.fBy)
			) {
				ErrorMsgBox ("The transform is impossible:\ncontradictory input data.");
				return;
			}
			break;
		}

		case 2:
			// TODO:
			assert (false);
			break;

		case 1:
			at.fAxx = 1;
			at.fAxy = 0;
			at.fAyx = 0;
			at.fAyy = 1;
			at.fBx = ptTo [0].x - ptFrom [0].x;
			at.fBy = ptTo [0].y - ptFrom [0].y;
			break;

		case 0:
			assert (false);
			return;
	}

	AffTransformMap (at);

	RefreshScales ();
	Refresh ();
}

static
int CALLBACK ATProc (HWND _hWnd, UINT _uMsg, WPARAM _wParam, LPARAM) {
	switch (_uMsg) {
		case WM_INITDIALOG: {
			return TRUE;
		}

		case WM_COMMAND:
			switch (LOWORD (_wParam)) {
				case IDC_POINT1: g_cPointIndex = 0; CMouseMode::SetPermanentCursor (g_hcurAT [0]); Redraw (); return TRUE;
				case IDC_POINT2: g_cPointIndex = 1; CMouseMode::SetPermanentCursor (g_hcurAT [1]); Redraw (); return TRUE;
				case IDC_POINT3: g_cPointIndex = 2; CMouseMode::SetPermanentCursor (g_hcurAT [2]); Redraw (); return TRUE;

				case IDC_DELETE_POINT1: g_bPairDefined [0] = false; RefreshATDialog (); Redraw (); return TRUE;
				case IDC_DELETE_POINT2: g_bPairDefined [1] = false; RefreshATDialog (); Redraw (); return TRUE;
				case IDC_DELETE_POINT3: g_bPairDefined [2] = false; RefreshATDialog (); Redraw (); return TRUE;

				case IDC_LOCATE_POINT1:	if (g_bPairDefined [0])	SetToCenter (g_ptFrom [0]); Refresh ();	return TRUE;
				case IDC_LOCATE_POINT2: if (g_bPairDefined [1]) SetToCenter (g_ptFrom [1]); Refresh ();	return TRUE;
				case IDC_LOCATE_POINT3: if (g_bPairDefined [2]) SetToCenter (g_ptFrom [2]); Refresh ();	return TRUE;

				case IDC_APPLY: CalculateAT ();	return TRUE;

				case IDC_AT_INFO: {
					char strText [256];
					::sprintf  (
						strText,
						"Net parameters of all previous\n"
						" map transformations:\n"
						"\n"
						"A11=%f\n"
						"A12=%f\n"
						"A21=%f\n"
						"A22=%f\n"
						"B1=%f\n"
						"B2=%f",
						g_atNet.fAxx, g_atNet.fAxy, g_atNet.fAyx, g_atNet.fAyy, g_atNet.fBx, g_atNet.fBy
					);
					::MessageBox (g_hWnd, strText, "Affine transformation", MB_OK | MB_ICONINFORMATION);
					return TRUE;
				}
			}
			break;

		case WM_MOVE: {
			RECT rect;
			::GetWindowRect (_hWnd, & rect);
			POINT p = {rect.left, rect.top};
			::ScreenToClient (g_hMapWnd, & p);
			if (p.x < 0)
				p.x = 0;
			if (p.y < 0)
				p.y = 0;

			SaveSetting ("AffTransX", DWORD (p.x));
			SaveSetting ("AffTransY", DWORD (p.y));
			break;
		}

		case WM_CLOSE:
			CMouseMode::Set (0);
			return TRUE;
	}

	return FALSE;
}

/////////////////////////////////

void CAffTransformMode::OnNewMap () {
	g_atNet.fAxx = g_atNet.fAyy = 1;
	g_atNet.fAxy = g_atNet.fAyx = 0;
	g_atNet.fBx  = g_atNet.fBy  = 0;

	g_cPointIndex = 0;

	g_bPairDefined [0] = false;
	g_bPairDefined [1] = false;
	g_bPairDefined [2] = false;

	RefreshATDialog ();
}

static bool g_bATActive;

void CAffTransformMode::OnEnterMode () {
	g_bATActive = true;

	::SetWindowPos (g_hwndAffineTransformation, HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
	::ShowWindow (g_hwndAffineTransformation, SW_SHOW);

	assert (g_cPointIndex < 3);
	SetPermanentCursor (g_hcurAT [g_cPointIndex]);
}

void CAffTransformMode::OnLeaveMode () {
	g_bATActive = false;

	::ShowWindow (g_hwndAffineTransformation, SW_HIDE);
}

void CAffTransformMode::OnLDown (int _x, int _y, DWORD _dwKeys) {
	assert (g_cPointIndex < 3);
	SetPermanentCursor (g_hcurATDrag);

	ScreenToPt (_x, _y, g_ptFrom [g_cPointIndex]);
	g_ptTo [g_cPointIndex] = g_ptFrom [g_cPointIndex];
	g_bPairDefined [g_cPointIndex] = true;

	RefreshATDialog ();
	Redraw ();

	CMouseMode::s_bDND = true;
	::SetCapture (g_hMapWnd);
}

void CAffTransformMode::OnLUp (int _x, int _y, DWORD _dwKeys) {
	assert (g_cPointIndex < 3);
	SetPermanentCursor (g_hcurAT [g_cPointIndex]);

	if (CMouseMode::s_bDND) {
		CMouseMode::s_bDND = false;
		::ReleaseCapture ();
	}

	ScreenToPt (_x, _y, g_ptTo [g_cPointIndex]);

	Redraw ();
}

void CAffTransformMode::OnMove (int _x, int _y, DWORD _dwKeys) {
	if (! CMouseMode::s_bDND)
		return;

	assert (g_cPointIndex < 3);
	SetPermanentCursor (g_hcurATDrag);

	if (g_bPairDefined [g_cPointIndex]) {
		const point_t ptOldTo = g_ptTo [g_cPointIndex];

		ScreenToPt (_x, _y, g_ptTo [g_cPointIndex]);

		// Optimize update rect.
		const long lX_From  = XToScreen (g_ptFrom [g_cPointIndex].x);
		const long lY_From  = YToScreen (g_ptFrom [g_cPointIndex].y);
		const long lX_To    = XToScreen (g_ptTo   [g_cPointIndex].x);
		const long lY_To    = YToScreen (g_ptTo   [g_cPointIndex].y);
		const long lX_OldTo = XToScreen (ptOldTo.x);
		const long lY_OldTo = YToScreen (ptOldTo.y);

		RECT rect1 = {lX_From, lY_From, lX_From + 1, lY_From + 1};
		RECT rect2 = {lX_To    - c_AB_SIZE, lY_To    - c_AB_SIZE, lX_To    + c_AB_SIZE + 1, lY_To    + c_AB_SIZE + 1};
		RECT rect3 = {lX_OldTo - c_AB_SIZE, lY_OldTo - c_AB_SIZE, lX_OldTo + c_AB_SIZE + 1, lY_OldTo + c_AB_SIZE + 1};

		RECT rect;
		::UnionRect (& rect, & rect1, & rect2);
		::UnionRect (& rect, & rect, & rect3);

		Redraw (& rect);
	}
}

void CAffTransformMode::OnEsc () {
	assert (g_cPointIndex < 3);
	g_bPairDefined [g_cPointIndex] = false;

	Redraw ();
}

void CAffTransformMode::OnPaint (HDC _hDC) {
	if (! g_bATActive)
		return;

	//
	// Draw shifted semi-transparent part of map.
	//
	if (g_bPairDefined [g_cPointIndex]) {
		const long lX_From = XToScreen (g_ptFrom [g_cPointIndex].x);
		const long lY_From = YToScreen (g_ptFrom [g_cPointIndex].y);
		const long lX_To   = XToScreen (g_ptTo   [g_cPointIndex].x);
		const long lY_To   = YToScreen (g_ptTo   [g_cPointIndex].y);

		if (lX_From != lX_To || lY_From != lY_To) {
			long w = c_AB_SIZE*2;
			long h = c_AB_SIZE*2;
			long sx = lX_From - c_AB_SIZE;
			long sy = lY_From - c_AB_SIZE;
			long dx = lX_To - c_AB_SIZE;
			long dy = lY_To - c_AB_SIZE;
			if (sx < 0) {
				dx -= sx;
				w += sx;
				sx = 0;
			} else if (sx + w >= g_lWidth)
				w = g_lWidth - sx;
			if (sy < 0) {
				dy -= sy;
				h += sy;
				sy = 0;
			} else if (sy + h >= g_lHeight)
				h = g_lHeight - sy;

			BLENDFUNCTION bf = {AC_SRC_OVER, 0, 96, 0};
			g_fpAlphaBlend (_hDC, dx, dy, w, h, g_hMemDC, sx, sy, w, h, bf);
		}
	}

	//
	// Draw arrows.
	//

	::SelectObject (_hDC, g_hFont);
	::SetBkMode (_hDC, TRANSPARENT);
	::SetTextAlign (_hDC, TA_CENTER | TA_BASELINE);

	for (size_t cPoint = 0; cPoint < 3; ++ cPoint) {
		if (! g_bPairDefined [cPoint])
			continue;

		::SelectObject (_hDC, cPoint == g_cPointIndex ? g_hDkRedThinPen : g_hRedThinPen);
		::SetTextColor (_hDC, cPoint == g_cPointIndex ? c_colDkRed : c_colRed);

		const long lX_From = XToScreen (g_ptFrom [cPoint].x);
		const long lY_From = YToScreen (g_ptFrom [cPoint].y);
		const long lX_To   = XToScreen (g_ptTo   [cPoint].x);
		const long lY_To   = YToScreen (g_ptTo   [cPoint].y);

		if (lX_From != lX_To || lY_From != lY_To) {
			::MoveToEx (_hDC, lX_From, lY_From, NULL);
			::LineTo   (_hDC, lX_To, lY_To);

			// Draw pointer.
			const double a = ::atan2 (lY_To - lY_From, lX_To - lX_From);
			const long x1 = lX_To - fast_floor (8*::cos (+.3 - a) + .5);
			const long y1 = lY_To + fast_floor (8*::sin (+.3 - a) + .5);
			const long x2 = lX_To - fast_floor (8*::cos (-.3 - a) + .5);
			const long y2 = lY_To + fast_floor (8*::sin (-.3 - a) + .5);

			::MoveToEx (_hDC, lX_To, lY_To, NULL);
			::LineTo   (_hDC, x1, y1);

			::MoveToEx (_hDC, lX_To, lY_To, NULL);
			::LineTo   (_hDC, x2, y2);
		} else {
			// Draw cross.
			::MoveToEx (_hDC, lX_To - 5, lY_To - 5, NULL);
			::LineTo   (_hDC, lX_To + 6, lY_To + 6);

			::MoveToEx (_hDC, lX_To - 5, lY_To + 5, NULL);
			::LineTo   (_hDC, lX_To + 6, lY_To - 6);
		}

		// Draw label.
		char strBuf [10];
		::itoa (cPoint + 1, strBuf, 10);
		const float d = ::_hypot (lX_From - lX_To, lY_From - lY_To);
		const long dx = d > 0 ? 0 + fast_floor ((lY_From - lY_To)*8/d + .5) : 1;
		const long dy = d > 0 ? 3 - fast_floor ((lX_From - lX_To)*8/d + .5) : -5;
		::TextOut (_hDC, (lX_From + lX_To)/2 + dx, (lY_From + lY_To)/2 + dy, strBuf, 1);
	}
}
