# include "stdafx.h"
# include "Settings.h"
# include "Map.h"
# include "Types.h"
# include "Levels.h"
# include "Load.h"
# include "Status.h"
# include "MapView.h"
# include "Scale.h"
# include "Position.h"
# include "Selection.h"
# include "Globals.h"

static
bool DrawRgn (HDC _hDC, const rgn_t & _rgn, POINT * _pts, bool _bTransparentPolygons) {
	const type_desc_t * const pTD = _rgn.pTypeDesc;
	assert (pTD);
	const bool bUnknown = pTD == g_types;

	bool bDrawLabel = false;

	// For each region element...
	for (size_t cElement = 0; cElement < _rgn.elements.size (); ++ cElement) {
		if (g_bStop)
			return false;

		const rgn_element_t & element = _rgn.elements [cElement];
		if (element.bDeleted)
			continue;

		const points_t & points = element.points;
		const size_t cPoints = points.size ();

		if (cPoints == 0)
			continue;

		if (! (_rgn.dwFlags & rgn_t::fWaypointOrTrack) && ! LevelIsVisible (element.cLevel))
			continue;

		// Calculate screen coordinates.
		for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
			const point_t & pt = points [cPoint];
			_pts [cPoint].x = XToScreen (pt.x);
			_pts [cPoint].y = YToScreen (pt.y);

			if (g_bStop)
				return false;
		}

		if (_rgn.IsPolygon ()) {
			if (element.fArea*g_kY*g_kX < 4)
				// The element is too small.
				continue;

			// Draw polygon.
			if (_bTransparentPolygons) {
				assert (pTD->hbrTransparentPolygon);
				::SelectObject (_hDC, element.bIsland ? g_hWhiteHatchedBrush : pTD->hbrTransparentPolygon);
			} else {
				assert (pTD->phBrush);
				::SelectObject (_hDC, element.bIsland ? g_hWhiteBrush : * (pTD->phBrush));
			}
			assert (pTD->phPen);
			::SelectObject (_hDC, * (pTD->phPen));
			::Polygon (_hDC, _pts, cPoints);

			bDrawLabel = true;
		}
		if (element.fLength*g_kY < 2)
			// The element is too short.
			continue;

		if (_rgn.dwFlags & rgn_t::fWaypointOrTrack) {
			::SelectObject (_hDC, g_hTrackPen);
			::Polyline (_hDC, _pts, cPoints);
		} else {
			// Draw polyline.
			assert (pTD->phPen);
			::SelectObject (_hDC, * (pTD->phPen));

			::Polyline (_hDC, _pts, cPoints);
			if (pTD->phPen2) {
				::SelectObject (_hDC, * (pTD->phPen2));
				::Polyline (_hDC, _pts, cPoints);
			}
		}

		bDrawLabel = true;
	}

	// Draw POI.
	if (pTD->phCursor)
		for (size_t cOrigin = 0; cOrigin < _rgn.origins.size (); ++ cOrigin) {
			const rgn_origin_t & origin = _rgn.origins [cOrigin];
			if (origin.bDeleted)
				continue;

			if (! (_rgn.dwFlags & rgn_t::fWaypointOrTrack) && ! LevelIsVisible (origin.cLevel))
				continue;

			const int x = XToScreen (origin.point.x);
			const int y = YToScreen (origin.point.y);

			if (_rgn.dwFlags & rgn_t::fWaypointOrTrack) {
				::DrawIconEx (_hDC, x - g_nCursorShiftX_Waypoint, y - g_nCursorShiftY_Waypoint, g_hcurWaypoint, 32, 32, 0, NULL, DI_NORMAL);
			} else
				::DrawIconEx (_hDC, x - pTD->nCursorShiftX, y - pTD->nCursorShiftY, * pTD->phCursor, 32, 32, 0, NULL, DI_NORMAL);

			bDrawLabel = true;
		}

	return bDrawLabel;
}

void map_t::Draw (HDC _hDC) const {
	RECT rect;
	rect.left = 0;
	rect.top = 0;
	rect.right = g_lWidth;
	rect.bottom = g_lHeight;

	if (rectBound.Invalid ()) {
		::FillRect (_hDC, & rect, g_hTmpBrush);
		return;
	}

	const bool bShowLabels = GetSetting ("ShowLabels", true);
	const bool bTransparentPolygons = GetSetting ("TransparentPolygons", true);

	SetStatus ("Rendering...");
	SetProgressStatus (0);

	rect_t rectVisible;
	GetVisibleRect (rectVisible);

	// Erase background
	::FillRect (_hDC, & rect, g_hWhiteBrush);

	// Draw the attached images first.
	for (images_t::const_iterator iImage = images.begin (); iImage != images.end (); ++ iImage) {
		const image_t & image = * iImage;

		if (! image.rect.Intersects (rectVisible) || image.bDeleted)
			continue;

		const long lImageX = XToScreen (image.rect.x0);
		const long lImageY = YToScreen (image.rect.y1);
		const long lImageW = fast_floor (image.rect.Width  ()*g_kX + .5);
		const long lImageH = fast_floor (image.rect.Height ()*g_kY + .5);
		if (lImageW >= 5 && lImageH >= 5) {
			bool bDrawBlank = false;
			if (lImageW >= image.lWidth*14 || lImageH >= image.lHeight*14) {
				// TODO: ???
				bDrawBlank = true;
			} else if (lImageW >= image.lWidth && lImageH >= image.lHeight) {
				// Use fast approximation for magnification case.
				// NOTE: HALFTONE is rather slow in this case.
				::SetStretchBltMode (_hDC, COLORONCOLOR);
				bDrawBlank = ! ::StretchBlt (_hDC, lImageX, lImageY, lImageW, lImageH, image.hMemDC, 0, 0, image.lWidth, image.lHeight, SRCCOPY);
			} else if (lImageW > image.lSmallWidth && lImageH > image.lSmallHeight) {
				// Use high quality approximation for the stretching case.
				::SetStretchBltMode (_hDC, HALFTONE);
				bDrawBlank = ! ::StretchBlt (_hDC, lImageX, lImageY, lImageW, lImageH, image.hMemDC, 0, 0, image.lWidth, image.lHeight, SRCCOPY);
			} else if (lImageW > 50 && lImageH > 50) {
				if (! image.bOverlay) {
					::SetStretchBltMode (_hDC, HALFTONE);
					bDrawBlank = ! ::StretchBlt (_hDC, lImageX, lImageY, lImageW, lImageH, image.hSmallMemDC, 0, 0, image.lSmallWidth, image.lSmallHeight, SRCCOPY);
				}
			} else
				bDrawBlank = true;

			if (bDrawBlank && ! image.bOverlay) {
				RECT rect;
				rect.left   = lImageX;
				rect.top    = lImageY;
				rect.right  = lImageX + lImageW;
				rect.bottom = lImageY + lImageH;
				::FillRect (_hDC, & rect, g_hGrayBrush);
			}
		}
	}

	::SetROP2 (_hDC, R2_COPYPEN);
	::SetBkMode (_hDC, TRANSPARENT);
	::SelectObject (_hDC, g_hFont);

	::SetBrushOrgEx (_hDC, -g_lX, g_lY, NULL);
	
	POINT * const pts = reinterpret_cast<POINT *> (::alloca (sizeof (POINT)*cMaxPointsInElement));

	// The list of regions with labels to draw them in second pass.
	typedef std::list<const rgn_t *> listLabeledRgns_t;
	listLabeledRgns_t listLabeledRgns;
	size_t cLabels = 0;
/*
	const size_t cCellWidth = 100;
	const size_t cCellHeight = 20;
	const size_t cCellsX = g_lWidth/cCellWidth + 1;
	const size_t cCellsY = g_lHeight/cCellHeight + 1;
	bool * const pLabelCells = reinterpret_cast<bool *> (::alloca (sizeof (bool)*cCellsX*cCellsY));
	for (size_t cCells = 0; cCells < cCellsX*cCellsY; ++ cCells)
		pLabelCells [cCells] = false;
*/
	// Draw regions (w/o labels).
	for (int cLayer = 3; cLayer >= 0; -- cLayer) {
		const rgns_t & rgns = layers [cLayer].rgns;
		if (rgns.empty ())
			continue;

		const size_t cRgns = layers [cLayer].cRgns;
		size_t cRgn = 0;
		for (rgns_t::const_iterator iRgn = rgns.begin (); iRgn != rgns.end (); ++ iRgn, ++ cRgn) {
			if (g_bStop)
				return;

			const rgn_t & rgn = * iRgn;
			if (rgn.IsDeleted ())
				continue;

			// Rough intersection check.
			if (rgn.rectBound.Intersects (rectVisible)) {
				const bool bDrawLabel = DrawRgn (_hDC, rgn, pts, bTransparentPolygons);

				// Prepare labels to be drawn in the 2nd pass.
				if (bDrawLabel && ! rgn.strLabel.empty () && bShowLabels) {
					bool bAddLabel = true;
/*
					if (rgn.bPtLabelDefined) {
						const int x = XToScreen (rgn.ptLabel.x);
						const int y = YToScreen (rgn.ptLabel.y);
						if (0 <= x && x < g_lWidth && 0 <= y && y < g_lHeight) {
							const size_t cIndexX = ((x + g_lX)/cCellWidth) % cCellsX;
							const size_t cIndexY = ((y + g_lY)/cCellHeight) % cCellsY;
							const size_t cIndex = cIndexX*cCellsY + cIndexY;
							assert (cIndex < cCellsX*cCellsY);
							if (pLabelCells [cIndex])
								bAddLabel = false;
							else
								pLabelCells [cIndex] = true;
						}
					}
*/
					// Exclude labels whose width > region's size.
					if (rgn.cLabelWidth != 0 && rgn.fSize2 != 0)
						bAddLabel = rgn.cLabelWidth*rgn.cLabelWidth < rgn.fSize2*g_kY*g_kY;

					if (bAddLabel) {
						listLabeledRgns.push_back (& rgn);
						++ cLabels;
					}
				}
			}

			if (bShowLabels)
				SetProgressStatus (fast_floor (12.5*cRgn/cRgns + 12.5*(3 - cLayer) + .5));
			else
				SetProgressStatus (fast_floor (25.*cRgn/cRgns + 25*(3 - cLayer) + .5));
		}
	}

	// Draw labels in the 2nd pass.
	if (bShowLabels) {
		const size_t cMaxLabels = g_lWidth*g_lHeight/2000;
		const size_t cExcess = cMaxLabels ? cLabels/cMaxLabels + 1 : cLabels;
		assert (cExcess >= 1);

		size_t cLabel = 0;
		for (listLabeledRgns_t::const_iterator i = listLabeledRgns.begin (); i != listLabeledRgns.end (); ++ i, ++ cLabel) {
			const rgn_t & rgn = ** i;

			if ((reinterpret_cast<size_t> (& rgn)/sizeof (rgn)) % cExcess != 0)
				continue;

			rgn.pTypeDesc->pfDrawLabel (_hDC, rgn);

			SetProgressStatus (fast_floor (50 + 50.*cLabel/cLabels + .5));
		}
	}

	SetProgressStatus (100);
	SetStatus ("Ready");
}

//////////////////////////////////////////

static
void DrawSelection (HDC _hDC, const rect_t & _rectVisible, const find_info_t & _Selection) {
	if (_Selection.pImage) {
		const image_t & image = * _Selection.pImage;

		if (! _rectVisible.Intersects (image.rect))
			return;

		const long lX = XToScreen (image.rect.x0);
		const long lY = YToScreen (image.rect.y1);
		const long lW = fast_floor (image.rect.Width  ()*g_kX + .5);
		const long lH = fast_floor (image.rect.Height ()*g_kY + .5);

		::SetBkMode (_hDC, TRANSPARENT);
		::SetBkColor (_hDC, c_colWhite);
		::SelectObject (_hDC, g_hSelectionBrush);
		::PatBlt (_hDC, lX, lY, lW, lH, PATINVERT);
		
		return;
	}

	if (! _Selection.pLayer)
		return;
	const rgn_t * const pSelectedRgn = & * _Selection.iRgn;

	CCritSec::CLock lock (g_cs);

	if (pSelectedRgn->IsDeleted ())
		return;
	if (! _rectVisible.Intersects (pSelectedRgn->rectBound))
		return;

	const type_desc_t * const pTD = pSelectedRgn->pTypeDesc;

	POINT * const pts = (POINT *) alloca (sizeof (POINT)*g_map.cMaxPointsInElement);

	const rgn_elements_t & elements = pSelectedRgn->elements;
	for (size_t cElement = 0; cElement < elements.size (); ++ cElement) {	
		const rgn_element_t & element = elements [cElement];
		if (element.bDeleted)
			continue;

		const points_t & points = element.points;
		const size_t cPoints = points.size ();

		if (_Selection.cElement != -1 && _Selection.cElement != cElement)
			continue;

		if (! (pSelectedRgn->dwFlags & rgn_t::fWaypointOrTrack) && ! LevelIsVisible (element.cLevel))
			continue;

		if (! _rectVisible.Intersects (element.rectBound))
			continue;

		// Calculate screen coordinates.
		for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
			const point_t & pt = points [cPoint];
			pts [cPoint].x = XToScreen (pt.x);
			pts [cPoint].y = YToScreen (pt.y);
		}

		if (pSelectedRgn->IsPolygon ()) {
			// Draw polygon.
			::SelectObject (_hDC, g_hNullPen);
			::SelectObject (_hDC, g_hSelectionBrush);
			::SetBkMode (_hDC, TRANSPARENT);
			::SetROP2 (_hDC, R2_XORPEN);

			::Polygon (_hDC, pts, cPoints);

			::SetROP2 (_hDC, R2_COPYPEN);
		} else {
			// Draw polyline.
			::SelectObject (_hDC, g_hSelectionPen);
			::SetBkMode (_hDC, OPAQUE);
			::SetBkColor (_hDC, c_colWhite);
			::SetROP2 (_hDC, R2_COPYPEN);

			::Polyline (_hDC, pts, cPoints);
		}
	}

	// Draw POI.
	if (pTD->phCursor) {
		const rgn_origins_t & origins = pSelectedRgn->origins; 
		for (size_t cOrigin = 0; cOrigin < origins.size (); ++ cOrigin) {
			const rgn_origin_t & origin = origins [cOrigin];
			if (origin.bDeleted)
				continue;

			if (_Selection.cOrigin != -1 && _Selection.cOrigin != cOrigin)
				continue;

			if (! (pSelectedRgn->dwFlags & rgn_t::fWaypointOrTrack) && ! LevelIsVisible (origin.cLevel))
				continue;

			if (! _rectVisible.Contains (origin.point))
				continue;

			// Calculate screen coordinates for POI and label.
			const int orig_x = XToScreen (origin.point.x);
			const int orig_y = YToScreen (origin.point.y);

			::SelectObject (_hDC, g_hBlackThinPen);
			::SelectObject (_hDC, g_hNullBrush);
			::SetBkMode (_hDC, OPAQUE);
			::SetBkColor (_hDC, c_colWhite);
			
			::Ellipse (_hDC, orig_x - 8, orig_y - 8, orig_x + 9, orig_y + 9);
		}
	}
}

void DrawSelection (HDC _hDC, const find_info_m_t & _Selection) {
	rect_t rectVisible;
	GetVisibleRect (rectVisible);

	for (size_t cSelItem = 0; cSelItem < g_Selection.size (); ++ cSelItem)
		DrawSelection (_hDC, rectVisible, g_Selection [cSelItem]);
}
