# include "stdafx.h"
# include "Map.h"
# include "Types.h"
# include "Datum.h"
# include "Load.h"
# include "Log.h"
# include "Status.h"
# include "Globals.h"

# include "GarminImg.h"

class CGarminImgLoader : public CMapLoader {
  public:
	// Override.
	bool IsValid (const char * _pData, size_t _cFileSize) const;
	bool Load (const char * _pData, size_t _cFileSize, const char * _strFileName, map_t * _pMap);

  private:
	void AddPoint (size_t _zone, USHORT _usType, size_t _ushKind, float _x, float _y, const char * _strLabel);
	void AddPoly  (size_t _zone, USHORT _usType, size_t _ushKind, points_t & _points, const char * _strLabel);

	void lfdecode (BYTE * _datasec, size_t _dataleng, tre2asec_t * tre2asec, size_t k, BYTE _bl1, points_t & _points, float _x0, float _y0);

	map_t * m_pMap;
	const char * m_pStart, * m_pEnd;

	// The counter of regions.
	size_t m_cOrdinal;

	file_refs_t m_file_refs;
};

//DECLARE_MAP_LOADER (CGarminImgLoader, "img", "Garmin map (*.img)\0*.img", false);

bool CGarminImgLoader::IsValid (const char * _pData, size_t _cFileSize) const {
	if (_cFileSize < 0x200)
		return false;

	// Decode the signature.
	const BYTE btXORMask = _pData [0x10] ^ BYTE ('D');
	char strSignature [6];
	for (size_t c = 0; c < sizeof (strSignature); ++ c)
		strSignature [c] = _pData [0x10 + c] ^ btXORMask;

	// Check the signature.
	if (::memcmp (strSignature, "DSKIMG", sizeof (strSignature)) == 0)
		return true;

	return false;
}

void CGarminImgLoader::AddPoint (size_t _zone, USHORT _usType, size_t _ushKind, float _x, float _y, const char * _strLabel) {
	assert (_zone < 4);

	layer_t & layer = m_pMap->layers [_ushKind];
	++ layer.cRgns;
	layer.rgns.push_back (rgn_t ());
	rgn_t & rgn = layer.rgns.back ();
	// Update the load order list.
	m_pMap->InsertRgnToSaveList (rgn);

	// Initialize the rgn.
	rgn.cBeginPos = -1;
	rgn.cFileLine = -1;
	rgn.cOrdinal = m_cOrdinal ++;
	rgn.pFileRef = & m_file_refs.back ();

	rgn.ushKind = _ushKind;
	rgn.ushType = _usType;
	rgn.pTypeDesc = GetTypeDesc (_usType, _ushKind);

	if (_strLabel)
		rgn.strLabel = _strLabel;

	rgn.origins.push_back (rgn_origin_t ());
	rgn_origin_t & origin = rgn.origins.back ();
	origin.cFileLine = -1;
	origin.cFileOffset = -1;
	origin.cLevel = _zone;
	origin.point.x = _x;
	origin.point.y = _y;
}

void CGarminImgLoader::AddPoly (size_t _zone, USHORT _usType, size_t _ushKind, points_t & _points, const char * _strLabel) {
	assert (_zone < 4);

	layer_t & layer = m_pMap->layers [_ushKind];
	++ layer.cRgns;
	layer.rgns.push_back (rgn_t ());
	rgn_t & rgn = layer.rgns.back ();
	// Update the load order list.
	m_pMap->InsertRgnToSaveList (rgn);

	// Initialize the rgn.
	rgn.cBeginPos = -1;
	rgn.cFileLine = -1;
	rgn.cOrdinal = m_cOrdinal ++;
	rgn.pFileRef = & m_file_refs.back ();

	rgn.ushKind = _ushKind;
	rgn.ushType = _usType;
	rgn.pTypeDesc = GetTypeDesc (_usType, _ushKind);

	if (_strLabel)
		rgn.strLabel = _strLabel;

	rgn.elements.push_back (rgn_element_t ());
	rgn_element_t & element = rgn.elements.back ();
	element.cFileLine = -1;
	element.cFileOffset = -1;
	element.cLevel = _zone;
	element.points.swap (_points);
}

bool CGarminImgLoader::Load (const char * _pData, size_t _cFileSize, const char * _strMapFileName, map_t * _pMap) {
	m_pMap   = _pMap;
	m_pStart = _pData;
	m_pEnd   = _pData + _cFileSize;

	m_file_refs.clear ();
	m_file_refs.push_back (file_ref_t ());
	m_file_refs.back ().strFileName = _strMapFileName;
	m_file_refs.back ().bAttachment = false;

	m_cOrdinal = 0;

	SetProgressStatus (0);

	//
	// Check protection.
	//

	// TODO: ask XOR code.
	const BYTE btXORMask = _pData [0x10] ^ BYTE ('D');
	if (btXORMask) {
		ReportWarning ("The map is protected.");
		return false;
	}

	//
	// Parse file header.
	//
	const CImgHeader * const pImgHeader = reinterpret_cast<const CImgHeader *> (_pData);

	const size_t cBlockSize = pImgHeader->GetBlockSize ();

	DWORD dwRGN_Start = 0, dwRGN_Length = 0;
	DWORD dwTRE_Start = 0, dwTRE_Length = 0;
	DWORD dwLBL_Start = 0, dwLBL_Length = 0;

	// Scan header sections.
	size_t i;
	for (i = 0x600; i < pImgHeader->dwDataOffset; i += sizeof (CImgIndexBlock)) {
		assert (sizeof (CImgIndexBlock) == 0x200);
		const CImgIndexBlock * const pIndexBlock = reinterpret_cast<const CImgIndexBlock *> (_pData + i);

		if (pIndexBlock->btOne != 1)
			break;

		if (pIndexBlock->btIndex != 0)
			continue;

		const size_t cStart = pIndexBlock->wSize*cBlockSize;
        const size_t cLength = pIndexBlock->dwLength;

		if (::memcmp (pIndexBlock->strType, "RGN", 3) == 0) {
			if (dwRGN_Start != 0) {
				ReportWarning (i, "Unexpected format: table RGN0 is duplicated.");
				return false;
			}
            dwRGN_Start = cStart;
            dwRGN_Length = cLength;
		} else if (::memcmp (pIndexBlock->strType, "TRE", 3) == 0) {
			if (dwTRE_Start != 0) {
				ReportWarning (i, "Unexpected format: table TRE0 is duplicated.");
				return false;
			}
            dwTRE_Start = cStart;
            dwTRE_Length = cLength;
		} else if (::memcmp (pIndexBlock->strType, "LBL", 3) == 0) {
			if (dwLBL_Start != 0) {
				ReportWarning (i, "Unexpected format: table LBL0 is duplicated.");
				return false;
			}
            dwLBL_Start = cStart;
            dwLBL_Length = cLength;
		}
	}
	if (dwRGN_Start == 0) {
		ReportWarning (0x600, "Unexpected format: table RGN0 is not found.");
		return false;
	}
	if (dwTRE_Start == 0) {
		ReportWarning (0x600, "Unexpected format: table TRE0 is not found.");
		return false;
	}
	if (dwLBL_Start == 0) {
		ReportWarning (0x600, "Unexpected format: table LBL0 is not found.");
		return false;
	}

	//
	// TRE part.
	//

	// Check TRE header.
	const CTreDataBlockHead * const pTRE_Header = reinterpret_cast<const CTreDataBlockHead *> (_pData + dwTRE_Start);
    const char * strInTre = _pData + dwTRE_Start + 0x6; // 6 bytes
    if (::memcmp (pTRE_Header->strLabel, "GARMIN TRE", sizeof (pTRE_Header->strLabel)) != 0) {
        ReportWarning (dwTRE_Start, "Unexpected format: TRE header is not found.");
		return false;
	}
	if (pTRE_Header->wHeaderLength != 0x78)
        ReportWarning ("Unexpected size of TRE header: %d bytes.", pTRE_Header->wHeaderLength);

	rect_t rectMapBound;
    rectMapBound.y1 = hexgrad_3byte (pTRE_Header->b3MapNorth);
	rectMapBound.x1 = hexgrad_3byte (pTRE_Header->b3MapEast);
	rectMapBound.y0 = hexgrad_3byte (pTRE_Header->b3MapSouth);
	rectMapBound.x0 = hexgrad_3byte (pTRE_Header->b3MapWest);

	// Feld fr TRE-Bereiche
	struct tresec_t {
		long lStart;
		long lSize;
		long dwExtra;
	};
	tresec_t tresec [7];
    // lschen Feld tresec
	::memset (tresec, 0, sizeof (tresec));
    // einlesen der TRE Bereichswerte Start/Lnge
//    tresec [0].lStart  = * reinterpret_cast<const long *> (_pData + dwTRE_Start + 0x21);
//    tresec [0].lSize   = * reinterpret_cast<const long *> (_pData + dwTRE_Start + 0x25);

    tresec [1].lStart  = * reinterpret_cast<const long *> (_pData + dwTRE_Start + 0x29);
    tresec [1].lSize   = * reinterpret_cast<const long *> (_pData + dwTRE_Start + 0x2D);

    tresec [2].lStart  = * reinterpret_cast<const long *> (_pData + dwTRE_Start + 0x31);
    tresec [2].lSize   = * reinterpret_cast<const long *> (_pData + dwTRE_Start + 0x35);
    tresec [2].dwExtra = * reinterpret_cast<const long *> (_pData + dwTRE_Start + 0x39);

    tresec [6].lStart  = * reinterpret_cast<const long *> (_pData + dwTRE_Start + 0x3F);
    tresec [6].lSize   = * reinterpret_cast<const long *> (_pData + dwTRE_Start + 0x43);

    tresec [3].lStart  = * reinterpret_cast<const long *> (_pData + dwTRE_Start + 0x4A);
    tresec [3].lSize   = * reinterpret_cast<const long *> (_pData + dwTRE_Start + 0x4E);
    tresec [3].dwExtra = * reinterpret_cast<const long *> (_pData + dwTRE_Start + 0x52);

    tresec [4].lStart  = * reinterpret_cast<const long *> (_pData + dwTRE_Start + 0x58);
    tresec [4].lSize   = * reinterpret_cast<const long *> (_pData + dwTRE_Start + 0x5C);
    tresec [4].dwExtra = * reinterpret_cast<const long *> (_pData + dwTRE_Start + 0x60);

    tresec [5].lStart  = * reinterpret_cast<const long *> (_pData + dwTRE_Start + 0x66);
    tresec [5].lSize   = * reinterpret_cast<const long *> (_pData + dwTRE_Start + 0x6A);
    tresec [5].dwExtra = * reinterpret_cast<const long *> (_pData + dwTRE_Start + 0x6E);

	long j, k, l, n;

	//*** TRE1 einlesen in tre1sec(x,y)
    // Anzahl eintrge in TRE1
	const size_t cZones = pTRE_Header->dwZonesLength/4;
	tre1sec_t * const tre1sec = new tre1sec_t [cZones]; // Feld fr TRE1:Zonenwerte
    j = cZones - 1;             // Reset
    for (i = dwTRE_Start + pTRE_Header->dwZonesOffset; i < dwTRE_Start + pTRE_Header->dwZonesOffset + pTRE_Header->dwZonesLength; i += 4) {
        tre1sec [j].lev = (BYTE) _pData [i];
        tre1sec [j].zf  = (BYTE) _pData [i + 1];
        tre1sec [j].n   = (DWORD (_pData [i + 3]) << 8) + _pData [i + 2];
        -- j;
    }

    // lange Eintrge in TRE2 zhlen
    long tre2lang = 0; // Anzahl der langen Eintrge in TRE2
	for (i = 2; i <= cZones; ++ i)
		tre2lang += tre1sec [i - 1].n;

	// *** TRE2 einlesen in tre2sec(x)
	//Feld fr TRE2:Zonendefinition
    const size_t c_tre2sec_size = tresec [1].lSize - 1;   // Lnge TRE2
    BYTE * tre2sec = (BYTE*) (_pData + dwTRE_Start + tresec [1].lStart);
/*
	// *** TRE3 einlesen in tre1sec(x)
    j = tresec [2].lSize/3;    // Anzahl eintrge in TRE3
    long * tre3sec = (long*)alloca (j); //Feld festlegen
    long k = 0;               // Reset
    for (i = dwTRE_Start + tresec [2].lStart; i <= dwTRE_Start + tresec [2].lStart + tresec [2].lSize - 1; i += 3) {
        tre3sec [k] = (DWORD (_pData [i + 2]) << 16) + (DWORD (_pData [i + 1]) << 8) + _pData [i];
        ++ k;           // Feldindex +1
    }*/

	// *** TRE4 einlesen in tre4sec(x)
	const size_t c_tre4sec_size = tresec [3].lSize/2;    // Anzahl eintrge in TRE4
    tre4sec_t * tre4sec = new tre4sec_t [c_tre4sec_size];     // Feld festlegen
    k = 0;                  // Reset
    for (i = dwTRE_Start + tresec [3].lStart; i <= dwTRE_Start + tresec [3].lStart + tresec [3].lSize - 1; i += 2) {
        tre4sec [k].bt1 = _pData [i];
        tre4sec [k].bt2 = _pData [i + 1];
        ++ k;
    }

	// *** TRE5 einlesen in tre5sec(x)
    const size_t c_tre5sec_size = tresec [4].lSize/2;   // Anzahl eintrge in TRE5
	tre5sec_t * tre5sec = new tre5sec_t [c_tre5sec_size]; // Feld festlegen
    k = 0;                  // Reset
    for (i = dwTRE_Start + tresec [4].lStart; i <= dwTRE_Start + tresec [4].lStart + tresec [4].lSize - 1; i += 2) {
        tre5sec [k].bt1 = _pData [i];
        tre5sec [k].bt2 = _pData [i + 1];
        ++ k;
    }

	// *** TRE6 einlesen in tre6sec(x)
    const size_t c_tre6sec_size = tresec [5].lSize/3;    // Anzahl eintrge in TRE6
	tre6sec_t * tre6sec = new tre6sec_t [c_tre6sec_size]; // Feld festlegen
    j = 0;                   // Reset
    for (i = dwTRE_Start + tresec [5].lStart; i <= dwTRE_Start + tresec [5].lStart + tresec [5].lSize - 1; i += 3) {
        BYTE hexb2 = _pData [i];
        BYTE hexb3 = _pData [i + 1];
        BYTE hexb1 = _pData [i + 2];
        tre6sec [j].sh1 = (DWORD (hexb2) << 8) + hexb1;
        tre6sec [j].sh2 = hexb3; // && 0xFF
        ++ j;           // Feldindex +1
    }

	//*** TRE2 dekodieren
    k = 1;
	const size_t c_tre2asec_size = tre2lang + tre1sec [0].n + 1;
	tre2asec_t * tre2asec = new tre2asec_t [c_tre2asec_size + 1]; //Feld fr TRE2 dekodiert
    //Lange Eintrge 16 byte(alle auer erste Zone)
    for (i = 0; i <= tre2lang*16 - 1; i += 16) {
		assert (k <= c_tre2asec_size);

        //+rgn1 offset
        tre2asec [k].l1 = DWORD(tre2sec [i + 2]) * 65536 + DWORD(tre2sec [i + 1]) * 256 + DWORD(tre2sec [i]);
        //inhalt
        tre2asec [k].l2 = tre2sec [i + 3];
        //ostwert
        tre2asec [k].l3 = DWORD(tre2sec [i + 6]) * 65536 + DWORD(tre2sec [i + 5]) * 256 + DWORD(tre2sec [i + 4]);
        //nordwert
        tre2asec [k].l4 = DWORD(tre2sec [i + 9]) * 65536 + DWORD(tre2sec [i + 8]) * 256 + DWORD(tre2sec [i + 7]);
        //zusatz 1
        tre2asec [k].l5 = DWORD(tre2sec [i + 11]) * 256 + DWORD(tre2sec [i + 10]);
        //zusatz 2
        tre2asec [k].l6 = DWORD(tre2sec [i + 13]) * 256 + DWORD(tre2sec [i + 12]);
        //zusatz 3
        tre2asec [k].l7 = DWORD(tre2sec [i + 15]) * 256 + DWORD(tre2sec [i + 14]);
        ++ k;
    }
    // Kurze Eintrge 14 byte(nur erste Zone)
    for (i = tre2lang*16; i <= tresec [1].lSize - 15; i += 14) {
        //+rgn1 offset
        tre2asec [k].l1 = DWORD(tre2sec [i + 2]) * 65536 + DWORD(tre2sec [i + 1]) * 256 + DWORD(tre2sec[i]);
        //inhalt
        tre2asec [k].l2 = tre2sec [i + 3];
        //ostwert
        tre2asec [k].l3 = DWORD(tre2sec [i + 6]) * 65536 + DWORD(tre2sec [i + 5]) * 256 + DWORD(tre2sec [i + 4]);
        //nordwert
        tre2asec [k].l4 = DWORD(tre2sec [i + 9]) * 65536 + DWORD(tre2sec [i + 8]) * 256 + DWORD(tre2sec [i + 7]);
        //zusatz 1
        tre2asec [k].l5 = DWORD(tre2sec [i + 11]) * 256 + DWORD(tre2sec [i + 10]);
        //zusatz 2
        tre2asec [k].l6 = DWORD(tre2sec [i + 13]) * 256 + DWORD(tre2sec [i + 12]);
        ++ k;
    }
    // Ende-Eintrag
	assert (k == c_tre2asec_size);
    // +rgn1 offset
    tre2asec [k].l1 = (DWORD (tre2sec [i + 2]) << 16) + (DWORD (tre2sec [i + 1]) << 8) + DWORD(tre2sec [i]);
    // inhalt
    tre2asec [k].l2 = tre2sec [i + 3];

    // Zonennummer+Zf
    k = 1;
//	m_pMap->Zoom.resize (c_tre1sec_size);
//	m_pMap->Levels.resize (c_tre1sec_size);
    for (i = cZones; i >= 1; -- i) {   //je Zone
		const BYTE lev = tre1sec [i - 1].lev;
		const BYTE zf =  tre1sec [i - 1].zf;

//		m_pMap->Zoom [i - 1] = lev;
//		m_pMap->Levels [i - 1] = zf;

        for (j = 1; j <= tre1sec [i - 1].n; ++ j) {    //Eintrag in TRE2
            tre2asec [k].l8 = lev;
            tre2asec [k].l9 = 1UL << (32 - zf);
            ++ k;
        }
    }

	//
	// LBL Abschnitt
	//

    // Test ob wirklich LBL
	const char * strInLbl = _pData + dwLBL_Start + 0x6; // 6 bytes
    if (::memcmp (strInLbl, "IN LBL", 6) != 0) {
        ReportWarning (dwLBL_Start, "kein LBL Bereich ab: %6s", strInLbl);
		return false;
	}
	// LBL Bereiche in lblsec-Feld eintragen
	short abheader = * reinterpret_cast<const short *> (_pData + dwLBL_Start);
    if (abheader <= 0) {
        ReportWarning (dwLBL_Start, "Wert von LBL-Header falsch: %d", abheader);
		return false;
	}

    // LBL unbekannt
	BYTE lblubk [6];             // unbekannte Teile von lbl
	::memcpy (lblubk, _pData + dwLBL_Start + 0xE, sizeof (lblubk));

	struct lblsec_t {
		long l1; // Start
		long l2; // Size
		long l3; // Extra 1
		long l4; // Extra 2
	};
	lblsec_t * lblsec = new lblsec_t [16];   // Feld fr LBL-Bereiche
	::memset (lblsec, 0, sizeof (sizeof (lblsec_t)*16));

    // einlesen der LBL Bereichswerte Start/Lnge
    lblsec [1].l1 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x15);
    lblsec [1].l2 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x19); 
    long tmplong1 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x1D);
    lblsec [1].l3 = tmplong1 - (short (tmplong1/65536)*65536);

    lblsec [2].l1 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x1F); 
    lblsec [2].l2 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x23); 
    lblsec [2].l3 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x27); 

    lblsec [3].l1 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x2D);
    lblsec [3].l2 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x31);
    lblsec [3].l3 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x35); 

    lblsec [4].l1 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x3B); 
    lblsec [4].l2 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x3F); 
    lblsec [4].l3 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x43);

    lblsec [5].l1 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x49);
    lblsec [5].l2 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x4D);
    lblsec [5].l3 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x51);

    lblsec [6].l1 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x57);
    lblsec [6].l2 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x5B);
    lblsec [6].l3 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x5F);

    lblsec [7].l1 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x64);
    lblsec [7].l2 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x68);
    lblsec [7].l3 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x6C);

    lblsec [8].l1 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x72);
	lblsec [8].l2 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x76);
    lblsec [8].l3 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x7A);

    lblsec [9].l1 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x80);
    lblsec [9].l2 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x84);
    lblsec [9].l3 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x88);

    lblsec [10].l1 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x8E);
    lblsec [10].l2 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x92);
    lblsec [10].l3 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x96);

    lblsec [11].l1 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0x9C);
    lblsec [11].l2 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0xA0);
    lblsec [11].l3 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0xA4);

    if (abheader > 0xAA) {
        lblsec [12].l1 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0xAA);
        tmplong1       = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0xAE);
        lblsec [12].l2 = tmplong1 - (short (tmplong1/65536)*65536);
    }

	// String American
	const char * lbl13 = NULL; // 9 bytes.
    if (abheader > 0xB0) {
        lblsec [13].l1 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0xB0);
        lblsec [13].l2 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0xB4);
		// *** lbl13
        lbl13 = _pData + dwLBL_Start + lblsec [13].l1;
    }

    if (abheader > 0xB8) {
        lblsec [14].l1 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0xB8);
        lblsec [14].l2 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0xBC);
        lblsec [14].l3 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0xC0);
    }
    if (abheader > 0xC4) {
        lblsec [15].l1 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0xC4);
        lblsec [15].l2 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0xC8);
        lblsec [15].l3 = * reinterpret_cast<const long *> (_pData + dwLBL_Start + 0xCC);
    }

	//*** LBL1 nach lbl1sec(n) laden
    long abstart = dwLBL_Start + lblsec [1].l1;
    long ableng = lblsec [1].l2;
    // einlesen Bytes aus Datei in Feld
    const BYTE * lbl1sec = reinterpret_cast<const BYTE *> (_pData + abstart); // Size=ableng.

	const size_t cBitsPerSymbol = lblsec [1].l3/256;
    if (cBitsPerSymbol != 6) {
        ReportWarning ("Only 6-bit coding is supported. Here is %d-bit one.", cBitsPerSymbol);
		return false;
	}

	std::map<long, std::string> labels;

    i = 0; // index fr bytes 
	j = 1; // j=index fr Texte
	BYTE bl1 = 0;

    while (i < ableng) {
        std::string lbl1text;
        tmplong1 = i;    // i=offset (merken 1. Byte im Text)
z1byte:
		SetProgressStatus (10.*i/ableng);
		if (g_bStop)
			return false;

        // 1. Zeichen aus (i)
        const short z1 = (lbl1sec [i] & 0xFC)/4;
        if (z1 > 0x2F) {
			if (tmplong1 != 0)
				labels [tmplong1] = lbl1text;
            ++ i;
			++ j;
			continue;
        }
        lbl1text += lblzeichen (bl1, z1);

        // 2. Zeichen aus (i) und (i+1)
        const short z2 = (lbl1sec [i] & 0x3) * 16 + (lbl1sec [i + 1] & 0xF0) / 16;
        if (z2 > 0x2F) {
			if (tmplong1 != 0)
				labels [tmplong1] = lbl1text;
            ++ i;
			++ j;
			continue;
        }
        lbl1text += lblzeichen (bl1, z2);

        // 3. Zeichen aus (i+1) und (i+2)
        const short z3 = (lbl1sec [i + 1] & 0xF) * 4 + (lbl1sec [i + 2] & 0xC0) / 64;
        if (z3 > 0x2F) {
			if (tmplong1 != 0)
				labels [tmplong1] = lbl1text;
            i += 2;
			++ j;
			continue;
        }
        lbl1text += lblzeichen (bl1, z3);

        // 4. Zeichen aus (i+2)
        const short z4 = lbl1sec [i + 2] & 0x3F;
        if (z4 > 0x2F) {
			if (tmplong1 != 0)
				labels [tmplong1] = lbl1text;
            i += 3;
			++ j;
			continue;
        }
        lbl1text += lblzeichen (bl1, z4);
        i += 3;
        goto z1byte; //nchsten drei Byte in selben Text
	}

	SetProgressStatus (10);

	//
	// RGN Abschnitt
	//

    const char * strInRgn = _pData + dwRGN_Start + 0x6; // 6 bytes
    if (::memcmp (strInRgn, "IN RGN", 6) != 0) {
        ReportWarning (dwRGN_Start, "kein RGN Bereich: %6s", strInRgn);
		return false;
	}
	// Headerlnge von RGN
	abheader = * reinterpret_cast<const short *> (_pData + dwRGN_Start);
    if (abheader != 0x1D) {
		ReportWarning ("Headerlnge von RGN nicht 1D: 0x%04x", abheader);
		return false;
	}

	// RGN unbekannt
	BYTE rgnubk [6];               // unbekannte Teile von rgn
	::memcpy (rgnubk, _pData + dwRGN_Start + 0xE, sizeof (rgnubk));

    // lschen Feld rgnsec
    // einlesen der LBL Bereichswerte Start/Lnge
	//Feld fr TRE-Bereiche
    const long rgnsec00 = * reinterpret_cast<const long *> (_pData + dwRGN_Start + 0x15);
	const long rgnsec01 = * reinterpret_cast<const long *> (_pData + dwRGN_Start + 0x19);

	//*** RGN1 nach rgn1sec(n) laden
    abstart = dwRGN_Start + rgnsec00;
    ableng = rgnsec01;
    BYTE * rgn1sec = (BYTE *) (_pData + abstart); // Size= ableng

	//*** RGN1 Abschnitte in tre2asec eintragen (start)
    for (k = 1; (size_t)k <= c_tre2asec_size - 1; ++ k) {  //je tre2asec Eintrag, nicht den Endeeintrag
		if (g_bStop)
			return false;

        for (i = 10; i <= 17; ++ i) {
            tre2asec [k].l10 = -1;
			tre2asec [k].l11 = -1;
			tre2asec [k].l12 = -1;
			tre2asec [k].l13 = -1;
			tre2asec [k].l14 = -1;
			tre2asec [k].l15 = -1;
			tre2asec [k].l16 = -1;
			tre2asec [k].l17 = -1;
		}

        if (tre2asec [k].l2 == 0)
			continue;
		//RGN leer

        short z1 = 0, z2 = 0, z3 = 0, z4 = 0;

        if (tre2asec [k].l2 & 0x80) z1 = 1;      //Flchen
        if (tre2asec [k].l2 & 0x40) z2 = 1;      //Linien
        if (tre2asec [k].l2 & 0x20) z3 = 1;      //Orte
        if (tre2asec [k].l2 & 0x10) z4 = 1;      //Punkte
        bl1 = z1 + z2 + z3 + z4 - 1;   //Anzahl der Zeiger

        long zeig1 = bl1 > 0 ? 256*DWORD (rgn1sec [tre2asec [k].l1 + 1]) + rgn1sec [tre2asec [k].l1] : 0;
        long zeig2 = bl1 > 1 ? 256*DWORD (rgn1sec [tre2asec [k].l1 + 3]) + rgn1sec [tre2asec [k].l1 + 2] : 0;
        long zeig3 = bl1 > 2 ? 256*DWORD (rgn1sec [tre2asec [k].l1 + 5]) + rgn1sec [tre2asec [k].l1 + 4] : 0;

    // Punkte ?
        if (z4 == 1)  // punkte da
            tre2asec [k].l10 = bl1*2 + tre2asec [k].l1; // beginn

    // Orte ?
        if (z3 == 1) {      // orte da
            if (z4 == 0)   // keine Punkte vor orten
                tre2asec [k].l12 = bl1*2 + tre2asec [k].l1; //beginn
            else            //punkte vor orten
                tre2asec [k].l12 = zeig1 + tre2asec [k].l1;  //beginn
        }

    // Linien ?
        if (z2 == 1) {      // linien da
            const BYTE bl2 = tre2asec [k].l2 & 0x30;    // maskieren orte+punkte
            if (bl2 == 0)     //keine punkte+orte vor linien
                tre2asec [k].l14 = bl1*2 + tre2asec [k].l1; //beginn
            else if (bl2 == 0x10 || bl2 == 0x20)  //punkte oder orte vor linien
                tre2asec [k].l14 = zeig1 + tre2asec [k].l1;  //beginn
            else if (bl2 == 0x30)         //punkte und orte vor linien
                tre2asec [k].l14 = zeig2 + tre2asec [k].l1;  //beginn
        }

    // Flchen ?
        if (z1 == 1) {      // flchen da
            const BYTE bl2 = tre2asec [k].l2 & 0x70;    //maskieren orte+punkte+linien
            if (bl2 == 0)     // keine punkte+orte+linien vor flchen
                tre2asec [k].l16 = bl1 * 2 + tre2asec [k].l1; // beginn
            else if (bl2 == 0x10 || bl2 == 0x20 || bl2 == 0x40) // punkte, orte oder linien vor flchen
                tre2asec [k].l16 = zeig1 + tre2asec [k].l1;  // beginn
            else if (bl2 == 0x30 || bl2 == 0x50 || bl2 == 0x60) // zwei ber. vor linien
                tre2asec [k].l16 = zeig2 + tre2asec [k].l1;  // beginn
            else if (bl2 == 0x70)     // punkte+orte+linien vor flchen
                tre2asec [k].l16 = zeig3 + tre2asec [k].l1;  // beginn
        }
	}

	//*** RGN1 Abschnitte in tre2asec eintragen (lnge)
    for (k = 1; (size_t)k <= c_tre2asec_size - 1; ++ k) {   //je tre2asec Eintrag, nicht den Endeeintrag
		if (g_bStop)
			return false;

        if (tre2asec [k].l2 == 0)
			continue;  //RGN leer

    //punkte
        if (tre2asec [k].l10 != -1) {    // punkte da
            if (tre2asec [k].l12 != -1)  // orte dahinter
                tre2asec [k].l11 = tre2asec [k].l12 - tre2asec [k].l10;
            else if (tre2asec [k].l14 != -1) // linien dahinter
                tre2asec [k].l11 = tre2asec [k].l14 - tre2asec [k].l10;
            else if (tre2asec [k].l16 != -1) // flchen dahinter
                tre2asec [k].l11 = tre2asec [k].l16 - tre2asec [k].l10;
            else
                tre2asec [k].l11 = tre2asec [k + 1].l1 - tre2asec [k].l10;
        }
    //orte
        if (tre2asec [k].l12 != -1) {    // orte da
            if (tre2asec [k].l14 != -1) // linien dahinter
                tre2asec [k].l13 = tre2asec [k].l14 - tre2asec [k].l12;
            else if (tre2asec [k].l16 != -1) // flchen dahinter
                tre2asec [k].l13 = tre2asec [k].l16 - tre2asec [k].l12;
            else
                tre2asec [k].l13 = tre2asec [k + 1].l1 - tre2asec [k].l12;
        }
    //linien
        if (tre2asec [k].l14 != -1) {   // linien da
            if (tre2asec [k].l16 != -1) // flchen dahinter
                tre2asec [k].l15 = tre2asec [k].l16 - tre2asec [k].l14;
            else
                tre2asec [k].l15 = tre2asec [k + 1].l1 - tre2asec [k].l14;
		}
    //flchen
        if (tre2asec [k].l16 != -1)    // flchen da
            tre2asec [k].l17 = tre2asec [k + 1].l1 - tre2asec [k].l16;
    }

	SetProgressStatus (15);

////////////////////////////

//Punkte
    for (k = 1; k <= c_tre2asec_size - 1; ++ k) {   // je tre2asec Eintrag, nicht den Endeeintrag
		SetProgressStatus (15 + 5.*k/c_tre2asec_size);
		if (g_bStop)
			return false;

        // RGN leer oder keine punkte oder nicht die Zone 00
        if (tre2asec [k].l2 == 0)
			continue;
        if (tre2asec [k].l10 == -1)
			continue;
		if (tre2asec [k].l8 != tre1sec [0].lev)
			continue;

		const long maxwert = tre2asec [k].l11/9;
        for (i = tre2asec [k].l10; i <= tre2asec [k].l10 + tre2asec [k].l11 - 1; l += 9) {
            // nord und ost
            const long y = wertlong (tre2asec, 256 * DWORD (rgn1sec [i + 7]) + rgn1sec [i + 6], k, tre2asec [k].l4);
            const long x = wertlong (tre2asec, 256 * DWORD (rgn1sec [i + 5]) + rgn1sec [i + 4], k, tre2asec [k].l3);
			const float fy = hexgrad_3byte (((BYTE *) & y) + 1);
			const float fx = hexgrad_3byte (((BYTE *) & x) + 1);

            // symbol
            long type = 256 * DWORD(rgn1sec [i]) + rgn1sec [i + 8];

            // Zone fr symbol aus tre6
			long zone = 0;
            for (l = 0; l <= c_tre6sec_size - 1; ++ l) {
                if (tre6sec [l].sh1 == type) {
                    zone = tre6sec [l].sh2;
                    break;
                }
            }

			const long lLabelOffset = 65536*DWORD (rgn1sec [i + 3]) + 256*DWORD (rgn1sec [i + 2]) + rgn1sec [i + 1];
			const char * strLabel = lLabelOffset ? labels [lLabelOffset].c_str () : NULL;

			for (size_t z = 0; z <= zone; ++ z)
				AddPoint (zone, type, 0, fx, fy, strLabel);
        }
    }
	SetProgressStatus (20);


// orte
	for (k = 1; (size_t)k <= c_tre2asec_size - 1; ++ k) {   //je tre2asec Eintrag, nicht den Endeeintrag
		SetProgressStatus (20 + 5*k/c_tre2asec_size);
		if (g_bStop)
			return false;

        // RGN leer oder keine orte oder nicht die Zone 00
        if (tre2asec [k].l2 == 0)
			continue;
        if (tre2asec [k].l12 == -1)
			continue;
		if (tre2asec [k].l8 != tre1sec [0].lev)
			continue;

		const long maxwert = tre2asec [k].l13/8; // anzahl punkte/orte
		for (i = tre2asec [k].l12; i <= tre2asec [k].l12 + tre2asec [k].l13 - 1; i += 8) {
            const long y = wertlong (tre2asec, 256 * DWORD (rgn1sec [i + 7]) + rgn1sec [i + 6], k, tre2asec [k].l4);
            const long x = wertlong (tre2asec, 256 * DWORD (rgn1sec [i + 5]) + rgn1sec [i + 4], k, tre2asec [k].l3);
			const float fy = hexgrad_3byte (((BYTE *) & y) + 1);
			const float fx = hexgrad_3byte (((BYTE *) & x) + 1);

            // symbol
            const long type = rgn1sec [i];

            // Zone fr symbol aus tre6
			long zone = 0;
            for (long l = 0; (size_t)l <= c_tre6sec_size - 1; ++ l)
                if (tre6sec [l].sh1 == 256 * type) {
                    zone = tre6sec [l].sh2;
                    break;
                }

			const long lLabelOffset = 65536*DWORD (rgn1sec [i + 3]) + 256*DWORD (rgn1sec [i + 2]) + rgn1sec [i + 1];
			const char * strLabel = lLabelOffset ? labels [lLabelOffset].c_str () : NULL;

			for (size_t z = 0; z <= zone; ++ z)
				AddPoint (z, type, 1, fx, fy, strLabel);
		}
	}
	SetProgressStatus (25);

// linien
    for (k = 1; (size_t)k <= c_tre2asec_size - 1; ++ k) { // je tre2asec Eintrag, nicht den Endeeintrag
		SetProgressStatus (25 + 5*k/c_tre2asec_size);
		if (g_bStop)
			return false;

        // RGN leer oder keine linien oder nicht die Zone 00
        if (tre2asec [k].l2 == 0)
			continue;
        if (tre2asec [k].l14 == -1)
			continue;
		if (tre2asec [k].l8 != tre1sec [0].lev)
			continue;

        // linien da
		i = tre2asec [k].l14;                          // startwert
		const long ableng = tre2asec [k].l14 + tre2asec [k].l15;   // endwert

        // 1. linienzeile, 1 mal je linie
        // Nummer
nextline:
		points_t points;

        // ost und nord
        const long y = wertlong (tre2asec, 256 * DWORD(rgn1sec [i + 7]) + rgn1sec [i + 6], k, tre2asec [k].l4); // nord
        const long x = wertlong (tre2asec, 256 * DWORD(rgn1sec [i + 5]) + rgn1sec [i + 4], k, tre2asec [k].l3); // ost
		const float fy = hexgrad_3byte (((BYTE *) & y) + 1);
		const float fx = hexgrad_3byte (((BYTE *) & x) + 1);

		point_t point;
		point.x = fx;
		point.y = fy;
		points.push_back (point);

        // linientyp
		const long type = rgn1sec [i] & 0x3F; // 2^7 und 2^6 entfernen

        // Zone fr typ aus tre4
		long zone = 0;
        for (l = 0; (size_t)l <= c_tre4sec_size - 1; ++ l)
            if (tre4sec [l].bt1 == type) {
                zone = tre4sec [l].bt2;
                break;
            }

        //text offset
		const long lLabelOffset = 65536*DWORD (rgn1sec [i + 3]) + 256*DWORD (rgn1sec [i + 2]) + rgn1sec [i + 1];
		const char * strLabel = lLabelOffset ? labels [lLabelOffset].c_str () : NULL;

        //ENDE der 1. linienzeile
        //Datenfeldlnge
		long dataleng; 
        if ((rgn1sec [i] & 0x80) == 0) {  // 1 byte lnge
            dataleng = rgn1sec [i + 8];   //datenlnge
            i += 9;                   //i auf beschreibung
        } else {                            //2 byte lnge
            dataleng = 256 * DWORD(rgn1sec [i + 9]) + rgn1sec [i + 8]; //datenlnge
            i += 10;                  //i auf beschreibung
        }
        // Datenfelddefinition
        bl1 = rgn1sec [i];
        // Datenfeld nach datasec()
		const size_t c_datasec_size = dataleng;
        BYTE * datasec = new BYTE [dataleng];  // Feld fr linien/flchen Daten
        n = 0;       // zhler in datasec
        ++ i;       // i=1. datenbyte
        dataleng += i;     // endezeiger
        while (i < dataleng) {
            datasec [n] = rgn1sec [i]; //daten aus rgn1sec
            ++ i;
			++ n;
        }

        // umrechnen der daten und ausgabe
        lfdecode (datasec, c_datasec_size - 1, tre2asec, k, bl1, points, fx, fy);
		for (size_t z = 0; z <= zone; ++ z) {
			points_t points2 = points;
			AddPoly (z, type, 2, points2, strLabel);
		}

        i = dataleng;  //i=nchste linie
        //test ob noch linien da
        if (i < ableng)
			goto nextline;
    }
	SetProgressStatus (30 + 5*k/c_tre2asec_size);

// flchen
    for (k = 1; k <= c_tre2asec_size - 1; ++ k) {   // je tre2asec Eintrag, nicht den Endeeintrag
		SetProgressStatus (30 + 20*k/c_tre2asec_size);
		if (g_bStop)
			return false;

        // RGN leer oder keine flchen oder nicht die Zone 00
        if (tre2asec [k].l2 == 0)
			continue;
        if (tre2asec [k].l16 == -1)
			continue;
		if (tre2asec [k].l8 != tre1sec [0].lev)
			continue;

        // flchen da
        i = tre2asec [k].l16;                          // startwert
        const long ableng = tre2asec [k].l16 + tre2asec [k].l17;   // endwert

        // 1. flchenzeile, 1 mal je flche
        // Nummer
nextfla:
		points_t points;

        // ost und nord
        const long y = wertlong (tre2asec, 256 * DWORD(rgn1sec [i + 7]) + rgn1sec [i + 6], k, tre2asec [k].l4); // nord
        const long x = wertlong (tre2asec, 256 * DWORD(rgn1sec [i + 5]) + rgn1sec [i + 4], k, tre2asec [k].l3); // ost
		const float fy = hexgrad_3byte (((BYTE *) & y) + 1);
		const float fx = hexgrad_3byte (((BYTE *) & x) + 1);

		point_t point;
		point.x = fx;
		point.y = fy;
		points.push_back (point);

        // flchentyp
        const long type = rgn1sec [i] & 0x7F;    // 2^7 entfernen

        // Zone fr typ aus tre5
		long zone = 0;
		for (l = 0; l <= c_tre5sec_size - 1; ++ l)
            if (tre5sec [l].bt1 == type) {
                zone = tre5sec [l].bt2;
				break;
			}

        // text offset
		const long lLabelOffset = 65536*DWORD (rgn1sec [i + 3]) + 256*DWORD (rgn1sec [i + 2]) + rgn1sec [i + 1];
		const char * strLabel = lLabelOffset ? labels [lLabelOffset].c_str () : NULL;

        // ENDE der 1. flchenzeile
        // Datenfeldlnge
		long dataleng;
        if ((rgn1sec [i] & 0x80) == 0) { // 1 byte lnge
            dataleng = rgn1sec [i + 8];   // datenlnge
            i += 9;                   // i auf beschreibung
        } else {                           // 2 byte lnge
            dataleng = 256 * DWORD(rgn1sec [i + 9]) + rgn1sec [i + 8]; // datenlnge
            i += 10;                  // i auf beschreibung
        }
        // Datenfelddefinition
        bl1 = rgn1sec [i];
        // Datenfeld nach datasec()
		const size_t c_datasec_size = dataleng;
        BYTE * datasec = new BYTE [dataleng];  // Feld fr linien/flchen Daten

        n = 0;           // zhler in datasec
        ++ i;             // i=1. datenbyte
        dataleng += i;     // endezeiger
        while (i < dataleng) {
            datasec [n] = rgn1sec [i]; // daten aus rgn1sec
            ++ i;
			++ n;
        }

        // umrechnen der daten und ausgabe
        lfdecode (datasec, c_datasec_size - 1, tre2asec, k, bl1, points, fx, fy);
		if (type != 0x4b) {
/*			for (size_t z = 0; z <= 2; ++ z) {
				points_t points2 = points;
				AddPoly (z, type, 3, points2, strLabel);
			}*/
			AddPoly (zone, type, 3, points, strLabel);
		}

        // diag
        i = dataleng;  // i=nchste flche

        // test ob noch flche da
        if (i < ableng)
			goto nextfla;
    }


	//
	//
	//

	SetProgressStatus (50);

	m_pMap->bNoHeader = false;

	return true;
}

static inline
void save_point (points_t & _points, float & _x0, float & _y0, long _x, long _y) {
	_x0 += hexgrad_3byte (((BYTE *) & _x) + 1);
	_y0 += hexgrad_3byte (((BYTE *) & _y) + 1);

	point_t point;
    point.x = _x0;
	point.y = _y0;
	_points.push_back (point);
}

//bl2=testfeld ob irgendein bit=1
// bl1=datenfeldbeschreibung (byte)
// bitn/o=anzahl bits fr nord/ost (byte)
// n=bitzhler (0-524xxx)
// vzn/o=vorzeichen:-1=fix(-);+1=fix(+);0=variabel (byte)
// k=tre2asec nr fr zonenfaktor
// beznord/ost=bezug fr 1. punkt
// istnord/ost=dif-wert fr ausgabe (long)
// tmpnord/ost=rechen-wert (long)
// datasec=daten der linie/flche in byte
// bitsec=feld fr bitstrom (max 524280 bit// S)
void CGarminImgLoader::lfdecode (
	BYTE * datasec, size_t _datasec_size, tre2asec_t * tre2asec, size_t k, BYTE bl1,
	points_t & _points, float x0, float y0
) {
	const size_t c_bitsec_size = (_datasec_size + 1) * 8 - 1;
    BYTE * bitsec = new BYTE [c_bitsec_size + 1];
	CByteArrayHolder bah (bitsec);

    // fllen des bitsec
    size_t n = 0;   // bitzhler
	size_t l;
    for (l = 0; l <= _datasec_size; ++ l) {   // databyte schleife
        bitsec [n]     = datasec [l] & 1;
        bitsec [n + 1] = (datasec [l] & 2) / 2;
        bitsec [n + 2] = (datasec [l] & 4) / 4;
        bitsec [n + 3] = (datasec [l] & 8) / 8;
        bitsec [n + 4] = (datasec [l] & 16) / 16;
        bitsec [n + 5] = (datasec [l] & 32) / 32;
        bitsec [n + 6] = (datasec [l] & 64) / 64;
        bitsec [n + 7] = (datasec [l] & 128) / 128;
        n += 8;
    }

    n = 0;   // reset bitzhler
    const long zf = tre2asec [k].l9;

    BYTE bitn, bito; // ?? number of bits ??
    float vzn, vzo; // May be 0, 1, -1.

// vorzeichen ost bestimmen
    if (bitsec [n] == 0) {   // variabel
        vzo = 0;
        bito = 1;
        ++ n;   // n auf vznord
    } else if (bitsec [n] == 1 && bitsec [n + 1] == 0) { // plus
        vzo = 1;
        bito = 0;
        n += 2;   // n auf vznord
    } else if (bitsec [n] == 1 && bitsec [n + 1] == 1) { // minus(kein 2k)
        vzo = -1;
        bito = 0;
        n += 2;   // n auf vznord
    }

// vorzeichen nord bestimmen
    if (bitsec [n] == 0) {   // variabel
        vzn = 0;
        bitn = 1;
        ++ n;   // n auf 1. datenbit
    } else if (bitsec [n] == 1 && bitsec [n + 1] == 0) { // plus
        vzn = 1;
		bitn = 0;
        n += 2;   // n auf 1. datenbit
    } else if (bitsec [n] == 1 && bitsec [n + 1] == 1) { // minus(kein 2k)
        vzn = -1;
		bitn = 0;
        n += 2;   // n auf 1. datenbit
    }

// bitanzahl
	{
		BYTE bittmp1 = (bl1 & 0xF0) / 16;
		if (bittmp1 > 9)
			bittmp1 = 2 * bittmp1 - 9;
		bitn += bittmp1 + 2;
	} {
		BYTE bittmp2 = (bl1 & 0xF);
		if (bittmp2 > 9)
			bittmp2 = 2 * bittmp2 - 9;
		bito += bittmp2 + 2;
	}

// konstanten (wert fr erkennung sonferfall)
    const float tmpsing1 = 1UL << (bito - 1);
    const float tmpsing2 = 1UL << (bitn - 1);

    long istost = 0;
    long tmpost = 0;
    long istnord = 0;
	long tmpnord = 0;

// auswahl der dekodierung
    if (vzo == 0 && vzn == 0) goto dek_ov_nv;
    if (vzo == 0 && vzn != 0) goto dek_ov_nf;
    if (vzo != 0 && vzn == 0) goto dek_of_nv;
    if (vzo != 0 && vzn != 0) goto dek_of_nf;
    ReportWarning ("wrong Suffix (+/-) !!!");
    return;

// ost=variabel nord=variabel
dek_ov_nv:
    // ostwert nach tmpost
    istost = 0;
o1:
	tmpost = 0;
    for (l = 0; l < bito; ++ l) {   // bitdurchlauf fr ostwert
		assert (n <= c_bitsec_size);
        tmpost |= bitsec [n]*(1UL << l);
        ++ n;       // nchstes bit
    }
    // test auf neg und solderfall
	assert (tmpost < 2*tmpsing1);
    if (tmpost < tmpsing1)   // pos
        istost += tmpost;
    else if (tmpost == tmpsing1) {    // sonderfall
        istost += tmpost - 1;
        goto o1;
    } else if (tmpost > tmpsing1)   // neg(2k)
        istost = (tmpost - 2*tmpsing1) - istost;

    // nordwert nach tmpnord
    istnord = 0;
n1:
	tmpnord = 0;
    for (l = 0; l < bitn; ++ l) {   // bitdurchlauf fr nordwert
		assert (n <= c_bitsec_size);
        tmpnord |= bitsec [n]*(1UL << l);
        ++ n;       // nchstes bit
    }
    // test auf neg und solderfall
	assert (tmpnord < 2*tmpsing2);
    if (tmpnord < tmpsing2)   // pos
        istnord += tmpnord;
    else if (tmpnord == tmpsing2) {    // sonderfall
        istnord += tmpnord - 1;
        goto n1;
    } else if (tmpnord > tmpsing2)   // neg(2k)
        istnord = (tmpnord - 2*tmpsing2) - istnord;

    // ausgabe
	istost *= zf;
	istnord *= zf;
    save_point (_points, x0, y0, istost, istnord);

    // haben noch ein ost+nordwert platz ?
    if (n + bito + bitn - 1 <= c_bitsec_size)
		goto dek_ov_nv;

	assert (int (c_bitsec_size - 8) <= int (n - 1) && n - 1 <= c_bitsec_size);
    return;


// ost=variabel nord=fix
dek_ov_nf:
    // ostwert nach tmpost
    istost = 0;
o2:
	tmpost = 0;
    for (l = 0; l <= bito - 1; ++ l) {  // bitdurchlauf fr ostwert
		assert (n <= c_bitsec_size);
        tmpost |= bitsec [n]*(1UL << l);
        ++ n;       // nchstes bit
    }
    // test auf neg und solderfall
	assert (tmpost < 2*tmpsing1);
    if (tmpost < tmpsing1)   // pos
        istost += tmpost;
    else if (tmpost == tmpsing1) {    // sonderfall
        istost += tmpost - 1;
        goto o2;
    } else if (tmpost > tmpsing1)     // neg(2k)
        istost = (tmpost - 2*tmpsing1) - istost;

    // nordwert nach istnord
    istnord = 0;
    for (l = 0; l <= bitn - 1; ++ l) {   // bitdurchlauf fr nordwert
		assert (n <= c_bitsec_size);
        istnord |= bitsec [n]*(1UL << l);
        ++ n;       // nchstes bit
    }
    istnord *= vzn;

    // ausgabe
	istost *= zf;
	istnord *= zf;
    save_point (_points, x0, y0, istost, istnord);

    // haben noch ein ost+nordwert platz ?
    if (n + bito + bitn - 1 <= c_bitsec_size)
		goto dek_ov_nf;

	assert (int (c_bitsec_size - 8) <= int (n - 1) && n - 1 <= c_bitsec_size);
    return;


// ost=fix nord=variabel
dek_of_nv:
    // ostwert nach istost
    istost = 0;
    for (l = 0; l <= bito - 1; ++ l) {   // bitdurchlauf fr ostwert
		assert (n <= c_bitsec_size);
        istost |= bitsec [n]*(1UL << l);
        ++ n;       // nchstes bit
    }
    istost *= vzo;

    // nordwert nach tmpnord
    istnord = 0; // zum zusammenzhlen beim sonderfall
n2:
	tmpnord = 0;
    for (l = 0; l <= bitn - 1; ++ l) {   // bitdurchlauf fr nordwert
		assert (n <= c_bitsec_size);
        tmpnord |= bitsec [n]*(1UL << l);
        ++ n;       // nchstes bit
    }
    // test auf neg und solderfall
	assert (tmpnord < 2*tmpsing2);
    if (tmpnord < tmpsing2)   // pos
        istnord += tmpnord;
    else if (tmpnord == tmpsing2) {      // sonderfall
        istnord += tmpnord - 1; // merken
        goto n2;                         // nchster nordwert !
    } else if (tmpnord > tmpsing2)   // neg(2k)
        istnord = (tmpnord - 2*tmpsing2) - istnord;

    // ausgabe
	istost *= zf;
	istnord *= zf;
    save_point (_points, x0, y0, istost, istnord);

    // haben noch ein ost+nordwert platz ?
    if (n + bito + bitn - 1 <= c_bitsec_size)
		goto dek_of_nv;

	assert (int (c_bitsec_size - 8) <= int (n - 1) && n - 1 <= c_bitsec_size);
    return;


// ost=fix nord=fix
dek_of_nf:
    // ostwert nach istost
    istost = 0;
    for (l = 0; l <= bito - 1; ++ l) {   // bitdurchlauf fr ostwert
		assert (n <= c_bitsec_size);
        istost |= bitsec [n]*(1UL << l);
        ++ n;       // nchstes bit
    }
    istost *= vzo;

    // nordwert nach istnord
    istnord = 0;
    for (l = 0; l <= bitn - 1; ++ l) {   // bitdurchlauf fr nordwert
		assert (n <= c_bitsec_size);
        istnord |= bitsec [n]*(1UL << l);
        ++ n;       // nchstes bit
    }
    istnord *= vzn;

    // ausgabe
	istost *= zf;
	istnord *= zf;
	save_point (_points, x0, y0, istost, istnord);

    // haben noch ein ost+nordwert platz ?
    if (n + bito + bitn - 1 <= c_bitsec_size)
		goto dek_of_nf;

	assert (int (c_bitsec_size - 8) <= int (n - 1) && n - 1 <= c_bitsec_size);
}
