# include "stdafx.h"
# include "Map.h"
# include "Types.h"
# include "Load.h"
# include "Log.h"
# include "Status.h"
# include "Globals.h"

//////////////////////////////////////////

class CMSExportTxtLoader : public CMapLoader {
  public:
	// Override.
	bool IsValid (const char * _pData, size_t _cFileSize) const;
	bool Load (const char * _pData, size_t _cFileSize, const char * _strFileName, map_t * _pMap);

  private:
	void ParsePoint (const char * _strData, point_t & _point);
	void ParseWaypoint (const char * _strData, const char * _strEnd);
	void ParseTrack (const char * _strData, const char * _strEnd);
	void ParseTrackpoint (const char * _strData, const char * _strEnd);

	map_t * m_pMap;
	const char * m_pStart, * m_pEnd;

	// The counter of current line.
	size_t m_cFileLine;
	//
	size_t m_cOrdinal;

	points_t m_points;
	size_t m_cTrackLine;
	size_t m_cTrackBeginPos;
	std::string m_strTrackLabel;
	void SaveTrack (const char * _strEnd);

	file_refs_t m_file_refs;
};

DECLARE_MAP_LOADER (CMSExportTxtLoader, "txt", "MapSource exported text (*.txt)\0*.txt", false);

bool CMSExportTxtLoader::IsValid (const char * _pData, size_t _cFileSize) const {
	if (! CheckAnsiFile (_pData, _cFileSize, true))
		return false;

	bool bGrid = false, bDatum = false, bHeader = false;
	size_t cFileLine = 0;

	const char * pLine = _pData;
	do {
		const char * const pNextLine = FindNextLine (pLine, _pData + _cFileSize);

		if (HasPrefix (pLine, pNextLine, "Grid\t", 5))
			bGrid = true;
		else if (HasPrefix (pLine, pNextLine, "Datum\t", 6))
			bDatum = true;
		else if (HasPrefix (pLine, pNextLine, "Header\t", 7))
			bHeader = true;

		if (pNextLine [-1] == '\n')
			// Count lines.
			++ cFileLine;

		if (cFileLine > 10)
			return bGrid && bDatum && bHeader;

		pLine = pNextLine;
	} while (pLine < _pData + _cFileSize && ! g_bStop);

	return false;
}

	// Find end of field.
static
const char * FindNextField (const char * _pBegin, const char * _pEnd) {
	for (const char * p = _pBegin; p < _pEnd; ++ p)
		if (* p == '\t')
			return p + 1 < _pEnd ? p + 1 : _pEnd;

	return _pEnd;
}

void CMSExportTxtLoader::ParsePoint (const char * _strData, point_t & _point) {
	char chNS, chEW;
	float fLat = -200, fLon = -100;
	if (::sscanf (_strData, "%c%f %c%f", & chNS, & fLat, & chEW, & fLon) != 4) {
		ReportWarning (_strData - m_pStart, "Unexpected format of waypoint coordinate.");
		return;
	}
	if (chEW == 'W')
		fLon = -fLon;
	if (chNS == 'S')
		fLat = -fLat;

	_point.x = fLon;
	_point.y = fLat;
}

struct symbol_t {
	const char * strKey;
	ULONG        ulType;
};

const symbol_t g_symbols [] = {
	{"Building",             0x06402},
	{"Campground",           0x10048}, // 0x02b03, 0x04800
	{"Car",                  0x04d00}, // ???
	{"Car Repair",           0x02f03},
	{"Forest",               0x0660a},
	{"Museum",               0x02c02},
	{"Park",                 0x0640e}, // 0x04900 0x02c06
	{"Residence",            0x06402}, // ?
	{"Skull and Crossbones", 0x05700},
	{"Summit",               0x06616},
	{"Swimming Area",        0x05400}, // 0x02d09
	{"Truck Stop",           0},
	{"Waypoint",             0},//0x00100},
};

const size_t c_cSymbols = sizeof (g_symbols)/sizeof (g_symbols [0]);

void CMSExportTxtLoader::ParseWaypoint (const char * _strData, const char * _strEnd) {
	rgn_t rgn;
	rgn.cBeginPos = _strData - m_pStart;
	rgn.cEndPos = _strEnd - m_pStart;
	rgn.cFileLine = m_cFileLine;
	rgn.cOrdinal = m_cOrdinal ++;

	rgn.dwFlags |= rgn_t::fWaypointOrTrack;

	const char * strDescription = FindNextField (_strData, _strEnd);
	if (strDescription == _strEnd) {
		ReportWarning (_strData - m_pStart, "Unexpected format of waypoint.");
		return;
	}

	rgn.strLabel.assign (_strData, strDescription);

	const char * const strType        = FindNextField (strDescription, _strEnd);
	const char * const strPosition    = FindNextField (strType,        _strEnd);
	const char * const strAltitude    = FindNextField (strPosition,    _strEnd);
	const char * const strDepth       = FindNextField (strAltitude,    _strEnd);
	const char * const strProximity   = FindNextField (strDepth,       _strEnd);
	const char * const strDisplayMode = FindNextField (strProximity,   _strEnd);
	const char * const strColor       = FindNextField (strDisplayMode, _strEnd);
	const char * const strSymbol      = FindNextField (strColor,       _strEnd);
	const char * const strFacility    = FindNextField (strSymbol,      _strEnd);
	if (strFacility == _strEnd) {
		ReportWarning (_strData - m_pStart, "Unexpected format of waypoint.");
		return;
	}

	point_t point;
	ParsePoint (strPosition, point);

	rgn.origins.push_back (rgn_origin_t ());
	rgn_origin_t & origin = rgn.origins.back ();
	origin.cLevel = 0;
	origin.point = point;

	// Find best type for the waypoint.
	rgn.pTypeDesc = & g_types [0];
	rgn.ushKind = 0;
	for (size_t cSymbol = 0; cSymbol < c_cSymbols; ++ cSymbol)
		if (::_strnicmp (g_symbols [cSymbol].strKey, strSymbol, strFacility - strSymbol - 1) == 0) {
			const USHORT ushType = g_symbols [cSymbol].ulType & 0xFFFF;
			rgn.ushKind = g_symbols [cSymbol].ulType >> 16;
			assert (rgn.ushKind < 4);
			rgn.pTypeDesc = GetTypeDesc (ushType, rgn.ushKind);
			break;
		}

	rgn.pFileRef = & m_file_refs.back ();

	//
	// Add the waypoint to the map.
	//

	m_pMap->layers [rgn.ushKind].rgns.push_back (rgn);
	++ m_pMap->layers [rgn.ushKind].cRgns;
	m_pMap->InsertRgnToSaveList (m_pMap->layers [rgn.ushKind].rgns.back ());
}

void CMSExportTxtLoader::SaveTrack (const char * _strEnd) {
	if (m_points.empty ())
		return;

	//
	// Add the track to the map.
	//

	m_pMap->layers [2].rgns.push_back (rgn_t ());
	++ m_pMap->layers [2].cRgns;

	rgn_t & rgn = m_pMap->layers [2].rgns.back ();

	m_pMap->InsertRgnToSaveList (rgn);

	//
	//
	//

	rgn.dwFlags |= rgn_t::fWaypointOrTrack;

	rgn.cBeginPos = m_cTrackBeginPos;
	rgn.cEndPos = _strEnd - m_pStart;
	rgn.cFileLine = m_cTrackLine;
	rgn.cOrdinal = m_cOrdinal ++;

	rgn.strLabel.swap (m_strTrackLabel);

	rgn.elements.push_back (rgn_element_t ());
	rgn_element_t & element = rgn.elements.back ();
	element.cFileLine = m_cTrackLine;
	element.cFileOffset = m_cTrackBeginPos;

	element.cLevel = 0;
	element.points.swap (m_points);

	rgn.pTypeDesc = & g_types [0];
	rgn.ushKind = 2;

	rgn.pFileRef = & m_file_refs.back ();
}

void CMSExportTxtLoader::ParseTrack (const char * _strData, const char * _strEnd) {
	assert (m_points.empty ());

	m_cTrackLine = m_cTrackLine;
	m_cTrackBeginPos = _strData - m_pStart;

	const char * const strLabel = FindNextField (_strData, _strEnd);
	const char * const strTime = FindNextField (strLabel, _strEnd);
	if (strLabel == _strEnd || strTime == _strEnd) {
		ReportWarning (_strData - m_pStart, "Unexpected format of track definition.");
		m_strTrackLabel = "";
	} else
		m_strTrackLabel.assign (strLabel, strTime - 1);
}

void CMSExportTxtLoader::ParseTrackpoint (const char * _strData, const char * _strEnd) {
	point_t point;
	ParsePoint (_strData, point);

	m_points.push_back (point);
}

bool CMSExportTxtLoader::Load (const char * _pData, size_t _cFileSize, const char * _strMapFileName, map_t * _pMap) {
	if (! CheckAnsiFile (_pData, _cFileSize, false))
		return false;

	m_pMap = _pMap;
	m_pStart = _pData;
	m_pEnd = _pData + _cFileSize;

	m_cOrdinal = 0;
	m_cFileLine = 1;
	m_cTrackLine = 1;
	m_cTrackBeginPos = 0;

	m_file_refs.clear ();
	m_file_refs.push_back (file_ref_t ());
	m_file_refs.back ().strFileName = _strMapFileName;
	m_file_refs.back ().bAttachment = true;

	SetProgressStatus (0);

	const char * pLine = _pData;
	do {
		const char * const pNextLine = FindNextLine (pLine, m_pEnd);

		if (pLine [0] != '\n' && pLine [0] != '\r') {
			if (HasPrefix (pLine, pNextLine, "Trackpoint\t", 11))
				ParseTrackpoint (pLine + 11, pNextLine);
			else if (HasPrefix (pLine, pNextLine, "Waypoint\t", 9))
				ParseWaypoint (pLine + 9, pNextLine);
			else if (HasPrefix (pLine, pNextLine, "Track\t", 6)) {
				SaveTrack (pLine);
				ParseTrack (pLine, pNextLine);
			} else if (HasPrefix (pLine, pNextLine, "Header\t", 7))
				SaveTrack (pLine);
			else if (HasPrefix (pLine, pNextLine, "Grid\t", 5)) {
				const char * const strGrid = pLine + 5;
				if (::memcmp (strGrid, "Lat/Lon hddd.ddddd", pNextLine - strGrid - 1) != 0) {
					ReportWarning (pLine - m_pStart, "Grid format is not supported. 'Lat/Lon hddd.ddddd' required.");
					return false;
				}

				m_pMap->cHeaderLen = pNextLine - m_pStart;
			} else if (HasPrefix (pLine, pNextLine, "Datum\t", 6)) {
				const char * const strDatum = pLine + 6;
				if (::memcmp (strDatum, "WGS 84", pNextLine - strDatum - 1) != 0)
					ReportWarning (pLine - m_pStart, "Datums other than WGS84 are not supported.");

				m_pMap->cHeaderLen = pNextLine - m_pStart;
			}
		}

		if (pNextLine [-1] == '\n')
			// Count lines.
			++ m_cFileLine;

		pLine = pNextLine;

		SetProgressStatus (fast_floor ((50.*(pLine - _pData))/_cFileSize + .5));
	} while (pLine < _pData + _cFileSize && ! g_bStop);

	SaveTrack (m_pEnd);

	SetProgressStatus (50);

	m_pMap->file_refs.swap (m_file_refs);

	// Suppress warning about empty Zoom etc.
	m_pMap->bNoHeader = true;

	return true;
}
