# include "stdafx.h"
# include "Map.h"
# include "Types.h"
# include "Load.h"
# include "Log.h"
# include "Levels.h"
# include "Status.h"
# include "Label.h"
# include "Scale.h"
# include "Globals.h"

static
bool PartContainsPart (const rgn_element_t & _Outer, const rgn_element_t & _Inner) {
	if (
		_Outer.cLevel != _Inner.cLevel ||
		_Outer.fArea < _Inner.fArea ||
		! _Outer.rectBound.Contains (_Inner.rectBound)
	)
		return false;

	assert (! _Outer.points.empty ());
	assert (! _Inner.points.empty ());
	return PolygonContainsPt (_Outer.points, _Inner.points [0]) > 0;
}

/////////////////////////////////////////////////////

	// The precision of rgn search.
const float c_fE_in_pixels = 4;

bool rgn_t::Contains (size_t _cLevel, const point_t & _point) const {
	const float c_fE = c_fE_in_pixels/g_kX;

	if (! elements.empty ()) {
		for (rgn_elements_t::const_iterator i = elements.begin (); i != elements.end (); ++ i)
			if (i->bDeleted)
				continue;
			if (_cLevel == i->cLevel) {
				if (IsPolyline ()) {
					if (PolylineContainsPt (i->points, _point, c_fE))
						return true;
				} else {
					if (PolygonContainsPt (i->points, _point) >= 0)
						return true;
				}
			}
	} else if (! origins.empty ()) {
		for (rgn_origins_t::const_iterator i = origins.begin (); i != origins.end (); ++ i)
			if (i->bDeleted)
				continue;
			if (_cLevel == i->cLevel)
				if (PointContainsPt (i->point, _point, c_fE))
					return true;
	}

	return false;
}

void rgn_t::RefreshAttributes (map_t & _map) {
	// Explore parts.
	const size_t cElements = elements.size ();
	if (cElements) {
		// Data used to count islands and average element area.
		float fArea = 0;
		size_t cNonIslandParts = 1;
		size_t cOuterPart = 0;

		for (size_t cElement = 0; cElement < cElements; ++ cElement) {
			rgn_element_t & element = elements [cElement];
			if (element.bDeleted)
				continue;

			points_t & points = element.points;
			const size_t cPoints = points.size ();

			if (cPoints > _map.cMaxPointsInElement)
				_map.cMaxPointsInElement = cPoints;

			// Determine the bounding rect.
			for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
				point_t & point = points [cPoint];

				element.rectBound.Extend (point);
			}
			rectBound.Extend (element.rectBound);

			element.bIsland = false;

			// For polygons only.
			if (IsPolygon ()) {
				// NOTE: count element's area before islands detection.
				element.fArea = GetArea (points);

				// Detect 'islands'.
				// NOTE: 'islands' are possible in cMap.exe output only; they are not allowed for cGPSMap.
				if (cElement > 0 && cPoints >= 3) {
					if (PartContainsPart (elements [cOuterPart], element)) {
						element.bIsland = true;
					} else {
						cOuterPart = cElement;
						++ cNonIslandParts;
					}
				}

				if (element.bIsland)
					fArea -= element.fArea;
				else
					fArea += element.fArea;
			} else
				element.fArea = 0;

			element.fLength = GetLength (points);
		}

		fSize2 = GetRgnSize (* this);
		fAveElementArea = IsPolygon () ? fArea/cNonIslandParts : 0;
	}

	// Explore origins.
	for (size_t cOrigin = 0; cOrigin < origins.size (); ++ cOrigin) {
		const rgn_origin_t & origin = origins [cOrigin];
		if (origin.bDeleted)
			continue;

		// Include origins in the bounding rect.
		rectBound.Extend (origin.point);
	}

	_map.rectBound.Extend (rectBound);

	// Prepare point for the label, if any.
	if (! strLabel.empty ())
		pTypeDesc->pfPrepareLabel (* this);
}

bool rgn_t::IsDeleted () const {
	return (dwFlags & fDeleted) || (pFileRef && pFileRef->bDeleted);
}

bool rgn_t::IsPolygon () const {
	return pTypeDesc == g_types ? (ushKind == 3) : (pTypeDesc->ushKind == 3);
}

bool rgn_t::IsPolyline () const {
	return pTypeDesc == g_types ? (ushKind == 2) : (pTypeDesc->ushKind == 2);
}

bool rgn_t::IsPoint () const {
	return pTypeDesc == g_types ? (ushKind == 1) : (pTypeDesc->ushKind == 1);
}

bool rgn_t::IsPOI () const {
	return pTypeDesc == g_types ? (ushKind == 0) : (pTypeDesc->ushKind == 0);
}

float GetRgnSize (const rgn_t & _rgn) {
	const size_t cElements = _rgn.elements.size ();
	if (! cElements)
		return 0;

	float fCosY;
	point_t p0, p1;
	bool b0Defined = false, b1Defined = false;

	float d2max = 0;

	// Find the most distant point for the first one.
	size_t cElement;
	for (cElement = 0; cElement < cElements; ++ cElement) {
		const rgn_element_t & element = _rgn.elements [cElement];
		if (element.bDeleted)
			continue;

		const points_t & points = element.points;
		const size_t cPoints = points.size ();
		if (! cPoints)
			continue;

		for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
			const point_t & point = points [cPoint];

			if (! b0Defined) {
				p0 = point;
				b0Defined = true;
				fCosY = ::cos (point.y*c_PI/180.);
				continue;
			}

			const float d2 = (p0.x - point.x)*(p0.x - point.x)*fCosY*fCosY + (p0.y - point.y)*(p0.y - point.y);
			if (d2 > d2max) {
				d2max = d2;
				p1 = point;
				b1Defined = true;
			}
		}
	}

	if (! b1Defined)
		return 0;

	// Find the most distant point for the first most distant one.
	for (cElement = 0; cElement < cElements; ++ cElement) {
		const rgn_element_t & element = _rgn.elements [cElement];
		if (element.bDeleted)
			continue;

		const points_t & points = element.points;
		const size_t cPoints = points.size ();
		if (! cPoints)
			continue;

		for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
			const point_t & point = points [cPoint];

			const float d2 = (p1.x - point.x)*(p1.x - point.x)*fCosY*fCosY + (p1.y - point.y)*(p1.y - point.y);
			if (d2 > d2max)
				d2max = d2;
		}
	}

	return d2max;
}

//////////////////////////////////////////////////

CCritSec g_cs;
map_t g_map;

void map_t::Init () {
	cHeaderLen = 0;
	header_data.clear ();

	cRgnLimit = 700;
	Levels.clear ();
	Zoom.clear ();

	for (size_t cLayer = 0; cLayer < 4; ++ cLayer)
		layers [cLayer].clear ();

	for (images_t::iterator iImage = images.begin (); iImage != images.end (); ++ iImage)
		iImage->clear ();
	images.clear ();
	file_refs.clear ();
	bNoHeader = false;

	rectBound = rect_t ();
	cMaxPointsInElement = 0;
	fCosY = 1;

	pFirstRgn = NULL;
	pLastRgn = NULL;

	cPrecOrder = 5;
}

void map_t::Swap (map_t & _other) {
	std::swap (cHeaderLen, _other.cHeaderLen);
	header_data.swap (_other.header_data);

	std::swap (cRgnLimit, _other.cRgnLimit);
	Zoom.swap (_other.Zoom);
	Levels.swap (_other.Levels);

	for (size_t cLayer = 0; cLayer < 4; ++ cLayer)
		layers [cLayer].swap (_other.layers [cLayer]);

	images.swap (_other.images);
	file_refs.swap (_other.file_refs);
	std::swap (bNoHeader, _other.bNoHeader);

	rectBound.swap (_other.rectBound);

	std::swap (cMaxPointsInElement, _other.cMaxPointsInElement);
	std::swap (fCosY, _other.fCosY);

	std::swap (pFirstRgn, _other.pFirstRgn);
	std::swap (pLastRgn, _other.pLastRgn);

	std::swap (cPrecOrder, _other.cPrecOrder);
}

void map_t::Join (map_t & _other) {
	if (bNoHeader && ! _other.bNoHeader) {
		Zoom.swap (_other.Zoom);
		Levels.swap (_other.Zoom);

		cHeaderLen = _other.cHeaderLen;
		header_data.swap (_other.header_data);
		cRgnLimit = _other.cRgnLimit;
	} else {
		if (Zoom != _other.Zoom) {
			// TODO: join Zoom etc
		}
		// TODO: join headers.
	}
	bNoHeader = bNoHeader && _other.bNoHeader;

	//
	// Join layers.
	//

	for (size_t cLayer = 0; cLayer < 4; ++ cLayer) {
		rgns_t & rgns = layers [cLayer].rgns;
		if (rgns.empty ()) {
			layers [cLayer].swap (_other.layers [cLayer]);
		} else {
			rgns.splice (rgns.begin (), _other.layers [cLayer].rgns);

			layers [cLayer].cRgns += _other.layers [cLayer].cRgns;
			assert (layers [cLayer].cRgns == layers [cLayer].rgns.size ());
			_other.layers [cLayer].cRgns = 0;

			layers [cLayer].cDeletedRgns += _other.layers [cLayer].cDeletedRgns;
			_other.layers [cLayer].cDeletedRgns = 0;
		}
	}

	// Join images.
	JoinImages (images, _other.images);

	// Join attachments.
	if (file_refs.empty ())
		file_refs.swap (_other.file_refs);
	else
		file_refs.splice (file_refs.begin (), _other.file_refs);


	// Extend rectBound.
	rectBound.Extend (_other.rectBound);

	if (cMaxPointsInElement < _other.cMaxPointsInElement)
		cMaxPointsInElement = _other.cMaxPointsInElement;

	fCosY = ::cos (rectBound.MeanY ()*c_PI/180.);

	//
	// Splice the load lists.
	//

	if (pLastRgn)
		pLastRgn->pNextRgn = _other.pFirstRgn;
	pLastRgn = _other.pLastRgn;
	_other.pFirstRgn = NULL;
	_other.pLastRgn = NULL;

	if (cPrecOrder < _other.cPrecOrder)
		cPrecOrder = _other.cPrecOrder;
}

/*
struct CompareRgnArea {
	bool operator () (const rgn_t & _1, const rgn_t & _2) const {
		return _1.fAvePartArea > _2.fAvePartArea;
	}
};*/

bool map_t::LoadMap (const char * _pData, size_t _cFileSize, const char * _strFileName, CMapLoader * _pLoader) {
	Init ();

	SetStatus ("Loading map...");

	assert (_pLoader);
	if (! _pLoader->Load (_pData, _cFileSize, _strFileName, this)) {
		SetStatus ("Failed.");
		return false;
	}

	//
	// Post-process the map.
	//

	bool * const pbRgnsInLevel = reinterpret_cast<bool *> (::alloca (sizeof (bool)*Zoom.size ()));
	size_t cLevel;
	for (cLevel = 0; cLevel < Zoom.size (); ++ cLevel)
		pbRgnsInLevel [cLevel] = false;

	bool bThereAreIslands = false;

	if (layers [0].cRgns + layers [1].cRgns + layers [2].cRgns + layers [3].cRgns > 65535)
		ReportWarning ("Number of regions exceeds 65536 (cGPSMap limitation).");

	for (int cLayer = 3; cLayer >= 0; -- cLayer) {
		layer_t & layer = layers [cLayer];
		rgns_t & rgns = layer.rgns;
		const size_t cRgns = layer.cRgns;
		assert (cRgns == rgns.size ());

		size_t cRgn = 0;
		for (rgns_t::iterator iRgn = rgns.begin (); iRgn != rgns.end (); ++ iRgn, ++ cRgn) {
			if (g_bStop)
				return false;

			rgn_t & rgn = * iRgn;
			if (rgn.IsDeleted ())
				continue;

			// Explore parts.
			size_t cElements = rgn.elements.size ();
			if (cElements) {
				if (cElements > cRgnLimit)
					ReportWarning (rgn.cBeginPos, "RgnLimit=%d exceeded: region contains %d elements.", cRgnLimit, cElements);

				// Data used to count islands and average element area.
				float fArea = 0;
				size_t cNonIslandParts = 1;
				size_t cOuterPart = 0;

				for (size_t cElement = 0; cElement < cElements; ++ cElement) {
					rgn_element_t & element = rgn.elements [cElement];
					if (element.bDeleted)
						continue;

					points_t & points = element.points;
					size_t cPoints = points.size ();

					if (cPoints > cMaxPointsInElement)
						cMaxPointsInElement = cPoints;

					bool bRemoveElement = false;

					pbRgnsInLevel [element.cLevel] = true;

					// Determine the bounding rect.
					for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
						point_t & point = points [cPoint];

						element.rectBound.Extend (point);

						if (g_bStop)
							return false;
					}
					rgn.rectBound.Extend (element.rectBound);

					element.bIsland = false;

					// For polygons only.
					if (cLayer == 3) {
						// NOTE: count element's area before islands detection.
						element.fArea = GetArea (points);

						// Detect 'islands'.
						// NOTE: 'islands' are possible in cMap.exe output only; they are not allowed for cGPSMap.
						if (cElement > 0 && cPoints >= 3) {
							if (PartContainsPart (rgn.elements [cOuterPart], element)) {
								element.bIsland = true;
								bThereAreIslands = true;
							} else {
								cOuterPart = cElement;
								++ cNonIslandParts;
							}
						}

						if (element.bIsland)
							fArea -= element.fArea;
						else
							fArea += element.fArea;

						if (cPoints >= 3 && points [0] == points [cPoints - 1]) {
							ReportWarning (element.cFileOffset, "First and last points of polygon are equal. The last point will be ignored.");

							points.pop_back ();
							-- cPoints;
						}

						if (cPoints <= 2) {
							ReportWarning (element.cFileOffset, "Polygon element has just %d point(s). The element will be ignored.", cPoints);
							bRemoveElement = true;
						}
					} else
						element.fArea = 0;

					element.fLength = GetLength (points);

					// For polylines only.
					if (cLayer == 2) {
						if (cPoints <= 1) {
							ReportWarning (element.cFileOffset, "Polyline element has just %d point(s). The element will be ignored.", cPoints);
							bRemoveElement = true;
						}
					}

					if (cPoints > 16384)
						ReportWarning (element.cFileOffset, "Region' element has %d point(s). This is Win9x limitation for rendering.", cPoints);

					if (element.cLevel >= Levels.size () - 1)
						ReportWarning (element.cFileOffset, "Region' element has incorrect level %d", element.cLevel);

					if (bRemoveElement) {
						rgn.elements.erase (rgn.elements.begin () + cElement);
						-- cElement;
						-- cElements;
					}
				}

				rgn.fSize2 = GetRgnSize (rgn);

				if (cLayer == 3)
					rgn.fAveElementArea = fArea/cNonIslandParts;
				else
					rgn.fAveElementArea = 0;
			}

			// Explore origins.
			for (size_t cOrigin = 0; cOrigin < rgn.origins.size (); ++ cOrigin) {
				const rgn_origin_t & origin = rgn.origins [cOrigin];
				if (origin.bDeleted)
					continue;

				pbRgnsInLevel [origin.cLevel] = true;

				// Include origins in the bounding rect.
				rgn.rectBound.Extend (origin.point);

				if (origin.cLevel >= Levels.size () - 1)
					ReportWarning (origin.cFileOffset, "Region' origin has incorrect level %d.", origin.cLevel);
			}

			rectBound.Extend (rgn.rectBound);

			if (cElements == 0 && rgn.origins.empty ()) {
				ReportWarning (rgn.cBeginPos, "Region does not contain any point. The region will be ignored.");

				rgn.dwFlags |= rgn_t::fDeleted;
				++ layer.cDeletedRgns;
			}

			if (! rgn.strLabel.empty () && ! IsLabelCorrect (rgn.strLabel.c_str ()))
				ReportWarning (rgn.cBeginPos, "Label contains symbols not allowed for 6-bit coding.");

			// Prepare point for the label.
			if (! rgn.strLabel.empty ())
				rgn.pTypeDesc->pfPrepareLabel (rgn);

			SetProgressStatus (fast_floor (50 + 12.5*cRgn/cRgns + 12.5*(3 - cLayer) + .5));
		}
	}

	// Sort RGN80 by fArea.
	layers [3].rgns.sort ();//CompareRgnArea ());

	// Proceed the attached images.
	for (images_t::iterator iImage = images.begin (); iImage != images.end (); ++ iImage)
		rectBound.Extend (iImage->rect);

	// Check zoom levels.
	if (! Zoom.empty ()) {
		for (cLevel = 0; cLevel < Zoom.size () - 1; ++ cLevel) {
			if (! pbRgnsInLevel [cLevel])
				ReportWarning ("Level %d is empty.", cLevel);

			if (Zoom [cLevel] >= Zoom [cLevel + 1])
				ReportWarning ("Zoom%d >= Zoom%d.", cLevel, cLevel + 1);
		}
		if (pbRgnsInLevel [cLevel])
			ReportWarning ("Zoom level %d is not empty.", cLevel);
	} else
		if (! bNoHeader)
			ReportWarning ("Zoom levels are not defined.");

	if (cMaxPointsInElement > 255)
		ReportWarning ("There are region elements with more than 255 points (cGPSMap limitation).");
	if (bThereAreIslands)
		ReportWarning ("There are polygons with nested elements (cGPSMap limitation).");

	// TODO: check datum.
	// TODO: check generalization.
	// TODO: check other cGPSMap limitations
// - Size of a single element cannot exceed maximum allowed size of region
//;   (maximum size of region can be calculated from TreSize*DegreeGrid for
//;   given level)
//
//LevelLimit = (optional) limit of element for a single region (value less or equal to 255)
//LevelFill

	if (cPrecOrder > 5)
		ReportWarning ("Input file defines some or all points with excessive precision. Significant digits will be truncated to 5.");

	fCosY = ::cos (rectBound.MeanY ()*c_PI/180.);

	SetProgressStatus (100);

	return true;
}

find_info_t map_t::FindRgn (const point_t & _p) {
	const float c_fE = c_fE_in_pixels/g_kX;

	rect_t rectE;
	rectE.x0 = _p.x - c_fE;
	rectE.x1 = _p.x + c_fE;
	rectE.y0 = _p.y - c_fE;
	rectE.y1 = _p.y + c_fE;

	find_info_t ret;

	for (images_t::iterator iImage = images.begin (); iImage != images.end (); ++ iImage) {
		image_t & image = * iImage;

		if (! image.bOverlay && ! image.bDeleted && image.rect.Contains (_p))
			ret.pImage = & image;
	}

	for (int cLayer = 3; cLayer >= 0; -- cLayer) {
		layer_t & layer = layers [cLayer];
		rgns_t & rgns = layer.rgns;
		for (rgns_t::iterator iRgn = rgns.begin (); iRgn != rgns.end (); ++ iRgn) {
			if (g_bStop)
				return ret;

			rgn_t & rgn = * iRgn;
			if (rgn.IsDeleted ())
				continue;

			if (! rgn.rectBound.Intersects (rectE))
				continue;

			for (size_t cOrigin = 0; cOrigin < rgn.origins.size (); ++ cOrigin) {
				const rgn_origin_t & origin = rgn.origins [cOrigin];
				if (origin.bDeleted)
					continue;

				if (! (rgn.dwFlags & rgn_t::fWaypointOrTrack) && ! LevelIsVisible (origin.cLevel))
					continue;

				if (PointContainsPt (origin.point, _p, c_fE)) {
					ret.pLayer = & layer;
					ret.iRgn = iRgn;
					ret.cElement = -1;
					ret.cOrigin = cOrigin;
					continue;
				}
			}
			
			for (size_t cElement = 0; cElement < rgn.elements.size (); ++ cElement) {
				const rgn_element_t & element = rgn.elements [cElement]; 
				if (element.bDeleted)
					continue;

				const points_t & points = element.points;
				if (points.empty ())
					continue;

				if (! element.rectBound.Contains (_p))
					continue;
				if (! (rgn.dwFlags & rgn_t::fWaypointOrTrack) && ! LevelIsVisible (element.cLevel))
					continue;

				if (rgn.IsPolyline ()) {
					if (PolylineContainsPt (points, _p, c_fE)) {
						ret.pLayer = & layer;
						ret.iRgn = iRgn;
						ret.cElement = cElement;
						ret.cOrigin = -1;
						continue;
					}
				} else if (rgn.IsPolygon ()) {
					if (PolygonContainsPt (points, _p) >= 0) {
						ret.pLayer = & layer;
						ret.iRgn = iRgn;
						ret.cElement = cElement;
						ret.cOrigin = -1;
						continue;
					}
				}
			}
		}
	}

	if (ret.pLayer)
		ret.pImage = NULL;

	return ret;
}

void map_t::ToWGS84 (datum_t _Datum, float _fHeight) {
	for (int cLayer = 3; cLayer >= 0; -- cLayer) {
		layer_t & layer = layers [cLayer];
		rgns_t & rgns = layer.rgns;
		for (rgns_t::iterator iRgn = rgns.begin (); iRgn != rgns.end (); ++ iRgn) {
			if (g_bStop)
				return;

			rgn_t & rgn = * iRgn;
			if (rgn.IsDeleted ())
				continue;

			// Update origins.
			for (size_t cOrigin = 0; cOrigin < rgn.origins.size (); ++ cOrigin) {
				rgn_origin_t & origin = rgn.origins [cOrigin];
				if (origin.bDeleted)
					continue;

				::ToWGS84 (_Datum, origin.point, _fHeight);
			}
			
			// Update elements of the rgn.
			for (size_t cElement = 0; cElement < rgn.elements.size (); ++ cElement) {
				rgn_element_t & element = rgn.elements [cElement]; 
				if (element.bDeleted)
					continue;

				points_t & points = element.points;
				const size_t cPoints = points.size ();

				for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint)
					::ToWGS84 (_Datum, points [cPoint], _fHeight);
			}

			// Update labels' positions.
			if (rgn.dwFlags & rgn_t::fPtLabelDefined)
				::ToWGS84 (_Datum, rgn.ptLabel, _fHeight);
		}
	}
}

void map_t::AffTransform (const aff_transform_t & _m) {
	// Update total bounding rectangle.
	rectBound = rect_t ();

	// NOTE: Take the attached images into account but do not transform as reference objects.
	for (images_t::iterator iImage = images.begin (); iImage != images.end (); ++ iImage)
		rectBound.Extend (iImage->rect);

	for (int cLayer = 3; cLayer >= 0; -- cLayer) {
		layer_t & layer = layers [cLayer];
		rgns_t & rgns = layer.rgns;
		for (rgns_t::iterator iRgn = rgns.begin (); iRgn != rgns.end (); ++ iRgn) {
			if (g_bStop)
				return;

			rgn_t & rgn = * iRgn;
			if (rgn.IsDeleted ())
				continue;

			if (rgn.dwFlags & rgn_t::fWaypointOrTrack) {
				// NOTE: Tracks and waypoints are not to be transformed as reference objects.
				rectBound.Extend (rgn.rectBound);
				continue;
			}

			// Update bounding rectangle of the rgn.
			rgn.rectBound = rect_t ();

			// Update origins.
			for (size_t cOrigin = 0; cOrigin < rgn.origins.size (); ++ cOrigin) {
				rgn_origin_t & origin = rgn.origins [cOrigin];
				if (origin.bDeleted)
					continue;

				_m.ApplyTo (origin.point);

				rgn.rectBound.Extend (origin.point);
			}
			
			// Update elements of the rgn.
			for (size_t cElement = 0; cElement < rgn.elements.size (); ++ cElement) {
				rgn_element_t & element = rgn.elements [cElement];
				if (element.bDeleted)
					continue;

				points_t & points = element.points;
				const size_t cPoints = points.size ();

				element.rectBound = rect_t ();
				for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
					point_t & point = points [cPoint];

					_m.ApplyTo (point);

					element.rectBound.Extend (point);
				}
				rgn.rectBound.Extend (element.rectBound);
			}

			// Update labels' positions.
			if (rgn.dwFlags & rgn_t::fPtLabelDefined)
				_m.ApplyTo (rgn.ptLabel);
			// TODO: fLabelAngle.

			rectBound.Extend (rgn.rectBound);
		}
	}
}