# ifndef __Map_H_
# define __Map_H_

//
// Map.h
//
// Declarations of all map-related types.
//

# include <vector>
# include <string>

# include "PtAlgo.h"
# include "Datum.h"


//
// Region element: 'elementary' polygon w/o self-intersections or polyline
// defined for givel zoom layer
// (corresponds to one DataX string in Polish format).
//
struct rgn_element_t {
	// Nodes of polygon or polyline.
	// NOTE: no self-intersections for polygons is assumed.
	points_t points;
	// The bounding rectangle.
	rect_t   rectBound;
	// The zoom level.
	size_t   cLevel;

	// The reference to source file.
	size_t   cFileLine;
	size_t   cFileOffset;

	// The linear size in degrees.
	float    fLength;
	// For polygons only - area in degrees^2.
	float    fArea;

	// For polygons only - whether is it nested 'hole' in another element.
	bool     bIsland;

	// True for deleted elements.
	bool     bDeleted;

	rgn_element_t () : bDeleted (false) {}
};

typedef std::vector<rgn_element_t> rgn_elements_t;


//
// Region's origin: coordinates of POI or point
// defined for givel zoom layer
// (corresponds to one OriginX string in Polish format).
//
struct rgn_origin_t {
	// The coordinates.
	point_t point;
	// The zoom level.
	size_t  cLevel;

	// The reference to source file.
	size_t  cFileLine;
	size_t  cFileOffset;

	// True for deleted origins.
	bool    bDeleted;

	rgn_origin_t () : bDeleted (false) {}
};

typedef std::vector<rgn_origin_t> rgn_origins_t;


struct map_t;
struct type_desc_t;
struct file_ref_t;

//
// Region: one or more elements or origins.
// (corresponds to one [RGNx0] section in Polish format).
//
struct rgn_t {
	// Rgn's type.
	const type_desc_t * pTypeDesc;
	// NOTE: Unknown types can't be found in CTypeDesc table; so, they should be saved here.
	USHORT ushType;
	USHORT ushKind;
	
	// Rgn's elementary polygons or polylines.
	rgn_elements_t elements;

	// Rgn's points.
	rgn_origins_t origins;

	// Rgn's label.
	std::string strLabel;
	// Position & orientation of the text (for visualization).
	point_t ptLabel;
	float   fLabelAngle;
	// Text width in pixels.
	size_t  cLabelWidth;

	// Distance^2 between max distant points.
	float fSize2;

	// Used to sort RGN80.
	float fAveElementArea;

	// The bounding rectangle.
	rect_t rectBound;

	// Reference to the source file.
	size_t cBeginPos;
	size_t cEndPos;
	size_t cFileLine;
	size_t cOrdinal;

	// The next rgn in load order.
	rgn_t * pNextRgn;

	std::string strComment;

	// Reference to the attacned file (for waypoints and tracks).
	file_ref_t * pFileRef;

	enum {
		fWaypointOrTrack     = 0x10,

		fDirIndicator        = 0x01,

		fDirIndicatorDefined = 0x02,
		fPtLabelDefined      = 0x04,

		// NOTE: it should correspond to layer_t::cDeletedRgns.
		fDeleted             = 0x08
	};
	DWORD dwFlags;

	bool IsDeleted () const;

	bool IsPolygon () const;
	bool IsPolyline () const;
	bool IsPoint () const;
	bool IsPOI () const;

	bool Contains (size_t _cLevel, const point_t & _point) const;

	// Updates bounding rect, area, size etc
	void RefreshAttributes (map_t & _map);

	rgn_t () :
		pTypeDesc (NULL),
		ushType (0),
		ushKind (0),
		cLabelWidth (0),
		fSize2 (0),
		fAveElementArea (0),
		cBeginPos (0),
		cEndPos (0),
		cFileLine (0),
		cOrdinal (0),
		pNextRgn (NULL),
		dwFlags (0),
		pFileRef (NULL)
	{}

	// NOTE: used to sort RGN80.
	bool operator < (const rgn_t & _other) const {
		return fAveElementArea > _other.fAveElementArea;
	}
/*
	bool operator == (const rgn_t & _other) const {
		if (fAvePartArea != _other.fAvePartArea)
			return false;
		return elements == _other.elements;
	}*/
};

typedef std::list<rgn_t> rgns_t;

float GetRgnSize (const rgn_t & _rgn);


//
// Rendering layer.
// NOTE: all objects in the layer are considered to be drawn from first to last one.
// In practice, there are just 4 layers: (1) POIs, (2) points, (3) polylines, (4) polygons.
// And just polygons may overlap, so the order of drawing inside their layer is meaningful for them.
//
struct layer_t {
	rgns_t rgns;
	// Total count of regions, including deleted ones.
	size_t cRgns;
	size_t cDeletedRgns;

	layer_t () : cRgns (0), cDeletedRgns (0) {}

	void clear () {
		rgns.clear ();
		cRgns = 0;
		cDeletedRgns = 0;
	}

	void swap (layer_t & _other) {
		rgns.swap (_other.rgns);
		std::swap (cRgns, _other.cRgns);
		std::swap (cDeletedRgns, _other.cDeletedRgns);
	}
};


//
// The source file description.
//
struct file_ref_t {
	// File name including full path.
	std::string strFileName;
	// True for deleted files.
	bool bDeleted;
	// True for attached files.
	bool bAttachment;

	file_ref_t () : bDeleted (false), bAttachment (true) {}
};

typedef std::list<file_ref_t> file_refs_t;


//
// The attached raster map.
//
struct image_t {
	// The bounding rectangle.
	rect_t  rect;

	// The bitmap with the map image.
	HBITMAP hBitmap;
	HDC     hMemDC;
	long    lWidth;
	long    lHeight;
	// The stretched bitmap used to speed up output.
	HBITMAP hSmallBitmap;
	HDC     hSmallMemDC;
	long    lSmallWidth;
	long    lSmallHeight;

	// File name including full path.
	std::string strFileName;

	// True for additional alpha-blended images rendered for overlay areas.
	bool bOverlay;

	// True for deleted image.
	bool bDeleted;

	image_t () :
		hBitmap (NULL), hMemDC (NULL),
		hSmallBitmap (NULL), hSmallMemDC (NULL),
		bOverlay (false),
		bDeleted (false)
	{}

	void MakeSmall ();

	void clear ();
	void swap (image_t & _other);
};

typedef std::list<image_t> images_t;
void JoinImages (images_t & _images1, images_t & _images2);


//
//
//
struct find_info_t {
	// NULL if not found.
	layer_t * pLayer;
	rgns_t::iterator iRgn;
	// -1 for all rgn' elements
	size_t cElement;
	// -1 for all rgn' origins
	size_t cOrigin;

	// NULL if not found.
	image_t * pImage;

	find_info_t () : pLayer (NULL), pImage (NULL) {}

	void clear () {
		pLayer = NULL;
		pImage = NULL;
	}

	bool empty () const {return pImage == NULL && pLayer == NULL;}

	bool operator == (const find_info_t & _other) const {
		if (pImage != NULL || _other.pImage != NULL)
			return  pImage == _other.pImage;
		if (pLayer != _other.pLayer)
			return false;
		if (pLayer == NULL)
			return true;
		return iRgn == _other.iRgn && cElement == _other.cElement && cOrigin == _other.cOrigin;
	}
};


class CMapLoader;

//
// The map data.
//
struct map_t {
	//
	// Header data.
	//

	// True for imported maps w/o [IMG ID] headers.
	bool bNoHeader;
	// The length of the header in source file in bytes.
	size_t cHeaderLen;

	//
	size_t cRgnLimit;

	// The zoom levels.
	std::vector<size_t> Levels;
	std::vector<size_t> Zoom;

	// The rest of definitions.
	std::vector<std::string> header_data;


	//
	// The map layers sorted in order of drawing:
	//   layers[0] is for POI (RGN10),
	//   layers[1] is for points (RGN20),
	//   layers[2] is for polylines (RGN40),
	//   layers[3] is for polygons (RGN80), in turn, sorted by fAveElementArea.
	//
	layer_t layers [4];


	// The raster images attached.
	images_t images;

	// Source files (including attached files with waypoints and tracks but excluding raster maps).
	file_refs_t file_refs;


	// The bounding rectangle for the whole map.
	rect_t rectBound;
	// Max number of points per rgn's element.
	size_t cMaxPointsInElement;
	// Cosine of the average map's latitude.
	float fCosY;


	// The head & tail of rgns list in load order.
	rgn_t * pFirstRgn, * pLastRgn;
	void InsertRgnToSaveList (rgn_t & _rgn) {
		if (pFirstRgn == NULL)
			pFirstRgn = & _rgn;
		if (pLastRgn != NULL)
			pLastRgn->pNextRgn = & _rgn;
		pLastRgn = & _rgn;
	}


	// The number of significant digits (for load/save).
	size_t cPrecOrder;


	// The constructor.
	map_t () {Init ();}
	void Init ();

	bool empty () const {
		return
			layers [0].rgns.empty () &&
			layers [1].rgns.empty () &&
			layers [2].rgns.empty () &&
			layers [3].rgns.empty () &&
			images.empty ();
	}
	size_t GetRgnCount () const {
		return layers [0].cRgns + layers [1].cRgns + layers [2].cRgns + layers [3].cRgns;
	}

	bool LoadMap (const char * _pData, size_t _cFileSize, const char * _strFileName, CMapLoader * _pLoader);

	void Swap (map_t & _other);
	void Join (map_t & _other);

	find_info_t FindRgn (const point_t & _pt);

	void Draw (HDC _hDC) const;

	void ToWGS84 (datum_t _Datum, float _fHeight);
	void AffTransform (const aff_transform_t & _m);
};


//
// The map in hand.
//
extern map_t g_map;

// The critical section to protect access to g_map.
# include "Wrappers.h"
extern CCritSec g_cs;

# endif // __Map_H_