# include "StdAfx.h"
# include "Resource.h"
# include "MovePoints.h"
# include "MapView.h"
# include "Selection.h"
# include "Position.h"
# include "Scale.h"
# include "Levels.h"
# include "MapView.h"
# include "UndoRedo.h"
# include "Label.h"
# include "Globals.h"

class CMovePointAction : public CEditAction {
  public:
	CMovePointAction (rgn_t * _pRgn, rgn_element_t * _pElement, size_t _cPointIndex, const point_t & _pointNew) :
		m_pRgn (_pRgn),
		m_pElement (_pElement),
		m_cPointIndex (_cPointIndex),
		m_pointOld (_pElement->points [_cPointIndex]),
		m_pointNew (_pointNew)
	{}

	virtual void GetName (std::string & _str) const;
	virtual bool CanUndo () const;
	virtual bool Do ();
	virtual void Undo ();
	virtual void Redo ();

  private:
	rgn_t * const m_pRgn;
	rgn_element_t * const m_pElement;
	const size_t m_cPointIndex;
	const point_t m_pointNew, m_pointOld;
};

void CMovePointAction::GetName (std::string & _str) const {
	_str.append ("move point [");
	GetHintText (* m_pRgn, _str);
	_str.append ("]");
}

bool CMovePointAction::CanUndo () const {return true;}

bool CMovePointAction::Do () {
	CCritSec::CLock lock (g_cs);

	m_pElement->points [m_cPointIndex] = m_pointNew;
	m_pRgn->RefreshAttributes (g_map);

	return true;
}

void CMovePointAction::Undo () {
	CCritSec::CLock lock (g_cs);

	m_pElement->points [m_cPointIndex] = m_pointOld;
	m_pRgn->RefreshAttributes (g_map);
}

void CMovePointAction::Redo () {
	CMovePointAction::Do ();
}

static inline
void MovePoint (rgn_t * _pRgn, rgn_element_t * _pElement, size_t _cPointIndex, const point_t & _new) {
	assert (_cPointIndex < _pElement->points.size ());
	PerformEditAction (new CMovePointAction (_pRgn, _pElement, _cPointIndex, _new));
}

/////////////////////////////////

class CDeletePointAction : public CEditAction {
  public:
	CDeletePointAction (rgn_t * _pRgn, rgn_element_t * _pElement, size_t _cPointIndex) :
		m_pRgn (_pRgn),
		m_pElement (_pElement),
		m_cPointIndex (_cPointIndex),
		m_point (_pElement->points [_cPointIndex])
	{}

	virtual void GetName (std::string & _str) const;
	virtual bool CanUndo () const;
	virtual bool Do ();
	virtual void Undo ();
	virtual void Redo ();

  private:
	rgn_t * const m_pRgn;
	rgn_element_t * const m_pElement;
	const size_t m_cPointIndex;
	const point_t m_point;
};

void CDeletePointAction::GetName (std::string & _str) const {
	_str.append ("delete point [");
	GetHintText (* m_pRgn, _str);
	_str.append ("]");
}

bool CDeletePointAction::CanUndo () const {return true;}

bool CDeletePointAction::Do () {
	CCritSec::CLock lock (g_cs);

	m_pElement->points.erase (m_pElement->points.begin () + m_cPointIndex);
	m_pRgn->RefreshAttributes (g_map);

	return true;
}

void CDeletePointAction::Undo () {
	CCritSec::CLock lock (g_cs);

	m_pElement->points.insert (m_pElement->points.begin () + m_cPointIndex, m_point);
	m_pRgn->RefreshAttributes (g_map);
}

void CDeletePointAction::Redo () {
	CDeletePointAction::Do ();
}

static inline
void DeletePoint (rgn_t * _pRgn, rgn_element_t * _pElement, size_t _cPointIndex) {
	assert (_cPointIndex < _pElement->points.size ());
	PerformEditAction (new CDeletePointAction (_pRgn, _pElement, _cPointIndex));
}

/////////////////////////////////

class CAddPointAction : public CEditAction {
  public:
	CAddPointAction (rgn_t * _pRgn, rgn_element_t * _pElement, size_t _cPointIndex) :
		m_pRgn (_pRgn),
		m_pElement (_pElement),
		m_cPointIndex (_cPointIndex)
	{}

	virtual void GetName (std::string & _str) const;
	virtual bool CanUndo () const;
	virtual bool Do ();
	virtual void Undo ();
	virtual void Redo ();

  private:
	rgn_t * const m_pRgn;
	rgn_element_t * const m_pElement;
	const size_t m_cPointIndex;
};

void CAddPointAction::GetName (std::string & _str) const {
	_str.append ("add point [");
	GetHintText (* m_pRgn, _str);
	_str.append ("]");
}

bool CAddPointAction::CanUndo () const {return true;}

bool CAddPointAction::Do () {
	CCritSec::CLock lock (g_cs);

	assert (m_cPointIndex + 1 < m_pElement->points.size ());

	// Make new point.
	const point_t & point1 = m_pElement->points [m_cPointIndex];
	const point_t & point2 = m_pElement->points [m_cPointIndex + 1];
	point_t point;
	point.x = (point1.x + point2.x)/2;
	point.y = (point1.y + point2.y)/2;

	m_pElement->points.insert (m_pElement->points.begin () + m_cPointIndex + 1, point);

	m_pRgn->RefreshAttributes (g_map);

	return true;
}

void CAddPointAction::Undo () {
	CCritSec::CLock lock (g_cs);

	m_pElement->points.erase (m_pElement->points.begin () + m_cPointIndex + 1);
	m_pRgn->RefreshAttributes (g_map);
}

void CAddPointAction::Redo () {
	CAddPointAction::Do ();
}

static inline
void AddPoint (rgn_t * _pRgn, rgn_element_t * _pElement, size_t _cPointIndex) {
	assert (_cPointIndex < _pElement->points.size () - 1);
	PerformEditAction (new CAddPointAction (_pRgn, _pElement, _cPointIndex));
}

/////////////////////////////////

static HCURSOR g_hcurMovePointDrag;
static HCURSOR g_hcurMovePoint;

static HPEN g_hpenPoint;
static HPEN g_hpenTmp;

HMENU g_hmenuPointPopup;

bool CPointsMode::Init () {
	g_hpenPoint = ::CreatePen (PS_SOLID, 1, c_colBlue);
	g_hpenTmp   = ::CreatePen (PS_SOLID, 1, c_colGray);

	g_hcurMovePoint     = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorMovePoint));
	g_hcurMovePointDrag = ::LoadCursor (g_hInst, MAKEINTRESOURCE (IDC_CursorMovePointDrag));

	g_hmenuPointPopup = ::GetSubMenu (::LoadMenu (g_hInst, MAKEINTRESOURCE (IDM_POPUP)), 3);
	assert (g_hmenuPointPopup);
	::SetMenuDefaultItem (g_hmenuPointPopup, ID_RGN_PROPERTIES, MF_BYCOMMAND);

	return true;
}

void CPointsMode::OnEnterMode () {
	SetPermanentCursor (g_hArrowCursor);
}

static rgn_elements_t::iterator g_iElement;
static size_t g_cPointIndex;

static
bool FindPoint (int _x, int _y) {
	if (g_Selection.size () != 1 || ! g_Selection [0].pLayer)
		return false;

	const size_t cLevel = GetZoomLevel (g_fScale);
	if (cLevel == -1)
		return false;

	rgn_t & rgn = * g_Selection [0].iRgn;
	for (rgn_elements_t::iterator i = rgn.elements.begin (); i != rgn.elements.end (); ++ i) {
		rgn_element_t & element = * i;
		if (element.bDeleted)
			continue;
		if (element.cLevel != cLevel)
			continue;

		const points_t & points = element.points;
		const size_t cPoints = points.size ();
		for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
			const point_t & point = points [cPoint];

			const long lX = XToScreen (point.x);
			const long lY = YToScreen (point.y);
			if (::labs (_x - lX) <= 3 && ::labs (_y - lY) <= 3) {
				g_iElement = i;
				g_cPointIndex = cPoint;
				return true;
			}
		}
	}

	return false;
}

void CPointsMode::OnLDown (int _x, int _y, DWORD _dwKeys) {
	const bool bStillContains = FindPoint (_x, _y);

	if (! bStillContains)
		SelectObjectAt (_x, _y, false, false);

	if (bStillContains && FindPoint (_x, _y))
		StartDND (_x, _y);
}

void CPointsMode::OnLUp (int _x, int _y, DWORD _dwKeys) {
	SetPermanentCursor (g_hArrowCursor);

	if (! CMouseMode::s_bDND)
		return;
	StopDND ();

	// Apply changes.
	if (g_Selection.size () == 1 && g_Selection [0].pLayer) {
		point_t point;
		ScreenToPt (_x, _y, point);
		MovePoint (& * g_Selection [0].iRgn, & * g_iElement, g_cPointIndex, point);
	}
}

void CPointsMode::OnMove  (int _x, int _y, DWORD _dwKeys) {
	if (! CMouseMode::s_bDND) {
		if (FindPoint (_x, _y))
			SetPermanentCursor (g_hcurMovePoint);
		else
			SetPermanentCursor (g_hArrowCursor);
		return;
	}

	SetPermanentCursor (g_hcurMovePointDrag);
	// TODO: optimize rect
	Redraw ();
}

void CPointsMode::OnPaint (HDC _hDC) {
	if (g_Selection.size () != 1 || ! g_Selection [0].pLayer)
		return;

	const size_t cLevel = GetZoomLevel (g_fScale);
	if (cLevel == -1)
		return;

	rgn_t & rgn = * g_Selection [0].iRgn;
	for (rgn_elements_t::iterator i = rgn.elements.begin (); i != rgn.elements.end (); ++ i) {
		rgn_element_t & element = * i;
		if (element.bDeleted)
			continue;
		if (element.cLevel != cLevel)
			continue;

		const points_t & points = element.points;
		const size_t cPoints = points.size ();
		for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
			if (cPoint == 0)
				// Highlight the first point by different color.
				::SelectObject (_hDC, g_hGrayThinPen);
			else if (cPoint == 1)
				::SelectObject (_hDC, g_hpenPoint);

			const point_t & point = points [cPoint];

			const long lX = XToScreen (point.x);
			const long lY = YToScreen (point.y);

			// Draw cross.
			::MoveToEx (_hDC, lX - 5, lY - 5, NULL);
			::LineTo   (_hDC, lX + 6, lY + 6);

			::MoveToEx (_hDC, lX - 5, lY + 5, NULL);
			::LineTo   (_hDC, lX + 6, lY - 6);
		}

	}

	if (! CMouseMode::s_bDND)
		return;
		
	// Draw 'rubber'.
	if (g_iElement->points.size () >= 2) {
		assert (g_cPointIndex < g_iElement->points.size ());

		::SelectObject (_hDC, g_hpenTmp);

		size_t cPrevPointIndex = g_cPointIndex - 1;
		if (g_cPointIndex == 0 && g_Selection [0].iRgn->IsPolygon ())
			cPrevPointIndex = g_iElement->points.size () - 1;
		if (cPrevPointIndex < g_iElement->points.size ()) {
			const point_t & point_p = g_iElement->points [cPrevPointIndex];
			const long lX_p = XToScreen (point_p.x);
			const long lY_p = YToScreen (point_p.y);

			::MoveToEx (_hDC, CMouseMode::s_lX_DND, CMouseMode::s_lY_DND, NULL);
			::LineTo   (_hDC, lX_p, lY_p);
		}

		size_t cNextPointIndex = g_cPointIndex + 1;
		if (g_cPointIndex >= g_iElement->points.size () - 1 && g_Selection [0].iRgn->IsPolygon ())
			cNextPointIndex = 0;
		if (cNextPointIndex < g_iElement->points.size ()) {
			const point_t & point_n = g_iElement->points [cNextPointIndex];
			const long lX_n = XToScreen (point_n.x);
			const long lY_n = YToScreen (point_n.y);

			::MoveToEx (_hDC, CMouseMode::s_lX_DND, CMouseMode::s_lY_DND, NULL);
			::LineTo   (_hDC, lX_n, lY_n);
		}
	}
}

void CPointsMode::OnRUp (int _x, int _y, DWORD _dwKeys) {
	if (! FindPoint (_x, _y))
		return;

	assert (g_Selection.size () == 1 && g_Selection [0].pLayer);
	const bool bCanDelete = g_Selection [0].iRgn->IsPolygon () ? g_iElement->points.size () >= 4 : g_iElement->points.size () >= 3;
	::EnableMenuItem (g_hmenuPointPopup, ID_EDIT_DELETE_POINT, (bCanDelete ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND);
	const bool bCanAdd = g_Selection [0].iRgn->IsPolygon () || g_cPointIndex < g_iElement->points.size () - 1;
	::EnableMenuItem (g_hmenuPointPopup, ID_EDIT_ADD_POINT,    (bCanAdd ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND);

	g_lRClickX = _x;
	g_lRClickY = _y;
	if (! g_Selection.empty ()) {
		POINT pt;
		::GetCursorPos (& pt);
		::TrackPopupMenu (
			g_hmenuPointPopup,
			TPM_LEFTALIGN | TPM_TOPALIGN | TPM_LEFTBUTTON,
			pt.x, pt.y,
			0, g_hWnd, NULL
		);
	}
}

bool CPointsMode::OnCommand (WORD _wID) {
	switch (_wID) {
		case ID_EDIT_DELETE_POINT:
			if (! FindPoint (g_lRClickX, g_lRClickY))
				return false;

			assert (g_Selection.size () == 1 && g_Selection [0].pLayer);
			DeletePoint (& * g_Selection [0].iRgn, & * g_iElement, g_cPointIndex);
			return true;

		case ID_EDIT_ADD_POINT:
			if (g_Selection.size () != 1 || ! g_Selection [0].pLayer)
				return false;
			AddPoint (& * g_Selection [0].iRgn, & * g_iElement, g_cPointIndex);
			return true;
	}

	return false;
}

void CPointsMode::OnLDblClk (int _x, int _y, DWORD _dwKeys) {
//	SelectRgn (GET_X_LPARAM (_lParam), GET_Y_LPARAM (_lParam));
//	if (! g_Selection.empty ())
//		ShowRgnProperties (g_Selection);
}