# include "stdafx.h"
# include "Map.h"
# include "Types.h"
# include "Load.h"
# include "Log.h"
# include "Status.h"
# include "Globals.h"

//////////////////////////////////////////

struct CWaypointHdr;
struct CTrackHdr;

class CMpsLoader : public CMapLoader {
  public:
	// Override.
	bool IsValid (const char * _pData, size_t _cFileSize) const;
	bool Load (const char * _pData, size_t _cFileSize, const char * _strFileName, map_t * _pMap);

  private:
	void ParseWaypoint (const CWaypointHdr * _pWP, const void * _pEnd);
	void ParseTrack (const CTrackHdr * _pTrack, const void * _pEnd);

	map_t * m_pMap;
	const char * m_pStart, * m_pEnd;

	//
	size_t m_cOrdinal;

	file_refs_t m_file_refs;
};

DECLARE_MAP_LOADER (CMpsLoader, "mps", "MapSource 4.x file (*.mps)\0*.mps", false);

bool CMpsLoader::IsValid (const char * _pData, size_t _cFileSize) const {
	if (_cFileSize < 0x2d)
		return false;
	return ::memcmp (_pData, "MsRcd", 5) == 0;
}

# ifdef _MSC_VER
#	pragma pack (push, 1)
# endif

struct CRecordHdr {
	DWORD dwLen;
	BYTE  btType;
};

struct CVersion : public CRecordHdr {
	// 0x2E - 3.02
	// 0x99 - 4.09
	// 0x9A - 4.10 beta 3
	BYTE btVersion;
	BYTE bt0;  // Always 0x01

	// Followed by three zero-terminated strings with MapSource builder's nikname
	// and date and time.
	// Example: "hales\0Nov 18 2002\014:11:40\0"
};

struct CWaypointHdr : public CRecordHdr {
	char strName [1];
};

struct CWaypoint2 {
	BYTE ba0 [9];   // 0x00
	BYTE ba1 [0xC]; // 0xFF

	WORD w0;        // 0x00

	DWORD dw0;      // 0xFF

	DWORD dwY; // In arcs.
	DWORD dwX;

	BYTE  btAltitude;  // 0 - no altitude, 1 - has altitude
	double dblAltitude; // In meters

	char strDescription [1];
};

struct CWaypoint3 {
	BYTE  bt0;      // zero

	BYTE   btProximity;  // 0 - no proximity, 1 - has proximity
	double dblProximity; // In meters.

	// 0 - symbol, 0x1 - symbol & name, 0x2 - symbol and description
	DWORD dwShow;  

	// 0 - unknown, 1 - black, 2 - dk red, 3 - dk green, 4 - dk yellow,
	// 5 - dk blue, 6 - dk magenta, 7 - dk cyan, 8 - lt gray, 9 - dk gray, 0x0A - red
	// 0x0B - green, 0x0C - yellow, 0x0D - blue, 0x0E - magenta, 0x0F - cyan, 0x10 -  white
	DWORD dwColor; 

	// 0 - marina, 0x0A - residence, 0x0e - Skull & Crossbones, 0x12 - waypoint,
    // 0x15 - navaid (amber),
	// 0x25 - boat ramp, 0x26 - campground, 0x2d - parking area, 0x2e - park,
	// 0x32 - swimming area, 0x3e - Truck stop, 0x38 - Car
	// 0x4c - Car repair, 0x50 - museum, 0x5e - building, 0x5d - bridge
	// 0x69 - Forest, 0x6a - summit, 0x6b - airport, 0x6c - heliport
	DWORD dwType;  

	DWORD dw0; // zero

	BYTE  btDepth; // 0 - no depth, 1 - has depth
	double dblDepth; // In 10x meters

	BYTE  ba0 [7]; // zero
};

# ifdef _MSC_VER
#	pragma pack (pop)
# endif

struct t2t_t {
	DWORD dwWaypointType;
	DWORD dwRGNType;
};

const t2t_t g_t2t [] = {
	{0x00, 0x02f09}, // marina
	{0x0A, 0x06402}, // residence -> building
	{0x0e, 0x05700}, // skull & crossbones -> danger area
	{0x12, 0},       // waypoint -> unknown
	{0x15, 0x01612}, // navaid (amber)
	{0x25, 0x04700}, // boat ramp
	{0x26, 0x10048}, // campground
	{0x2d, 0x02f0b}, // parking area -> parking
	{0x2e, 0x0640e}, // park
	{0x32, 0x02d09}, // swimming area -> swimming
	{0x3e, 0x04d00}, // truck stop -> parking
	{0x38, 0x04d00}, // car -> parking
	{0x4c, 0x02f03}, // car repair -> auto repair
	{0x50, 0x02c02}, // museum -> museum/history
	{0x5e, 0x06402}, // building
	{0x5d, 0x06401}, // bridge
	{0x69, 0x0660a}, // forest
	{0x6a, 0x06616}, // summit
	{0x6b, 0x05900}, // airport
	{0x6c, 0x05904}, // heliport
};

const size_t g_cT2TSize = sizeof (g_t2t)/sizeof (g_t2t [0]);

static
const DWORD FindType (DWORD _dwType) {
	size_t a = 1, b = g_cT2TSize - 1;
	while (true) {
		const DWORD dwA = g_t2t [a].dwWaypointType;
		if (dwA == _dwType)
			return g_t2t [a].dwRGNType;

		const DWORD dwB = g_t2t [b].dwWaypointType;
		if (dwB == _dwType)
			return g_t2t [b].dwRGNType;

		if (b - a <= 1)
			return 0;

		const size_t m = (a + b)/2;

		const DWORD dwM = g_t2t [m].dwWaypointType;
		if (dwM < _dwType)
			a = m;
		else
			b = m;
	}
}


void CMpsLoader::ParseWaypoint (const CWaypointHdr * _pWP, const void * _pEnd) {
	const size_t cNameLen = ::strlen (_pWP->strName) + 1;
	const CWaypoint2 * const p2 = reinterpret_cast<const CWaypoint2 *> (reinterpret_cast<const BYTE *> (_pWP->strName) + cNameLen);

	const size_t cDescriptionLen = ::strlen (p2->strDescription);
	const CWaypoint3 * const p3 = reinterpret_cast<const CWaypoint3 *> (reinterpret_cast<const BYTE *> (p2->strDescription) + cDescriptionLen);

	point_t point;
	point.x = p2->dwX*180./(1UL << 31);
	point.y = p2->dwY*180./(1UL << 31);

	const DWORD dwType = FindType (p3->dwType);

	/////////////////////////////

	rgn_t rgn;
	rgn.cBeginPos = (char *) _pWP - m_pStart;
	rgn.cEndPos = (char *) _pEnd - m_pStart;
	rgn.cFileLine = 0;
	rgn.cOrdinal = m_cOrdinal ++;

	rgn.dwFlags |= rgn_t::fWaypointOrTrack;

	rgn.strLabel = _pWP->strName;

	rgn.origins.push_back (rgn_origin_t ());
	rgn_origin_t & origin = rgn.origins.back ();
	origin.cLevel = 0;
	origin.point = point;

	// Find best type for the waypoint.
	rgn.pTypeDesc = GetTypeDesc (dwType & 0xFFFF, dwType >> 16);
	rgn.ushType = dwType & 0xFFFF;
	rgn.ushKind = dwType >> 16;

	rgn.pFileRef = & m_file_refs.back ();

	//
	// Add the waypoint to the map.
	//

	m_pMap->layers [rgn.ushKind].rgns.push_back (rgn);
	++ m_pMap->layers [rgn.ushKind].cRgns;
	m_pMap->InsertRgnToSaveList (m_pMap->layers [rgn.ushKind].rgns.back ());
}

# ifdef _MSC_VER
#	pragma pack (push, 1)
# endif

struct CTrackHdr : public CRecordHdr {
	char strName [1];
};

struct CTrackPoint {
	DWORD dwY;        // In arcs.
	DWORD dwX;

	BYTE  btHasHeight; // 1 - has height, 0 - not defined
	double dfHeight;   // In meters.

	BYTE  btHasTime;  // 0 - no time info, 1 - has time info (NOTE: compressed tracks have no time)
	DWORD dwDateTime; // In seconds elapsed since 1 Jan 1970

	BYTE  ba [9];     // 0
};

struct CTrack2 {
	BYTE  btDisplayOnGPS; // 1 - display, 0 - hide
	DWORD dwColor;    // 0 - black etc (see above)

	DWORD dwPoints;
	CTrackPoint points [1];
};

# ifdef _MSC_VER
#	pragma pack (pop)
# endif

void CMpsLoader::ParseTrack (const CTrackHdr * _pTrack, const void * _pEnd) {
	const size_t cNameLen = ::strlen (_pTrack->strName) + 1;

	const CTrack2 * const p2 = reinterpret_cast<const CTrack2 *> (reinterpret_cast<const BYTE *> (_pTrack->strName) + cNameLen);

	points_t points (p2->dwPoints);

	const CTrackPoint * pPoint = p2->points;
	for (size_t cPoint = 0; cPoint < p2->dwPoints; ++ cPoint) {
		point_t & point = points [cPoint];
		point.x = pPoint->dwX*180./(1UL << 31);
		point.y = pPoint->dwY*180./(1UL << 31);

		++ pPoint;
		if (pPoint + 1 > _pEnd)
			break;
	}
	assert ((void *) pPoint == _pEnd);

	//
	// Add the track to the map.
	//

	m_pMap->layers [2].rgns.push_back (rgn_t ());
	++ m_pMap->layers [2].cRgns;

	rgn_t & rgn = m_pMap->layers [2].rgns.back ();

	m_pMap->InsertRgnToSaveList (rgn);

	//
	rgn.dwFlags |= rgn_t::fWaypointOrTrack;

	rgn.cBeginPos = (char *) _pTrack - m_pStart;
	rgn.cEndPos = (char *) _pEnd - m_pStart;
	rgn.cFileLine = 0;
	rgn.cOrdinal = m_cOrdinal ++;

	rgn.strLabel = _pTrack->strName;

	rgn.elements.push_back (rgn_element_t ());
	rgn_element_t & element = rgn.elements.back ();
	element.cFileLine = 0;
	element.cFileOffset = (char *) _pTrack - m_pStart;

	//
	element.cLevel = 0;
	element.points.swap (points);

	rgn.pTypeDesc = & g_types [0];
	rgn.ushKind = 2;

	rgn.pFileRef = & m_file_refs.back ();
}

bool CMpsLoader::Load (const char * _pData, size_t _cFileSize, const char * _strMapFileName, map_t * _pMap) {
	m_pMap = _pMap;
	m_pStart = _pData;
	m_pEnd = _pData + _cFileSize;

	m_cOrdinal = 0;

	m_file_refs.clear ();
	m_file_refs.push_back (file_ref_t ());
	m_file_refs.back ().strFileName = _strMapFileName;
	m_file_refs.back ().bAttachment = true;

	SetProgressStatus (0);

	const char * pRecord = _pData + 0xD;
	do {
		if (g_bStop)
			return false;

		const CRecordHdr * const pHdr = reinterpret_cast<const CRecordHdr *> (pRecord);
		const char * const pNextRecord = pRecord + sizeof (CRecordHdr) + pHdr->dwLen;

		switch (pHdr->btType) {
			case 'A': { // Format version.
				const CVersion * const pVer = static_cast<const CVersion *> (pHdr);
				if (pVer->btVersion < 0x99) {
					ReportWarning ("The file is created by an obsolete version of MapSource. MapSource 4.09 or later is required.");
					return false;
				} else if (pVer->btVersion > 0x9A)
					ReportWarning ("The file is created by a version of MapSource later than 4.10. Such format was not tested, and result of file loading may be unpredictable.");
				break;
			}

			case 'W': { // Waypoint
				const CWaypointHdr * const pWP = static_cast<const CWaypointHdr *> (pHdr);
				ParseWaypoint (pWP, pNextRecord);
				break;
			}

			case 'T': { // Track
				const CTrackHdr * const pTrack = static_cast<const CTrackHdr *> (pHdr);
				ParseTrack (pTrack, pNextRecord);
				break;
			}

			case 'R':  // Route
			case 'L':  // Map

			case 'V': {	// Unknown - appears at end as single WORD.
				break;
			}
		}

		pRecord = pNextRecord;

		SetProgressStatus (50.*(pRecord - _pData)/_cFileSize);
	} while (pRecord < m_pEnd);

	SetProgressStatus (50);

	m_pMap->file_refs.swap (m_file_refs);

	// Suppress warning about empty Zoom etc.
	m_pMap->bNoHeader = true;

	return true;
}
