# include "stdafx.h"
# include "Map.h"
# include "Datum.h"
# include "Load.h"
# include "Log.h"
# include "Status.h"
# include "MapView.h"
# include "MathAlgo.h"
# include "Globals.h"

//////////////////////////////////////////

# include "MiniGDI+.h"

HMODULE g_hGDIPlusDll;

typedef Status (WINAPI * pfGdiplusStartup_t) (OUT ULONG_PTR *token, const GdiplusStartupInput *input, OUT GdiplusStartupOutput *output);
static pfGdiplusStartup_t g_pfGdiplusStartup;

typedef VOID (WINAPI * pfGdiplusShutdown_t) (ULONG_PTR token);
static pfGdiplusShutdown_t g_pfGdiplusShutdown;

typedef GpStatus (WINGDIPAPI * pfGdipCreateFromHDC_t) (HDC hdc, GpGraphics **graphics);
static pfGdipCreateFromHDC_t g_pfGdipCreateFromHDC;

typedef GpStatus (WINGDIPAPI * pfGdipDeleteGraphics_t) (GpGraphics *graphics);
static pfGdipDeleteGraphics_t g_pfGdipDeleteGraphics;

typedef GpStatus (WINGDIPAPI * pfGdipLoadImageFromFile_t) (GDIPCONST WCHAR* filename, GpImage **image);
static pfGdipLoadImageFromFile_t g_pfGdipLoadImageFromFile;

typedef GpStatus (WINGDIPAPI * pfGdipDisposeImage_t) (GpImage *image);
static pfGdipDisposeImage_t g_pfGdipDisposeImage;

typedef GpStatus (WINGDIPAPI * pfGdipGetImageWidth_t) (GpImage *image, UINT *width);
static pfGdipGetImageWidth_t g_pfGdipGetImageWidth;

typedef GpStatus (WINGDIPAPI * pfGdipGetImageHeight_t) (GpImage *image, UINT *height);
static pfGdipGetImageHeight_t g_pfGdipGetImageHeight;

typedef GpStatus (WINGDIPAPI * pfGdipDrawImageRectRectI_t) (
	GpGraphics *graphics, GpImage *image,
	INT dstx, INT dsty, INT dstwidth, INT dstheight,
	INT srcx, INT srcy, INT srcwidth, INT srcheight,
	GpUnit srcUnit, GDIPCONST GpImageAttributes* imageAttributes, DrawImageAbort callback, VOID * callbackData
);
static pfGdipDrawImageRectRectI_t g_pfGdipDrawImageRectRectI;

bool InitGdiPlus () {
	if (g_hGDIPlusDll)
		return true;

	ReportText ("Initializing GDI+ API...");

	// First try to load from system directory.
	// NOTE: under Windows XP, it is required to use the system-supplied gdiplus.dll.
	char strSystem [MAX_PATH];
	::GetSystemDirectory (strSystem, MAX_PATH);
	::strcat (strSystem, "\\gdiplus.dll");
	g_hGDIPlusDll = ::LoadLibraryEx (strSystem, NULL, LOAD_WITH_ALTERED_SEARCH_PATH);
	if (! g_hGDIPlusDll) {
		// Second, try local one.
		g_hGDIPlusDll = ::LoadLibrary ("gdiplus.dll");
		if (g_hGDIPlusDll == NULL) {
			// TODO: Extract from resource?
			ReportWarning ("Failed to load GdiPlus.dll.");
			return false;
		}
	}

	// Find entry points.
	g_pfGdiplusStartup         = reinterpret_cast<pfGdiplusStartup_t>         (::GetProcAddress (g_hGDIPlusDll, "GdiplusStartup"));
	g_pfGdiplusShutdown        = reinterpret_cast<pfGdiplusShutdown_t>        (::GetProcAddress (g_hGDIPlusDll, "GdiplusShutdown"));
	g_pfGdipCreateFromHDC      = reinterpret_cast<pfGdipCreateFromHDC_t>      (::GetProcAddress (g_hGDIPlusDll, "GdipCreateFromHDC"));
	g_pfGdipDeleteGraphics     = reinterpret_cast<pfGdipDeleteGraphics_t>     (::GetProcAddress (g_hGDIPlusDll, "GdipDeleteGraphics"));
	g_pfGdipLoadImageFromFile  = reinterpret_cast<pfGdipLoadImageFromFile_t>  (::GetProcAddress (g_hGDIPlusDll, "GdipLoadImageFromFile"));
	g_pfGdipDisposeImage       = reinterpret_cast<pfGdipDisposeImage_t>       (::GetProcAddress (g_hGDIPlusDll, "GdipDisposeImage"));
	g_pfGdipGetImageWidth      = reinterpret_cast<pfGdipGetImageWidth_t>      (::GetProcAddress (g_hGDIPlusDll, "GdipGetImageWidth"));
	g_pfGdipGetImageHeight     = reinterpret_cast<pfGdipGetImageHeight_t>     (::GetProcAddress (g_hGDIPlusDll, "GdipGetImageHeight"));
	g_pfGdipDrawImageRectRectI = reinterpret_cast<pfGdipDrawImageRectRectI_t> (::GetProcAddress (g_hGDIPlusDll, "GdipDrawImageRectRectI"));
	if (
		! g_pfGdiplusStartup ||
		! g_pfGdiplusShutdown ||
		! g_pfGdipCreateFromHDC ||
		! g_pfGdipDeleteGraphics ||
		! g_pfGdipLoadImageFromFile ||
		! g_pfGdipDisposeImage ||
		! g_pfGdipGetImageWidth ||
		! g_pfGdipGetImageHeight ||
		! g_pfGdipDrawImageRectRectI
	) {
		ReportWarning ("Cannot find entry points in gdiplus.dll.");
		return false;
	}

	//Initialize GDI+.
	ULONG_PTR token;
	GdiplusStartupInput si;
	Status s = g_pfGdiplusStartup (& token, & si, NULL);
	if (s != 0) {
		ReportWarning ("Failed to initialize GDI+. Error code: %d.", s);
		g_hGDIPlusDll = NULL;
		return false;
	}

	return true;
}

static
BOOL CALLBACK DrawImageAbortProc (VOID *) {return g_bStop ? TRUE : FALSE;}

bool LoadImageByGdiPlus (const char * _strFileName, const HDC _hWndDC, const HDC _hMemDC, HBITMAP & _hBitmap, size_t & _cWidth, size_t & _cHeight) {
	// Convert file name to Unicode.
	WCHAR wstrFileName [_MAX_PATH + 1];
	::MultiByteToWideChar (::GetACP (), 0, _strFileName, -1, wstrFileName, _MAX_PATH);

	// Load GDI+.
	if (! InitGdiPlus ())
		return false;

	// Load image from file.
	GpImage * pImage = NULL;
	const GpStatus s = g_pfGdipLoadImageFromFile (wstrFileName, & pImage);
	if (s != 0) {
		if (s == UnknownImageFormat)
			ReportWarning ("Unknown image format.");
		else
			ReportWarning ("GDI+ failed to load image. Error code: %d.", s);

		if (pImage)
			g_pfGdipDisposeImage (pImage);
		return false;
	}

	// Get image's dimensions in pixels.
	g_pfGdipGetImageWidth (pImage, & _cWidth);
	g_pfGdipGetImageHeight (pImage, & _cHeight);

	SetProgressStatus (5);

	// Prepare temporary bitmap.
	assert (! _hBitmap);
	_hBitmap = ::CreateCompatibleBitmap (_hWndDC, _cWidth, _cHeight);
	if (_hBitmap == NULL) {
		ReportWarning ("Not enough memory to render image.");
		ReportText ("Try to change current video mode to diminish color depth.");

		if (pImage)
			g_pfGdipDisposeImage (pImage);
		return false;
	}
	::SelectObject (_hMemDC, _hBitmap);

	SetProgressStatus (6);

	// Render the map image to the temporary bitmap.
	GpGraphics * pGraphics = NULL;
	g_pfGdipCreateFromHDC (_hMemDC, & pGraphics);
	const GpStatus rs = g_pfGdipDrawImageRectRectI (pGraphics, pImage, 0, 0, _cWidth, _cHeight, 0, 0, _cWidth, _cHeight, UnitPixel, NULL, DrawImageAbortProc, NULL);
	if (rs != Ok)
		ReportWarning ("Failed to render image (error code=%d).", s);
	g_pfGdipDeleteGraphics (pGraphics);

	// Destroy image.
	g_pfGdipDisposeImage (pImage);

	return rs == Ok;
}

//////////////////////////////////////////

bool LoadImageByLoadPicture (const char * _strFileName, const HDC _hWndDC, const HDC _hMemDC, HBITMAP & _hBitmap, size_t & _cWidth, size_t & _cHeight) {
	ReportText ("Trying to load map image using LoadPicture() API...");

	// Check image file type.
	const char * strFileExt = ::strrchr (_strFileName, '.');
	if (strFileExt == NULL) {
		ReportWarning ("Map image type is not supported.");
		return false;
	}
	++ strFileExt;
	if (
		::stricmp (strFileExt, "bmp") != 0 &&
		::stricmp (strFileExt, "jpg") != 0 &&
		::stricmp (strFileExt, "gif") != 0
	) {
		ReportWarning ("Image type '.%s' is not supported by LoadPicture() API.", strFileExt);
		return false;
	}

	// Load image file.
	const CHandle hFile = ::CreateFile (_strFileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if (hFile == NULL || hFile == INVALID_HANDLE_VALUE) {
		ReportWarning ("Can't open map image fule: %s.", _strFileName);
		return false;
	}

	const DWORD dwSize = ::GetFileSize (hFile, NULL);
	HGLOBAL hGlobal = ::GlobalAlloc (GMEM_MOVEABLE, dwSize);
	void * pFile = ::GlobalLock (hGlobal);
	DWORD dwRead = 0;
	::ReadFile (hFile, pFile, dwSize, & dwRead, NULL);
	if (dwRead < dwSize) {
		::GlobalUnlock (hGlobal);
		::GlobalFree (hGlobal);

		ReportWarning ("Can't read map image file: %s.", _strFileName);
		return false;
	}

	CCom com;

	IStream * pStream = NULL;
	::CreateStreamOnHGlobal (hGlobal, TRUE, & pStream);
	CUnknownPtr pStreamGuard (pStream);

	IPicture * pPicture = NULL;
	HRESULT hr = ::OleLoadPicture (pStream, dwSize, FALSE, IID_IPicture, reinterpret_cast<void **> (& pPicture));
	CUnknownPtr pPictureGuard (pPicture);
	if (FAILED (hr)) {
		ReportWarning ("Failed to read map image: '%s'. Error code: 0x%08x.", _strFileName, hr);
		return false;
	}

	SetProgressStatus (5);

	//
	// Transform the image.
	//

	// Get image dimensions.
	OLE_XSIZE_HIMETRIC hmWidth;
	pPicture->get_Width (& hmWidth);
	_cWidth = (hmWidth + 1)*::GetDeviceCaps (_hMemDC, LOGPIXELSX)/2540;

	OLE_YSIZE_HIMETRIC hmHeight;
	pPicture->get_Height (& hmHeight);
	_cHeight = (hmHeight + 1)*::GetDeviceCaps (_hMemDC, LOGPIXELSY)/2540;

	// Create temporary bitmap used as source buffer.
	// NOTE: It should be DIB to have fast access to its bits.
	assert (_hBitmap);
	_hBitmap = ::CreateCompatibleBitmap (_hWndDC, _cWidth, _cHeight);
	::SelectObject (_hMemDC, _hBitmap);

	SetProgressStatus (6);

	// Render the map image to the temporary bitmap.
	hr = pPicture->Render (_hMemDC, 0, _cHeight, _cWidth, -(int)_cHeight, 0, 0, hmWidth, hmHeight, NULL);
	if (FAILED (hr)) {
		ReportWarning ("Failed to render map image: '%s'. Error code: 0x%08x.", _strFileName, hr);
		return false;
	}

	return true;
}

//////////////////////////////////////////

class COziMapLoader : public CMapLoader {
  public:
	// Override.
	bool IsValid (const char * _pData, size_t _cFileSize) const;
	bool Load (const char * _pData, size_t _cFileSize, const char * _strFileName, map_t * _pMap);

  private:
	const char * ParseLine (const char * _p);

	map_t * m_pMap;
	const char * m_pStart, * m_pEnd;

	// The counter of current line.
	size_t m_cFileLine;

	// The map image file name.
	char m_strImageName [MAX_PATH];

	enum {
		ptSimple, ptMercator, ptGauss, ptUTM, ptLambert, ptUnknown
	} m_Projection;

	datum_t m_Datum;

	point_t m_points [4];
	float   m_lScreenX [4];
	float   m_lScreenY [4];
	DWORD   m_dwPointDefined;

	rect_t  m_rectBound;
};

DECLARE_MAP_LOADER (COziMapLoader, "map", "OziExplorer map (*.map)\0*.map", false);

bool COziMapLoader::IsValid (const char * _pData, size_t _cFileSize) const {
	if (! CheckAnsiFile (_pData, _cFileSize, true))
		return false;

	return HasPrefix (_pData, _pData + _cFileSize, "OziExplorer Map Data File", 25);
}

const char * COziMapLoader::ParseLine (const char * _p) {
	const char * const pNextLine = FindNextLine (_p, m_pEnd);

	if (_p [0] != '\n' && _p [0] != '\r') {
		char * const pLine = reinterpret_cast<char *> (::alloca (pNextLine - _p));
		::memcpy (pLine, _p, pNextLine - _p - 1);
		pLine [pNextLine - _p - 1] = '\0';

		if (m_cFileLine == 3) {
			// Get map image file name.
			::memcpy (m_strImageName, _p, pNextLine - _p - 1);
			m_strImageName [pNextLine - _p - 1] = '\0';
		} else if (m_cFileLine == 5) {
			if (HasPrefix (_p, pNextLine, "WGS 84,", 7))
				m_Datum = dtWGS84;
			else if (HasPrefix (_p, pNextLine, "Pulkovo 1942", 12))
				m_Datum = dtPulkovo42;
			else
				m_Datum = dtUnknown;
		}

		if (HasPrefix (_p, pNextLine, "Map Projection,", 15)) {
			if (HasPrefix (_p + 15, pNextLine, "Mercator,", 9))
				m_Projection = ptMercator;
			else if (HasPrefix (_p + 15, pNextLine, "Latitude/Longitude,", 19))
				m_Projection = ptSimple;
			else if (HasPrefix (_p + 15, pNextLine, "Lambert Conformal Conic,", 24))
				m_Projection = ptLambert;
			else if (HasPrefix (_p + 15, pNextLine, "Transverse Mercator,", 20))
				m_Projection = ptGauss;
			else {
				ReportWarning (_p - m_pStart, "Unsupported projection type.");
				m_Projection = ptSimple;
			}
		} else if (HasPrefix (_p, pNextLine, "MMPNUM,", 7)) {
			int i = 0;
			if (::sscanf (pLine + 7, "%d", & i) != 1)
				ReportWarning (_p - m_pStart, "Unexpected format of MMPNUM.");
			if (i != 4)
				ReportWarning (_p - m_pStart, "Unexpected value of MMPNUM. Only 4 is supported.");
		} else if (HasPrefix (_p, pNextLine, "MMPXY,", 6)) {
			int i = 0;
			POINT pt;
			if (::sscanf (pLine + 6, "%d,%d,%d", & i, & pt.x, & pt.y) != 3)
				ReportWarning (_p - m_pStart, "Unexpected format of MMPXY line.");
			m_lScreenX [i - 1] = pt.x;
			m_lScreenY [i - 1] = pt.y;
			m_dwPointDefined |= 1 << (i - 1);
		} else if (HasPrefix (_p, pNextLine, "MMPLL,", 6)) {
			int i = 0;
			point_t point;
			if (::sscanf (pLine + 6, "%d,%f,%f", & i, & point.x, & point.y) != 3)
				ReportWarning (_p - m_pStart, "Unexpected format of MMPLL line.");
			if (1 <= i && i <= 4)
				// TODO: check the point' range.
				m_points [i - 1] = point;
		}
	}

	if (pNextLine [-1] == '\n')
		// Count lines.
		++ m_cFileLine;

	return pNextLine;
}

bool COziMapLoader::Load (const char * _pData, size_t _cFileSize, const char * _strMapFileName, map_t * _pMap) {
	m_pMap = _pMap;
	m_pStart = _pData;
	m_pEnd = _pData + _cFileSize;

	//
	// Parse .map file.
	//

	m_cFileLine = 1;
	m_strImageName [0] = '\0';
	m_rectBound = rect_t ();
	m_Projection = ptSimple;
	m_dwPointDefined = 0;
	m_Datum = dtUnknown;

	if (! CheckAnsiFile (_pData, _cFileSize, false))
		return false;

	SetProgressStatus (0);

	const char * pLine = _pData;
	do {
		pLine = ParseLine (pLine);
//		SetProgressStatus (fast_floor ((10.*(pLine - _pData))/_cFileSize + .5));
	} while (pLine < _pData + _cFileSize && ! g_bStop);

	SetProgressStatus (1);

	// Check all required fields are defined.
	if (m_dwPointDefined != 0xF) {
		ReportWarning ("Not enough points (4 expected in MMPXY and MMPLL).");
		return false;
	}

	//
	// Convert datum.
	//

	if (m_Datum == dtUnknown)
		ReportWarning ("Some unsupported datum is used. No datum conversion performed.");
	else if (m_Datum != dtWGS84) {
		ReportText ("Performing datum conversion to WGS84...");

//		ReportWarning ("Elevation is unknown. 0 m will be used.");
		for (size_t cPoint = 0; cPoint < 4; ++ cPoint)
			ToWGS84 (m_Datum, m_points [cPoint], 0);
	}

	//
	// Make calculations.
	//

	ReportText ("Performing projection calculations...");

	size_t cPoint;
	for (cPoint = 0; cPoint < 4; ++ cPoint)
		m_rectBound.Extend (m_points [cPoint]);
	
	if (m_Projection != ptSimple && m_Projection != ptGauss && m_Projection != ptUTM) {
		ReportWarning ("The projection type is not supported.");
		return false;
	}

	float Axx, Axy, Bx, Dxxy = 0;
	float Ayx, Ayy, By, Dyxx = 0;
	if (m_Projection == ptSimple/* || m_Projection == ptMercator*/) {
		if (
			! Solve3x3 (m_points, m_lScreenX, Axx, Axy, Bx) ||
			! Solve3x3 (m_points, m_lScreenY, Ayx, Ayy, By)
		) {
			ReportWarning ("Some points belong to a line.");
			return false;
		}
	} else if (m_Projection == ptGauss || m_Projection == ptUTM) {
		if (
			! Solve4x4_XY (m_points, m_lScreenX, Axx, Axy, Bx, Dxxy) ||
			//! Solve4x4_XX (m_points, m_lScreenY, Ayx, Ayy, By, Dyxx)
			! Solve3x3 (m_points, m_lScreenY, Ayx, Ayy, By)
		) {
			ReportWarning ("Some points belong to a line.");
			return false;
		}
	}

	// Check distorsion of the map.
	for (cPoint = 0; cPoint < 4; ++ cPoint) {
		const float fScreenDX = m_lScreenX [cPoint] - (Bx + Axx*m_points [cPoint].x + Axy*m_points [cPoint].y + Dxxy*m_points [cPoint].x*m_points [cPoint].y);
		const float fScreenDY = m_lScreenY [cPoint] - (By + Ayx*m_points [cPoint].x + Ayy*m_points [cPoint].y + Dyxx*m_points [cPoint].x*m_points [cPoint].x);
		if (fScreenDX*fScreenDX + fScreenDY*fScreenDY >= 1)
			ReportWarning ("Distorsion in (%d) image corner is %.1f, %.1f (in pixels).", cPoint + 1, fScreenDX, fScreenDY);
	}

	//
	// Load map image file.
	//

	ReportText ("Loading map image '%s'...", m_strImageName);

	// Check if image file exists.
	{
		CHandle hFile = ::CreateFile (m_strImageName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, NULL);
		if (hFile == NULL || hFile == INVALID_HANDLE_VALUE) {
			ReportText ("Failed to open '%s'.", m_strImageName);

			// Make image name local for folder with _strMapFileName.
			char strLocalImageName [MAX_PATH];
			::strcpy (strLocalImageName, _strMapFileName);
			char * strMapPathEnd = ::strrchr (strLocalImageName, '\\');
			if (strMapPathEnd)
				++ strMapPathEnd;
			else 
				strMapPathEnd = strLocalImageName;

			const char * strImageName = ::strrchr (m_strImageName, '\\');
			if (strImageName)
				++ strImageName;
			else
				strImageName = m_strImageName;

			::strcpy (strMapPathEnd, strImageName);

			ReportText ("Trying '%s'...", strLocalImageName);

			// Try the 'local' name.
			hFile = ::CreateFile (strLocalImageName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, NULL);
			if (hFile == NULL || hFile == INVALID_HANDLE_VALUE) {
				ReportWarning ("Can't open map image: %s.", m_strImageName);
				return false;
			}

			::strcpy (m_strImageName, strLocalImageName);
		}
	}

	const CClientDC hWndDC (g_hMapWnd);

	const CDCHolder hMemDC1 = ::CreateCompatibleDC (hWndDC);
	HBITMAP hBmp1 = NULL;
	size_t cWidthInPixels = 0, cHeightInPixels = 0;
	if (
		! LoadImageByGdiPlus     (m_strImageName, hWndDC, hMemDC1, hBmp1, cWidthInPixels, cHeightInPixels) &&
		! LoadImageByLoadPicture (m_strImageName, hWndDC, hMemDC1, hBmp1, cWidthInPixels, cHeightInPixels)
	) {
		if (hBmp1)
			::DeleteObject (hBmp1);

		ReportWarning ("Failed to render image file: '%s'.", m_strImageName);
		return false;
	}
	const CBitmapHolder hBmp1_Holder (hBmp1);

	if (cWidthInPixels <= 1 || cHeightInPixels <= 1) {
		ReportWarning ("Image should have dimensions > 1 pixel.");
		return false;
	}

	if (g_bStop)
		return false;

	//
	// Transform the image.
	//

	SetProgressStatus (8);

	// TODO:
	const size_t cWidthInPixels2 = cWidthInPixels;
	const size_t cHeightInPixels2 = cHeightInPixels;

	// Create bimap to keep image ready for rendering (i.e. transformed to the destination projection).
	// NOTE: it should be DDB to provide fast rendering using ::StretchBlt().
	CDCHolder hMemDC2 = ::CreateCompatibleDC (hWndDC);
	CBitmapHolder hBmp2 = ::CreateCompatibleBitmap (hWndDC, cWidthInPixels2, cHeightInPixels2);
	if (hBmp2 == NULL) {
		ReportWarning ("Not enough memory to transform image.");
		ReportText ("Try to change current video mode to diminish color depth.");
		return false;
	}
	::SelectObject (hMemDC2, hBmp2);

	SetProgressStatus (9);

	// Access source bits to perform projection's transform.
	BITMAPINFOHEADER bmih;
	::memset (& bmih, 0, sizeof (bmih));
	bmih.biSize = sizeof (bmih);
	bmih.biWidth = cWidthInPixels;
	bmih.biHeight = cHeightInPixels;
	bmih.biPlanes = 1;
	bmih.biBitCount = 24;
	bmih.biCompression = BI_RGB;
	bmih.biClrUsed = 0;
	bmih.biClrImportant = 0;

	const size_t cScanWidth1 = Align4 (3*cWidthInPixels);
	BYTE * const pBits1 = new BYTE [cScanWidth1*cHeightInPixels];
	if (pBits1 == NULL) {
		ReportWarning ("Not enough memory.");
		return false;
	}
	CByteArrayHolder pBits1_Holder (pBits1);
	for (size_t cScanLine = 0; cScanLine < cHeightInPixels; cScanLine += 100) {
		if (g_bStop)
			return false;

		::GetDIBits (
			hMemDC1, hBmp1, 
			cScanLine, cHeightInPixels - cScanLine > 100 ? 100 : cHeightInPixels - cScanLine,
			pBits1 + cScanLine*cScanWidth1, reinterpret_cast <BITMAPINFO *> (& bmih), DIB_RGB_COLORS
		);

		SetProgressStatus (fast_floor (10 + 10.*cScanLine/cHeightInPixels + .5));
	}

	// Prepare destination bits.
	const size_t cScanWidth2 = Align4 (3*cWidthInPixels2);
	BYTE * const pBits2 = new BYTE [cScanWidth2*cHeightInPixels2];
	if (pBits2 == NULL) {
		ReportWarning ("Not enough memory.");
		return false;
	}
	CByteArrayHolder pBits2_Holder (pBits2);
	::memset (pBits2, 255, cScanWidth2*cHeightInPixels2);

	SetProgressStatus (20);
	ReportText ("Performing image transform...");

	// Re-scale the rect to take right and bottom edges into the visible area.
	m_rectBound.x1 += m_rectBound.Width  ()/(cWidthInPixels2 - 1);
	m_rectBound.y1 += m_rectBound.Height ()/(cHeightInPixels2 - 1);

	// Perform the transform.
	for (size_t y2 = 0; y2 < cHeightInPixels2; ++ y2) {
		if (g_bStop)
			return false;

		// Latitude, in degrees.
		const float fY2 = m_rectBound.y0 + y2*m_rectBound.Height ()/cHeightInPixels2;
		for (size_t x2 = 0; x2 < cWidthInPixels2; ++ x2) {
			// Longitude, in degrees.
			const float fX2 = m_rectBound.x0 + x2*m_rectBound.Width ()/cWidthInPixels2;

			// Coordinates inside pBits1.
			const float iX1 = Axx*fX2 + Axy*fY2 + Bx + Dxxy*fX2*fY2 + .5;
			if (iX1 < 0)
				continue;
			const size_t x1 = fast_floor (iX1);
			if (x1 > cWidthInPixels - 1)
				continue;
			const float iY1 = Ayx*fX2 + Ayy*fY2 + By + Dyxx*fX2*fX2 + .5;
			if (iY1 < 0)
				continue;
			const size_t y1 = fast_floor (iY1);
			if (y1 > cHeightInPixels - 1)
				continue;

			const BYTE * const pPixel1 = pBits1 + x1*3 + (cHeightInPixels - 1 - y1)*cScanWidth1;
			BYTE       * const pPixel2 = pBits2 + x2*3 + y2*cScanWidth2;
			pPixel2 [0] = pPixel1 [0];
			pPixel2 [1] = pPixel1 [1];
			pPixel2 [2] = pPixel1 [2];
		}

		SetProgressStatus (fast_floor (20 + 80.*y2/cHeightInPixels2 + .5));
	}

	::SetDIBits (hMemDC2, hBmp2, 0, cHeightInPixels2, pBits2, reinterpret_cast <BITMAPINFO *> (& bmih), DIB_RGB_COLORS);

	// Shift the visible rectangle to make binding to the center of points.
	m_rectBound.x0 -= .5f*m_rectBound.Width  ()/(cWidthInPixels2 - 1);
	m_rectBound.x1 -= .5f*m_rectBound.Width  ()/(cWidthInPixels2 - 1);
	m_rectBound.y0 -= .5f*m_rectBound.Height ()/(cHeightInPixels2 - 1);
	m_rectBound.y1 -= .5f*m_rectBound.Height ()/(cHeightInPixels2 - 1);

	// Append the resulting image to the map.
	m_pMap->images.push_back (image_t ());
	image_t & image = m_pMap->images.back ();
	image.rect    = m_rectBound;
	image.hBitmap = hBmp2.Detach ();
	image.hMemDC  = hMemDC2.Detach ();
	image.lWidth  = cWidthInPixels2;
	image.lHeight = cHeightInPixels2;

	image.MakeSmall ();

	image.strFileName = _strMapFileName;

	// Suppress warning about empty Zoom etc.
	m_pMap->bNoHeader = true;

	m_pMap->cHeaderLen = _cFileSize;

	SetProgressStatus (100);

	return true;
}
