# include "stdafx.h"
# include "Map.h"
# include "Types.h"
# include "Datum.h"
# include "Load.h"
# include "Log.h"
# include "Status.h"
# include "Globals.h"

class COziPltLoader : public CMapLoader {
  public:
	// Override.
	bool IsValid (const char * _pData, size_t _cFileSize) const;
	bool Load (const char * _pData, size_t _cFileSize, const char * _strFileName, map_t * _pMap);

  private:
	void ParseTrackpoint (const char * _strData, const char * _strEnd);

	map_t * m_pMap;
	const char * m_pStart, * m_pEnd;

	// The counter of current line.
	size_t m_cFileLine;
	//
	size_t m_cOrdinal;

	datum_t m_Datum;

	points_t m_points;
	size_t m_cTrackLine;
	size_t m_cTrackBeginPos;
	std::string m_strTrackLabel;
	void SaveTrack (const char * _strEnd);

	file_refs_t m_file_refs;
};

DECLARE_MAP_LOADER (COziPltLoader, "plt", "OziExplorer track (*.plt)\0*.plt", false);

bool COziPltLoader::IsValid (const char * _pData, size_t _cFileSize) const {
	if (! CheckAnsiFile (_pData, _cFileSize, true))
		return false;

	return HasPrefix (_pData, _pData + _cFileSize, "OziExplorer Track Point File", 28);
}

void COziPltLoader::SaveTrack (const char * _strEnd) {
	if (m_points.empty ())
		return;

	//
	// Add the track to the map.
	//

	m_pMap->layers [2].rgns.push_back (rgn_t ());
	++ m_pMap->layers [2].cRgns;

	rgn_t & rgn = m_pMap->layers [2].rgns.back ();

	m_pMap->InsertRgnToSaveList (rgn);

	//
	//
	//

	rgn.dwFlags |= rgn_t::fWaypointOrTrack;

	rgn.cBeginPos = m_cTrackBeginPos;
	rgn.cEndPos = _strEnd - m_pStart;
	rgn.cFileLine = m_cTrackLine;
	rgn.cOrdinal = m_cOrdinal ++;

	rgn.strLabel = m_strTrackLabel;

	rgn.elements.push_back (rgn_element_t ());
	rgn_element_t & element = rgn.elements.back ();
	element.cFileLine = m_cTrackLine;
	element.cFileOffset = m_cTrackBeginPos;

	element.cLevel = 0;
	element.points.swap (m_points);

	rgn.pTypeDesc = & g_types [0];
	rgn.ushKind = 2;

	rgn.pFileRef = & m_file_refs.back ();
}

	// Find end of field.
static
const char * FindNextField (const char * _pBegin, const char * _pEnd) {
	for (const char * p = _pBegin; p < _pEnd; ++ p)
		if (* p == ',')
			return p + 1 < _pEnd ? p + 1 : _pEnd;

	return _pEnd;
}

void COziPltLoader::ParseTrackpoint (const char * _strData, const char * _strEnd) {
	//
	// Parse fields.
	//

	const char * const strLat = _strData;
	const char * const strLon = FindNextField (strLat, _strEnd);
	const char * const strBegin = FindNextField (strLon, _strEnd);
	if (strBegin == _strEnd || strLat == _strEnd || strLon == _strEnd) {
		ReportWarning (_strData - m_pStart, "Unexpected format of trackpoint.");
		return;
	}

	if (strBegin [0] != '0') {
		SaveTrack (_strData);

		m_cTrackLine = m_cFileLine;
		m_cTrackBeginPos = _strData - m_pStart;
	}

	point_t point;
	const char * pEnd = NULL;
	if (
		! ParseFloat (strLat, _strEnd, & pEnd, point.y) ||
		! ParseFloat (strLon, _strEnd, & pEnd, point.x)
	) {
		ReportWarning (_strData - m_pStart, "Unexpected format of trckpoint data.");
		return;
	}
	m_points.push_back (point);
}

bool COziPltLoader::Load (const char * _pData, size_t _cFileSize, const char * _strMapFileName, map_t * _pMap) {
	if (! CheckAnsiFile (_pData, _cFileSize, false))
		return false;

	m_pMap = _pMap;
	m_pStart = _pData;
	m_pEnd = _pData + _cFileSize;

	//
	// Parse .plt file.
	//

	m_cFileLine = 1;
	m_cOrdinal = 0;
	m_Datum = dtUnknown;

	m_cTrackLine = 1;
	m_cTrackBeginPos = 0;
	m_strTrackLabel = _strMapFileName;

	m_file_refs.clear ();
	m_file_refs.push_back (file_ref_t ());
	m_file_refs.back ().strFileName = _strMapFileName;
	m_file_refs.back ().bAttachment = true;

	SetProgressStatus (0);

	const char * pLine = _pData;
	do {
		const char * const pNextLine = FindNextLine (pLine, m_pEnd);

		if (pLine [0] != '\n' && pLine [0] != '\r') {
			if (m_cFileLine == 2) {
				if (HasPrefix (pLine, pNextLine, "WGS 84", 6))
					m_Datum = dtWGS84;
				else if (HasPrefix (pLine, pNextLine, "Pulkovo 1942", 12))
					m_Datum = dtPulkovo42;
				else
					m_Datum = dtUnknown;
			} else if (m_cFileLine >= 7) {
				ParseTrackpoint (pLine, pNextLine);

				if (m_cFileLine == 7) {
					m_pMap->cHeaderLen = pLine - m_pStart;

					m_cTrackLine = m_cFileLine;
					m_cTrackBeginPos = pLine - m_pStart;
				}
			}
		}

		if (pNextLine [-1] == '\n')
			// Count lines.
			++ m_cFileLine;

		pLine = pNextLine;

		SetProgressStatus (fast_floor ((50.*(pLine - _pData))/_cFileSize + .5));
	} while (pLine < _pData + _cFileSize && ! g_bStop);

	SetProgressStatus (50);

	SaveTrack (m_pEnd);

	// Convert datum.
	if (m_Datum == dtUnknown)
		ReportWarning ("Some unsupported datum is used. No datum conversion performed.");
	else if (m_Datum != dtWGS84) {
		ReportText ("Performing datum conversion to WGS84...");
		m_pMap->ToWGS84 (m_Datum, 0);
	}

	m_pMap->file_refs.swap (m_file_refs);

	// Suppress warning about empty Zoom etc.
	m_pMap->bNoHeader = true;

	return true;
}