# include "stdafx.h"
# include "Map.h"
# include "Types.h"
# include "Datum.h"
# include "Load.h"
# include "Log.h"
# include "Status.h"
# include "Globals.h"

class COziWptLoader : public CMapLoader {
  public:
	// Override.
	bool IsValid (const char * _pData, size_t _cFileSize) const;
	bool Load (const char * _pData, size_t _cFileSize, const char * _strFileName, map_t * _pMap);

  private:
	void ParseWaypoint (const char * _strData, const char * _strEnd);
	const char * ParseLine (const char * _p);

	map_t * m_pMap;
	const char * m_pStart, * m_pEnd;

	// The counter of current line.
	size_t m_cFileLine;
	//
	size_t m_cOrdinal;

	datum_t m_Datum;

	file_refs_t m_file_refs;
};

DECLARE_MAP_LOADER (COziWptLoader, "wpt", "OziExplorer waypoints (*.wpt)\0*.wpt", false);

bool COziWptLoader::IsValid (const char * _pData, size_t _cFileSize) const {
	if (! CheckAnsiFile (_pData, _cFileSize, true))
		return false;

	return HasPrefix (_pData, _pData + _cFileSize, "OziExplorer Waypoint File", 25);
}

	// Find end of field.
static
const char * FindNextField (const char * _pBegin, const char * _pEnd) {
	for (const char * p = _pBegin; p < _pEnd; ++ p)
		if (* p == ',')
			return p + 1 < _pEnd ? p + 1 : _pEnd;

	return _pEnd;
}

void COziWptLoader::ParseWaypoint (const char * _strData, const char * _strEnd) {
	//
	// Parse fields.
	//

	const char * const strName = FindNextField (_strData, _strEnd);
	const char * const strLat = FindNextField (strName, _strEnd);
	const char * const strLon = FindNextField (strLat, _strEnd);
	if (strName == _strEnd || strLat == _strEnd || strLon == _strEnd) {
		ReportWarning (_strData - m_pStart, "Unexpected format of waypoint.");
		return;
	}

	//
	// Create new POI.
	//

	rgn_t rgn;
	rgn.cBeginPos = _strData - m_pStart;
	rgn.cEndPos = _strEnd - m_pStart;
	rgn.cFileLine = m_cFileLine;
	rgn.cOrdinal = m_cOrdinal ++;

	rgn.dwFlags |= rgn_t::fWaypointOrTrack;

	// Get name.
	rgn.strLabel = strName;
	// Replace 209 with comma.
	for (size_t c = 0; rgn.strLabel [c]; ++ c)
		if (rgn.strLabel [c] == (char) 209)
			rgn.strLabel [c] = ',';

	rgn.origins.push_back (rgn_origin_t ());
	rgn_origin_t & origin = rgn.origins.back ();
	origin.cLevel = 0;

	// Parse coordinates.
	const char * pEnd = NULL;
	if (
		! ParseFloat (strLat, _strEnd, & pEnd, origin.point.y) ||
		! ParseFloat (strLon, _strEnd, & pEnd, origin.point.x)
	) {
		ReportWarning (_strData - m_pStart, "Unexpected format of waypoint data.");
		return;
	}

	// Set type to unknown.
	rgn.pTypeDesc = & g_types [0];
	rgn.ushKind = 0;

	rgn.pFileRef = & m_file_refs.back ();

	//
	// Add the waypoint to the map.
	//

	m_pMap->layers [rgn.ushKind].rgns.push_back (rgn);
	++ m_pMap->layers [rgn.ushKind].cRgns;
	m_pMap->InsertRgnToSaveList (m_pMap->layers [rgn.ushKind].rgns.back ());
}

const char * COziWptLoader::ParseLine (const char * _p) {
	const char * const pNextLine = FindNextLine (_p, m_pEnd);

	if (_p [0] != '\n' && _p [0] != '\r') {
		if (m_cFileLine == 2) {
			if (HasPrefix (_p, pNextLine, "WGS 84", 6))
				m_Datum = dtWGS84;
//			else if (HasPrefix (_p, pNextLine, "Pulkovo 1942", 12))
//				m_Datum = dtPulkovo42;
			else
				m_Datum = dtUnknown;
		} else if (m_cFileLine >= 5) {
			ParseWaypoint (_p, pNextLine);

			if (m_cFileLine == 5)
				m_pMap->cHeaderLen = _p - m_pStart;
		}
	}

	if (pNextLine [-1] == '\n')
		// Count lines.
		++ m_cFileLine;

	return pNextLine;
}

bool COziWptLoader::Load (const char * _pData, size_t _cFileSize, const char * _strMapFileName, map_t * _pMap) {
	if (! CheckAnsiFile (_pData, _cFileSize, false))
		return false;

	m_pMap = _pMap;
	m_pStart = _pData;
	m_pEnd = _pData + _cFileSize;

	//
	// Parse .wpt file.
	//

	m_cFileLine = 1;
	m_cOrdinal = 0;
	m_Datum = dtUnknown;

	m_file_refs.clear ();
	m_file_refs.push_back (file_ref_t ());
	m_file_refs.back ().strFileName = _strMapFileName;
	m_file_refs.back ().bAttachment = true;

	SetProgressStatus (0);

	const char * pLine = _pData;
	do {
		pLine = ParseLine (pLine);
		SetProgressStatus (fast_floor ((50.*(pLine - _pData))/_cFileSize + .5));
	} while (pLine < _pData + _cFileSize && ! g_bStop);

	SetProgressStatus (50);

	// Convert datum.
	if (m_Datum == dtUnknown)
		ReportWarning ("Some unsupported datum is used. No datum conversion performed.");
	else if (m_Datum != dtWGS84) {
		ReportText ("Performing datum conversion to WGS84...");
		m_pMap->ToWGS84 (m_Datum, 0);
	}

	m_pMap->file_refs.swap (m_file_refs);

	// Suppress warning about empty Zoom etc.
	m_pMap->bNoHeader = true;

	return true;
}