# include "stdafx.h"
# include "PtAlgo.h"
# include "Globals.h"

void aff_transform_t::ApplyTo (point_t & _p) const {
	const float x = _p.x, y = _p.y;

	_p.x = fAxx*x + fAxy*y + fBx;
	_p.y = fAyx*x + fAyy*y + fBy;
}

void aff_transform_t::Multiply (const aff_transform_t _at2, aff_transform_t & _result) const {
	assert (& _result != this);

	_result.fAxx = fAxx*_at2.fAxx + fAxy*_at2.fAyx;
	_result.fAxy = fAxx*_at2.fAxy + fAxy*_at2.fAyy;
	_result.fAyx = fAyx*_at2.fAxx + fAyy*_at2.fAyx;
	_result.fAyy = fAyx*_at2.fAxy + fAyy*_at2.fAyy;

	_result.fBx = fBx + fAxx*_at2.fBx + fAxy*_at2.fBy;
	_result.fBy = fBy + fAyx*_at2.fBx + fAyy*_at2.fBy;
}

void aff_transform_t::Reverse (aff_transform_t & _result) const {
	assert (& _result != this);

	const float D = fAxx*fAyy - fAxy*fAyx;
	_result.fAxx = fAyy/D;
	_result.fAxy = -fAxy/D;
	_result.fAyx = -fAyx/D;
	_result.fAyy = fAxx/D;

	_result.fBx = - (_result.fAxx*fBx + _result.fAxy*fBy);
	_result.fBy = - (_result.fAyx*fBx + _result.fAyy*fBy);
}

//////////////////////////////////////////////////////////

float GetArea (const points_t & _points) {
	const size_t cPoints = _points.size ();
	if (cPoints <= 2)
		return 0;

	float fArea = 0;
	for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
		const point_t & p0 = _points [cPoint];
		const point_t & p1 = _points [cPoint + 1 == cPoints ? 0 : cPoint + 1];

		fArea += (p1.x - p0.x)*(p0.y + p1.y);
	}

	return ::fabs (fArea)*.5;
}

float GetLength (const points_t & _points) {
	const size_t cPoints = _points.size ();
	if (cPoints <= 1)
		return 0;

	const float fCosY = ::cos (_points [0].y*c_PI/180.);

	float fLengthX2 = 0, fLengthY2 = 0;
	for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
		const point_t & p0 = _points [cPoint];
		const point_t & p1 = _points [cPoint + 1 == cPoints ? 0 : cPoint + 1];

		fLengthX2 += (p1.x - p0.x)*(p1.x - p0.x);
		fLengthY2 += (p1.y - p0.y)*(p1.y - p0.y);
	}

	return ::sqrt (fLengthX2*fCosY*fCosY + fLengthY2);
}

bool PointContainsPt (const point_t & _point, const point_t & _pt, float _e) {
	return ::fabs (_point.x - _pt.x)*::cos (_point.y*c_PI/180.) <= _e && ::fabs (_point.y - _pt.y) <= _e;
}

bool PolylineContainsPt (const points_t & _pts, const point_t & _pt, float _e) {
	if (! _pts.empty ()) {
		const float fCosY = ::cos (_pts [0].y*c_PI/180.);
		for (size_t c = 0; c < _pts.size () - 1; ++ c) {
			const point_t & a = _pts [c];
			const point_t & b = _pts [c + 1];

			const float dx = (b.x - a.x)*fCosY;
			const float dy = b.y - a.y;
			const float r = ::sqrt (dx*dx + dy*dy);
			const float fSinA = dy/r;
			const float fCosA = dx/r;
			const float t = (_pt.x - a.x)*fCosY*fCosA + (_pt.y - a.y)      *fSinA;
			const float d = (_pt.y - a.y)      *fCosA - (_pt.x - a.x)*fCosY*fSinA;
			if (-_e < d && d < _e && -_e < t && t < r + _e)
				return true;  
		}
	}
	return false;
}

///////////////////////////////////////////////////////////////
//
// The code was obtained from pGPSMap written by Olexa Ryznik (www.olexa.au)
//
// Returns 1 if point is entirely within polygon,
// 0 if it is on its border and
// -1 if it is outside polygon.
int PolygonContainsPt (const points_t & _polygon, const point_t & _p) {
	bool bResult = false;

	int iLastTouchingLinkSide = 0;
	size_t cFirstPoint = 0;
	bool bFirstPointPassed = false;

	const size_t cPoints = _polygon.size ();
	for (size_t cPoint = 1; cPoint != cFirstPoint + 1 || ! bFirstPointPassed; ++ cPoint) {
		const point_t & p0 = _polygon [(cPoint - 1) % cPoints];

		if (cPoint >= cPoints)
			cPoint -= cPoints;
		const point_t & p1 = _polygon [cPoint];

		if (
			(p0.y < _p.y && _p.y < p1.y) ||
			(p0.y > _p.y && _p.y > p1.y)
		) {
			// Chain crosses point's latitude.
			const float x = p0.x + (_p.y - p0.y)*(p1.x - p0.x)/(p1.y - p0.y);
			if (x == _p.x)
				// Chain crosses the point.
				return 0;
			else if (x < _p.x)
				// Chain passes under the point.
				bResult = ! bResult;

			bFirstPointPassed = true;
		} else if (p0.y == _p.y && p1.y == _p.y) {
			// Chain is entirely within point's latitude.
			if (
				(p0.x <= _p.x && p1.x >= _p.x) ||
				(p0.x >= _p.x && p1.x <= _p.x)
			)
				// Chain crosses the point.
				return 0;

			if (cFirstPoint == cPoint - 1 && p1.x < _p.x)
				// There was no any link that crosses point's latitude or finishes at it yet.
				++ cFirstPoint;
		} else if (p0.y != _p.y && p1.y == _p.y) {
			// Chain finishes at point's latitude.
			if (p1.x == _p.x)
				// Chain crosses the point.
				return 0;
			else if (p1.x < _p.x)
				// Remember last touching chain side.
				iLastTouchingLinkSide = p0.y < _p.y ? -1 : 1;
			bFirstPointPassed = true;
		} else if (p0.y == _p.y && p1.y != _p.y) {
			// Chain starts at point's latitude.
			if (p0.x == _p.x)
				// Chain crosses the point.
				return 0;
			else if (p0.x < _p.x)
				if (iLastTouchingLinkSide == 0)
					// There was no touching chain yet.
					// We should loop through the polygon 'till this point.
					cFirstPoint = cPoint;
				else if (
					iLastTouchingLinkSide == -1 && p1.y > _p.y ||
					iLastTouchingLinkSide == 1 && p1.y < _p.y
				)
					// This liks with previous touching chain together cross point's latitude.
					bResult = ! bResult;
		} else
			// Chain does not cross point's latitude.
			bFirstPointPassed = true;
	}
	return bResult ? 1 : -1;
}

/*
	// Remove all cuts whose points are outside the given rect.
void _SimplifyPolygon (const points_t & _polygon, const rect_t & _rct, points_t & _ret) {
	assert (_polygon.size () >= 2);

	bool bPrevInside = _rct.Contains (_polygon [0]);

	const size_t cPoints = _polygon.size ();
	_ret.reserve (cPoints/2);
	for (size_t cPoint = 1; cPoint <= cPoints; ++ cPoint) {
		const point_t & p = _polygon [cPoint == cPoints ? 0 : cPoint];
		const bool bInside = _rct.Contains (p);

		if (bInside || bPrevInside)
			_ret.push_back (p);

		bPrevInside = bInside;
	}
}

bool PolygonContainsPolygon (const points_t & _Outer, const points_t & _Inner, const rect_t & _rctInnerBound) {
	return false;

	assert (_Outer.size () >= 2 && _Inner.size () >= 2);
	assert (_rctInnerBound.y0 <= _rctInnerBound.y1);

	points_t OuterSimple;
	_SimplifyPolygon (_Outer, _rctInnerBound, OuterSimple);
	if (OuterSimple.size () <= 2)
		return false;

	const size_t cPoints = _Inner.size ();
	for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint)
		if (PolygonContainsPt (OuterSimple, _Inner [cPoint]) <= 0)
			return false;
	return true;
}*/
