# ifndef __PtAlgo_H_
# define __PtAlgo_H_

//
// PtAlgo.h
//
// Definitions of simple geometry types: point, rectangle etc.
// NOTE: 4-byte float is used everywhere to save memory.
//

# include <vector>


//
// Point.
//
struct point_t {
	// Longitude and latitude, in degrees.
	float x, y;

	bool operator == (const point_t & _other) const {return x == _other.x && y == _other.y;}
};

typedef std::vector<point_t> points_t;

float GetArea (const points_t & _points);
float GetLength (const points_t & _points);
bool PointContainsPt (const point_t & _point, const point_t & _pt, float _e);
int PolygonContainsPt (const points_t & _polygon, const point_t & _p);
bool PolylineContainsPt (const points_t & _pts, const point_t & _pt, float _e);


//
// Parameters of affine transformation.
//
struct aff_transform_t {
	// The matrix elements.
	float fAxx, fAxy, fAyx, fAyy;
	// The shift vector.
	float fBx, fBy;

	void ApplyTo (point_t & _p) const;
	void Multiply (const aff_transform_t _at2, aff_transform_t & _result) const;
	void Reverse (aff_transform_t & _atReverse) const;
};


//
// Rectangle.
//
struct rect_t {
	// In degrees.
	float x0, y0, x1, y1;

	rect_t () : x0 (200), x1 (-200), y0 (200), y1 (-200) {}

	void swap (rect_t & _other) {
		std::swap (x0, _other.x0);
		std::swap (y0, _other.y0);
		std::swap (x1, _other.x1);
		std::swap (y1, _other.y1);
	}

	float Width  () const {return x1 > x0 ? x1 - x0 : 0;}
	float Height () const {return y1 > y0 ? y1 - y0 : 0;}

	float MeanX () const {return (x0 + x1)/2;}
	float MeanY () const {return (y0 + y1)/2;}

	bool Empty () const {return x1 <= x0 || y1 <= y0;}
	bool Invalid () const {return x1 < x0 || y1 < y0;}

	bool Intersects (const rect_t & _rect) const {
		if (x1 < _rect.x0 || x0 > _rect.x1)
			return false;
		if (y1 < _rect.y0 || y0 > _rect.y1)
			return false;
		return true;
	}

	bool Contains (const point_t & _point) const {
		return _point.x >= x0 && _point.x <= x1 && _point.y >= y0 && _point.y <= y1;
	}

	bool Contains (const rect_t & _rect) const {
		return x0 <= _rect.x0 && x1 >= _rect.x1 && y0 <= _rect.y0 && y1 >= _rect.y1;
	}

	void Extend (const point_t & _point) {
		if (_point.x < x0) x0 = _point.x;
		if (_point.x > x1) x1 = _point.x;
		if (_point.y < y0) y0 = _point.y;
		if (_point.y > y1) y1 = _point.y;
	}

	void Extend (const rect_t & _rect) {
		if (_rect.x0 < x0) x0 = _rect.x0;
		if (_rect.x1 > x1) x1 = _rect.x1;
		if (_rect.y0 < y0) y0 = _rect.y0;
		if (_rect.y1 > y1) y1 = _rect.y1;
	}

	static void Intersection (rect_t & _rectDest, const rect_t & _rectSrc1, const rect_t & _rectSrc2) {
		_rectDest.x0 = _rectSrc1.x0 > _rectSrc2.x0 ? _rectSrc1.x0 : _rectSrc2.x0;
		_rectDest.x1 = _rectSrc1.x1 < _rectSrc2.x1 ? _rectSrc1.x1 : _rectSrc2.x1;
		_rectDest.y0 = _rectSrc1.y0 > _rectSrc2.y0 ? _rectSrc1.y0 : _rectSrc2.y0;
		_rectDest.y1 = _rectSrc1.y1 < _rectSrc2.y1 ? _rectSrc1.y1 : _rectSrc2.y1;
	}
};

# endif // __PtAlgo_H_