# include "stdafx.h"
# include "Map.h"
# include "Types.h"
# include "Load.h"
# include "Log.h"
# include "Status.h"
# include "Globals.h"

# ifdef _MSC_VER
#	pragma pack (push, 1)
# endif

struct CFieldDesc {
	char  strName [11];
	BYTE  btType; // 'C' - string, 'N' - integer, 'L' - boolean, 'D' - date, 'M' - mark
	DWORD dwAddress;
	BYTE  btLen;
	BYTE  btDecOrdinal;
	WORD  wReserved0;
	BYTE  btWorkAreaID;
	WORD  wReserved1;
	BYTE  btSetFields;
	BYTE  baReserved [8];
};

struct CDbfHdr {
	BYTE btHasDbt; // 0x03 - no marks, 0x83 - have marks in .dbt

	BYTE btYear;
	BYTE btMonth;
	BYTE btDay;

	DWORD dwNumRecords;
	WORD wHdrSize;
	WORD wRecordSize;

	BYTE baReserved0 [3];
	BYTE baReserved1 [13];
	BYTE baReserved2 [4];

	CFieldDesc FieldDescriptors [1];
};

# ifdef _MSC_VER
#	pragma pack (pop)
# endif

class CDbfFile {
  public:
	CDbfFile (const char * _strFileName);
	~CDbfFile ();

	bool CantOpen () const {return m_pData == NULL;}

	const char * GetFieldName (size_t _cField) const {
		const CDbfHdr * const pHdr = reinterpret_cast<const CDbfHdr *> (m_pData);
		return pHdr->FieldDescriptors [_cField].strName;
	}

	size_t GetFieldLength (size_t _cField) const {
		const CDbfHdr * const pHdr = reinterpret_cast<const CDbfHdr *> (m_pData);
		return pHdr->FieldDescriptors [_cField].btLen;
	}

	const char * GetFieldS (size_t _cRecord, size_t _cField) const {
		assert (_cField < m_vecFieldOffsets.size ());
		return m_pRecords + _cRecord*m_cRecordSize + m_vecFieldOffsets [_cField] + 1;
	}

	int GetFieldN (size_t _cRecord, size_t _cField) const {
		const size_t cLen = GetFieldLength (_cField);

		char * const pBuf = reinterpret_cast<char *> (alloca ((cLen + 1)*sizeof (char)));
		::memcpy (pBuf, GetFieldS (_cRecord, _cField), cLen*sizeof (char));
		pBuf [cLen] = '\0';

		return ::atoi (pBuf);
	}

	size_t GetRecords () const {
		return (m_cSize - (m_pRecords - m_pData))/m_cRecordSize;
	}

	bool RecordExists (size_t _cRecord) const {
		return m_pRecords [_cRecord*m_cRecordSize] == ' ';
	}

  private:
	CHandle m_hFile;
	CHandle m_hFileMapping;
	const char * m_pData;
	size_t m_cSize;

	const char * m_pRecords;
	size_t m_cRecordSize;

	std::vector<size_t> m_vecFieldOffsets;
};

CDbfFile::CDbfFile (const char * _strFileName) :
	m_hFile (::CreateFile (_strFileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, NULL)),
	m_hFileMapping (NULL),
	m_pData (NULL),
	m_cSize (0),
	m_pRecords (NULL),
	m_cRecordSize (0)
{
	if (m_hFile == NULL || m_hFile == INVALID_HANDLE_VALUE)
		return;

	m_cSize = ::GetFileSize (m_hFile, NULL);

	m_hFileMapping = ::CreateFileMapping (m_hFile, NULL, PAGE_READONLY, 0, 0, NULL);
	m_pData = reinterpret_cast<char *> (::MapViewOfFile (m_hFileMapping, FILE_MAP_READ, 0, 0, 0));
	if (! m_pData)
		return;

	const CDbfHdr * const pHdr = reinterpret_cast<const CDbfHdr *> (m_pData);
	m_pRecords = m_pData + pHdr->wHdrSize;
	m_cRecordSize = pHdr->wRecordSize;

	const size_t cFields = (pHdr->wHdrSize - 32 - 1)/32;
	m_vecFieldOffsets.resize (cFields);
	size_t cOffset = 0;
	for (size_t cField = 0; cField < cFields; ++ cField) {
		if (pHdr->FieldDescriptors [cField].strName [0] == 0xD)
			break;

		m_vecFieldOffsets [cField] = cOffset;
		cOffset += pHdr->FieldDescriptors [cField].btLen;
	}
}

CDbfFile::~CDbfFile () {
	if (m_pData)
		::UnmapViewOfFile (m_pData);
}

//////////////////////////////////////////

struct CShpPolygon;

class CShpLoader : public CMapLoader {
  public:
	// Override.
	bool IsValid (const char * _pData, size_t _cFileSize) const;
	bool Load (const char * _pData, size_t _cFileSize, const char * _strFileName, map_t * _pMap);

  private:
	void ParsePolygon (const CShpPolygon * const _pPolygon, const char * _strEnd, USHORT _ushType, const char * _strLabel);

	map_t * m_pMap;
	const char * m_pStart, * m_pEnd;

	//
	size_t m_cOrdinal;

	file_refs_t m_file_refs;
};

//DECLARE_MAP_LOADER (CShpLoader, "shp", "ArcView GIS shapefile (*.shp)\0*.shp", false);

# ifdef _MSC_VER
#	pragma pack (push, 1)
# endif

const DWORD c_dwShpSignature = 0x0a270000;

enum shp_type_t {
	shpPoint    = 1,
	shpArc      = 2,
	shpPolygon  = 5,
	shpPointSet = 8
};

struct CShpBox {
	double dX0;       // Bounding rectangle.
	double dY0;
	double dX1;
	double dY1;
};

struct CShpHdr {
	DWORD dwSignature;      // Always == c_dwShpSignature.

	DWORD dwaReserved1 [5]; // Fill with zero'es.

	DWORD dwFileLen;        // In WORDs, big endian!
	DWORD dwVersion;        // 1000
	DWORD dwType;           // ? shp_type_t
	CShpBox box;            // Bounding rectangle.

	DWORD dwaReserved2 [8]; // Fill with zero'es.
};

struct CShpRecordHdr {
	DWORD dwOrdinal; // 1-based, big endian.
	DWORD dwLen;     // In WORDs, big endian.
	DWORD dwType;    // From shp_type_t.
};

struct CShpPointData {
	double dX;
	double dY;
};

struct CShpPoint : public CShpRecordHdr {
	CShpPointData point;
};

struct CShpPointSet : public CShpRecordHdr {
	CShpBox box;
	DWORD dwPoints;
	CShpPointData points [1];
};

struct CShpPolygon : public CShpRecordHdr {
	CShpBox box;

	DWORD dwNumParts;
	DWORD dwNumPoints;
	DWORD dwaParts [1];
};

# ifdef _MSC_VER
#	pragma pack (pop)
# endif

bool CShpLoader::IsValid (const char * _pData, size_t _cFileSize) const {
	assert (sizeof (CShpHdr) == 100);
	if (_cFileSize < sizeof (CShpHdr))
		return false;
	return reinterpret_cast<const CShpHdr *> (_pData)->dwSignature == c_dwShpSignature;
}

static
DWORD Big2LowEndian (DWORD _dw) {
	const BYTE * const ba = reinterpret_cast<BYTE *> (& _dw);
	return (ba [0] << 24) + (ba [1] << 16) + (ba [2] << 8) + ba [3];
}

void CShpLoader::ParsePolygon (const CShpPolygon * const _pPolygon, const char * _strEnd, USHORT _ushType, const char * _strLabel) {
	m_pMap->layers [3].rgns.push_back (rgn_t ());
	++ m_pMap->layers [3].cRgns;

	rgn_t & rgn = m_pMap->layers [3].rgns.back ();

	m_pMap->InsertRgnToSaveList (rgn);

	rgn.cBeginPos = (char *) _pPolygon - m_pStart;
	rgn.cEndPos = _strEnd - m_pStart;
	rgn.cFileLine = 0;
	rgn.cOrdinal = m_cOrdinal ++;

	rgn.strLabel = _strLabel;

	rgn.pTypeDesc = GetTypeDesc (_ushType, 3);
	rgn.ushType = _ushType;
	rgn.ushKind = 3;

	rgn.pFileRef = & m_file_refs.back ();

	const DWORD * const pdwParts = _pPolygon->dwaParts;
	const CShpPointData * pPoints = reinterpret_cast<const CShpPointData *> (pdwParts + _pPolygon->dwNumParts + 1);
	for (size_t cPart = 0; cPart < _pPolygon->dwNumParts; ++ cPart) {
		const DWORD dwPoints = cPart == _pPolygon->dwNumParts - 1 ? _pPolygon->dwNumPoints - pdwParts [cPart] : pdwParts [cPart + 1] - pdwParts [cPart];

		const char * const pPartStart = reinterpret_cast<const char *> (pPoints);

		points_t points (dwPoints);
		for (size_t cPoint = 0; cPoint < dwPoints; ++ cPoint) {
			point_t & point = points [cPoint];
			point.x = pPoints->dX;
			point.y = pPoints->dY;

			++ pPoints;
		}

		rgn.elements.push_back (rgn_element_t ());
		rgn_element_t & element = rgn.elements.back ();
		element.cFileLine = 0;
		element.cFileOffset = pPartStart - m_pStart;
		element.cLevel = 0;
		element.points.swap (points);
	}
}

bool CShpLoader::Load (const char * _pData, size_t _cFileSize, const char * _strMapFileName, map_t * _pMap) {
	m_pMap = _pMap;
	m_pStart = _pData;
	m_pEnd = _pData + _cFileSize;

	m_cOrdinal = 0;

	m_file_refs.clear ();
	m_file_refs.push_back (file_ref_t ());
	m_file_refs.back ().strFileName = _strMapFileName;
	m_file_refs.back ().bAttachment = true;

	SetProgressStatus (0);

	//
	// Prepare access to the associated .dbf file.
	//

	char strDbfName [MAX_PATH + 1];
	::strncpy (strDbfName, _strMapFileName, MAX_PATH);
	char * const pExt = ::strrchr (strDbfName, '.');
	if (pExt == NULL)
		::strcat (strDbfName, ".dbf");
	else
		::strcpy (pExt + 1, "dbf");

	CDbfFile dbf (strDbfName);
	if (dbf.CantOpen ()) {
		ReportWarning ("Can't open or read DBF file: '%s'.", strDbfName);
		return false;
	}
	const size_t cRecords = dbf.GetRecords ();

	// TODO: ask user.
	size_t cTypeFieldIndex = 2;
	size_t cLabelFieldIndex = 3;

	const size_t cLabelFieldLen = dbf.GetFieldLength (cLabelFieldIndex);
	char * const strLabelBuf = reinterpret_cast<char *> (::alloca ((cLabelFieldLen + 1)*sizeof (char)));

	//
	// Parse .shp
	//

	const CShpHdr * pHdr = reinterpret_cast<const CShpHdr *> (_pData);

	const char * pRecord = reinterpret_cast<const char *> (pHdr + 1);
	size_t cRecordCounter = 0;
	do {
		if (g_bStop)
			return false;

		USHORT ushType;
		if (cRecordCounter < cRecords) {
			// Omit deleted records in .dbf.
			while (! dbf.RecordExists (cRecordCounter))
				++ cRecordCounter;

			// Get type of region from .dbf.
			ushType = dbf.GetFieldN (cRecordCounter, cTypeFieldIndex);

			// Get label of region from .dbf.
			const char * const strLabel = dbf.GetFieldS (cRecordCounter, cLabelFieldIndex);
			::strncpy (strLabelBuf, strLabel, cLabelFieldLen);
			strLabelBuf [cLabelFieldLen] = '\0';
			// Cut ending spaces.
			for (char * p = strLabelBuf + cLabelFieldLen - 1; p >= strLabelBuf; -- p)
				if (* p == ' ')
					* p = '\0';
				else
					break;

			++ cRecordCounter;
		} else {
			ReportWarning ("Not enough records in DBF file.");
			ushType = 0;
			strLabelBuf [0] = '\0';
		}

		// Get next record in .shp.
		const CShpRecordHdr * const pRecHdr = reinterpret_cast<const CShpRecordHdr *> (pRecord);
		const char * const pNextRecord = pRecord + 8 + Big2LowEndian (pRecHdr->dwLen)*sizeof (WORD);

		switch (pRecHdr->dwType) {
			case shpPoint: {
				break;
			}

			case shpPointSet: {
				break;
			}

			case shpArc: {
				break;
			}

			case shpPolygon:
			case 0x1f:
				ParsePolygon (static_cast<const CShpPolygon *> (pRecHdr), pNextRecord, ushType, strLabelBuf);
				break;
		}

		pRecord = pNextRecord;

		SetProgressStatus (50.*(pRecord - _pData)/_cFileSize);
	} while (pRecord < m_pEnd);

	SetProgressStatus (50);

	m_pMap->file_refs.swap (m_file_refs);

	// TODO: ask user.
	m_pMap->Zoom.resize (2);
	m_pMap->Zoom [0] = 0;
	m_pMap->Zoom [1] = 1;
	m_pMap->Levels.resize (2);
	m_pMap->Levels [0] = 24;
	m_pMap->Levels [1] = 23;

	return true;
}
