# include "stdafx.h"
# include "Status.h"
# include "Map.h"
# include "Types.h"
# include "Load.h"
# include "Log.h"
# include "Globals.h"

class CTxtLoader : public CMapLoader {
  public:
	// Override.
	bool IsValid (const char * _pData, size_t _cFileSize) const;
	bool Load (const char * _pData, size_t _cFileSize, const char * _strFileName, map_t * _pMap);

  private:
	enum GroupType {
		GT_NONE, GT_NONE_START,
		GT_IMG_ID,
		GT_RGN10, GT_RGN20, GT_RGN40, GT_RGN80,
		GT_PLT, GT_WPT, GT_SHP, GT_FILE, GT_DICTIONARY, GT_MAP, GT_COUNTRIES, GT_REGIONS, GT_CITIES,
		GT_UNKN
	} m_GroupType;

	bool ParseFloat (const char * _pValue, const char * _pEnd, const char ** _ppEnd, float & _fRet);
	bool ParsePoint (const char * _pValue, const char * _pEnd, point_t & _pt, const char ** _ppEnd);
	void ParsePoints (const char * _pValue, const char * _pEnd, points_t & _points);
	const char * ParseLine (const char * _p);
	void AddRgn (GroupType _gt, size_t _ushKind, const char * _p);

	map_t * m_pMap;
	const char * m_pStart, * m_pEnd;

	// The layer of the current region.
	size_t m_ushKind;
	// The counter of current line.
	size_t m_cFileLine;
	// The counter of regions.
	size_t m_cOrdinal;
	// True if [IMG ID] is defined.
	bool m_bHeader;
	// The size of point in source text.
	size_t m_cPointLen;
	// Pointer to the probable comment beginning.
	const char * m_pComment;

	// The max number of digits in mantissa.
	size_t m_cPrecOrder;

	file_refs_t m_file_refs;
};

DECLARE_DEFAULT_MAP_LOADER (CTxtLoader, "txt", "Polish format (*.txt)\0*.txt", true);

	// Find value beginning.
static
const char * GetValue (const char * _pKey, const char * _pEnd) {
	if (_pKey >= _pEnd)
		return _pEnd;
	const char * const p = reinterpret_cast<char *> (::memchr (_pKey, '=', _pEnd - _pKey));
	return p ? p + 1 : _pEnd;
}

	// TODO: check _pEnd.
	// TODO: check remainder of the line.
static
unsigned long ParseType (const char * _pValue, const char * _pEnd) {
	if (_pValue [0] == '0' && _pValue [1] == 'x') {
		char * p = NULL;
		return ::strtoul (_pValue + 2, & p, 16);
	} else {
		char * p = NULL;
		return ::strtoul (_pValue, & p, 10);
	}
}

bool CTxtLoader::ParseFloat (const char * _pValue, const char * _pEnd, const char ** _ppEnd, float & _fRet) {
	float r = 0;
	bool bMinus = false;
	bool bHasDigits = false;
	const char * p = _pValue;

	// Omit non-digits.
	while (p < _pEnd && ! (('0' <= * p && * p <= '9') || * p == '-' || * p == '.'))
		++ p;

	// Get sign.
	if (p < _pEnd && * p == '-') {
		++ p;
		bMinus = true;
	}

	// Get whole part of the digit.
	while (p < _pEnd && '0' <= * p && * p <= '9') {
		r *= 10;
		r += (* p) - '0';
		++ p;
		bHasDigits = true;
	}

	// Get decimal dot.
	if (p < _pEnd && * p != '.') {
		* _ppEnd = p;
		_fRet = bMinus ? -r : r;
		return bHasDigits;
	}
	++ p;
	const char * pFraction = p;

	// Get fractional part of the digit.
	float f = 0;
	float pw = 0.1f;
	while (p < _pEnd && '0' <= * p && * p <= '9') {
		f += ((* p) - '0')*pw;
		++ p;
		pw *= .1f;
		bHasDigits = true;
	}

	// Update max precision.
	if ((int) m_cPrecOrder < p - pFraction)
		m_cPrecOrder = p - pFraction;

	* _ppEnd = p;
	_fRet = bMinus ? -(r + f) : (r + f);

	return bHasDigits;
}

bool CTxtLoader::ParsePoint (const char * _pValue, const char * _pEnd, point_t & _pt, const char ** _ppEnd) {
	// Get opening bracket.
	if (_pValue >= _pEnd || * _pValue != '(') {
		* _ppEnd = _pEnd;
		return false;
	}

	// Get latitude.
	if (! ParseFloat (_pValue + 1, _pEnd, _ppEnd, _pt.y)) {
		ReportWarning (_pValue - m_pStart, "Latitude value is expected. The point will be ignored.");
		* _ppEnd = _pEnd;
		return false;
	}

	// Get delimiting comma.
	if (* _ppEnd >= _pEnd || ** _ppEnd != ',') {
		ReportWarning (_pValue - m_pStart, "Comma between latitude and longitude is expected. The point will be ignored.");
		* _ppEnd = _pEnd;
		return false;
	}

	// Get longitude.
	if (! ParseFloat ((* _ppEnd) + 1, _pEnd, _ppEnd, _pt.x)) {
		ReportWarning (_pValue - m_pStart, "Longitude value is expected. The point will be ignored.");
		* _ppEnd = _pEnd;
		return false;
	}

	// Get closing bracket.
	if (* _ppEnd >= _pEnd || ** _ppEnd != ')') {
		ReportWarning (_pValue - m_pStart, "Closing bracket for point is expected. The point will be ignored.");
		* _ppEnd = _pEnd;
		return false;
	}
	++ (* _ppEnd);

	// Check coordinates.
	if (180 < _pt.x || _pt.x < -180) {
		// TODO: get reminder for /180 ?
		ReportWarning (_pValue - m_pStart, "Invalid longitude value: %.5f. The point will be ignored.", _pt.x);
		return false;
	}
	if (90 <= _pt.y || _pt.y <= -90) {
		ReportWarning (_pValue - m_pStart, "Invalid latitude value: %.5f. The point will be ignored.", _pt.y);
		return false;
	}

	return true;
}

void CTxtLoader::ParsePoints (const char * _pValue, const char * _pEnd, points_t & _points) {
	for (const char * p = _pValue; p < _pEnd; ++ p) {
		point_t pt;
		if (! ParsePoint (p, _pEnd, pt, & p))
			break;
		_points.push_back (pt);

		if (p >= _pEnd || * p != ',')
			break;
	}
}

void CTxtLoader::AddRgn (GroupType _gt, size_t _ushKind, const char * _p) {
	if (! m_bHeader)
		ReportWarning (_p - m_pStart, "[IMG ID] section should be defined first.");

	m_GroupType = _gt;
	m_ushKind = _ushKind;

	// Create new rgn.
	layer_t & layer = m_pMap->layers [_ushKind];
	++ layer.cRgns;
	layer.rgns.push_back (rgn_t ());

	// Initialize the rgn.
	rgn_t & rgn = layer.rgns.back ();
	rgn.cBeginPos = (m_pComment ? m_pComment : _p) - m_pStart;
	rgn.ushKind   = m_ushKind;
	rgn.cFileLine = m_cFileLine;
	rgn.cOrdinal  = m_cOrdinal ++;
	rgn.pFileRef  = & m_file_refs.back ();

	if (m_pComment)
		rgn.strComment.assign (m_pComment, _p - m_pComment);

	// Update the load order list.
	m_pMap->InsertRgnToSaveList (rgn);
}

const char * CTxtLoader::ParseLine (const char * _p) {
	const char * const pNextLine = FindNextLine (_p, m_pEnd);

	switch (_p [0]) {
		case ';':
			if (_p + 1 < m_pEnd && _p [1] == '@') {
				// Comment contains extended data.
				if (
					HasPrefix (_p + 2, pNextLine, "IMAGE,\"", 7) ||
					HasPrefix (_p + 2, pNextLine, "FILE,\"", 6)
				) {
					const char * strFileNameBegin = reinterpret_cast<char *> (::memchr (_p + 2, '\"', pNextLine - _p - 1));
					if (strFileNameBegin)
						++ strFileNameBegin;
					else {
						ReportWarning (_p - m_pStart, "Unexpected format of attachment definition.");
						break;
					}
					const char * const strFileNameEnd = reinterpret_cast<char *> (::memchr (strFileNameBegin, '\"', pNextLine - strFileNameBegin - 1));
					if (strFileNameEnd == NULL || strFileNameEnd > strFileNameBegin + MAX_PATH) {
						ReportWarning (_p - m_pStart, "Unexpected format of attachment definition.");
						break;
					}

					char strFileName [MAX_PATH + 1];
					::memcpy (strFileName, strFileNameBegin, strFileNameEnd - strFileNameBegin);
					strFileName [strFileNameEnd - strFileNameBegin] = '\0';

					ReportText ("");
					ReportText ("Loading attached file '%s'...", strFileName);

					map_t mapImage;
					bool bTypeCanBeSaved;
					if (! CMapLoader::LoadFile (strFileName, & mapImage, bTypeCanBeSaved)) {
						ErrorMsgBox ("Can't load the attached file\n%s\n\nSee Error Log for details.", strFileName);
						break;
					}
					m_pMap->Join (mapImage);

					ReportText ("");
				}
			} else {
				// Ordinary comment.
				if (! m_pComment)
					m_pComment = _p;
			}
			break;

		case '\n':
		case '\r':
			break;

		case ' ':
		case '\t':
			ReportWarning (_p - m_pStart, "Line begins with space or tabstop. It will be ignored.");
			break;

		case '[': {
			if (HasPrefix (_p + 1, pNextLine, "END", 3)) {
				switch (m_GroupType) {
					case GT_NONE:
					case GT_NONE_START:
						ReportWarning (_p - m_pStart, "[END] has not matching section.");
						break;

					case GT_IMG_ID:
						m_pMap->cHeaderLen = pNextLine - m_pStart;
						break;

					case GT_RGN10:
					case GT_RGN20:
					case GT_RGN40:
					case GT_RGN80:
						m_pMap->layers [m_ushKind].rgns.back ().cEndPos = pNextLine - m_pStart;
						break;
				}
				m_GroupType = GT_NONE;
				m_pComment = NULL;
				break;
			} 
			if (m_GroupType != GT_NONE && m_GroupType != GT_NONE_START) {
				ReportWarning (_p - m_pStart, "Section without closing [END].");
				m_pComment = NULL;
			}

			if (HasPrefix (_p + 1, pNextLine, "RGN10]", 6))
				AddRgn (GT_RGN10, 0, _p);
			else if (HasPrefix (_p + 1, pNextLine, "RGN20]", 6))
				AddRgn (GT_RGN20, 1, _p);
			else if (HasPrefix (_p + 1, pNextLine, "RGN40]", 6))
				AddRgn (GT_RGN40, 2, _p);
			else if (HasPrefix (_p + 1, pNextLine, "RGN80]", 6))
				AddRgn (GT_RGN80, 3, _p);
			else if (HasPrefix (_p + 1, pNextLine, "IMG ID]", 7)) {
				if (m_bHeader)
					ReportWarning (_p - m_pStart, "[IMG ID] section is already defined.");
				m_GroupType = GT_IMG_ID;
				m_bHeader = true;
			} else if (HasPrefix (_p + 1, pNextLine, "PLT]", 4)) {
				ReportWarning (_p - m_pStart, "[PLT] section is not supported. It will be ignored.");
				// TODO: save content
				m_GroupType = GT_PLT;
			} else if (HasPrefix (_p + 1, pNextLine, "WPT]", 4)) {
				ReportWarning (_p - m_pStart, "[WPT] section is not supported. It will be ignored.");
				// TODO: save content
				m_GroupType = GT_WPT;
			} else if (HasPrefix (_p + 1, pNextLine, "SHP]", 4)) {
				ReportWarning (_p - m_pStart, "[SHP] section is not supported. It will be ignored.");
				// TODO: save content
				m_GroupType = GT_SHP;
			} else if (HasPrefix (_p + 1, pNextLine, "FILE]", 5)) {
				ReportWarning (_p - m_pStart, "[FILE] section is not supported. It will be ignored.");
				// TODO: save content
				m_GroupType = GT_FILE;
			} else if (HasPrefix (_p + 1, pNextLine, "DICTIONARY]", 11)) {
				ReportWarning (_p - m_pStart, "[DICTIONARY] section is not supported. It will be ignored.");
				// TODO: save content
				m_GroupType = GT_DICTIONARY;
			} else if (HasPrefix (_p + 1, pNextLine, "MAP]", 4)) {
				ReportWarning (_p - m_pStart, "[MAP] section is not supported. It will be ignored.");
				// TODO: save content
				m_GroupType = GT_MAP;
			} else if (HasPrefix (_p + 1, pNextLine, "Countries]", 10)) {
				ReportWarning (_p - m_pStart, "[Countries] section is not supported. It will be ignored.");
				// TODO: save content
				m_GroupType = GT_COUNTRIES;
			} else if (HasPrefix (_p + 1, pNextLine, "Regions]", 8)) {
				ReportWarning (_p - m_pStart, "[Regions] section is not supported. It will be ignored.");
				// TODO: save content
				m_GroupType = GT_REGIONS;
			} else if (HasPrefix (_p + 1, pNextLine, "Cities]", 7)) {
				ReportWarning (_p - m_pStart, "[Cities] section is not supported. It will be ignored.");
				// TODO: save content
				m_GroupType = GT_CITIES;
			} else {
				char strBuf [32];
				const size_t cLen = pNextLine - _p - 1 > 31 ? 31 : pNextLine - _p - 1;
				::strncpy (strBuf, _p, cLen);
				strBuf [cLen] = '\0';
				ReportWarning (_p - m_pStart, "Unknown section type: \'%s\'. The section will be ignored.", strBuf);

				m_GroupType = GT_UNKN;
			}
			
			break;
		}

		default: {
			if (m_GroupType == GT_NONE) {
				char strBuf [32];
				const size_t cLen = pNextLine - _p - 1 > 31 ? 31 : pNextLine - _p - 1;
				::strncpy (strBuf, _p, cLen);
				strBuf [cLen] = '\0';

				ReportWarning (_p - m_pStart, "Some text is out of section: \'%s\'.", strBuf);

				break;
			}

			const char * const pKey = _p;
			const char * const pValue = GetValue (_p, pNextLine);

			switch (m_GroupType) {
				case GT_UNKN:
				case GT_PLT:
				case GT_WPT:
				case GT_SHP:
					break;

				case GT_IMG_ID: {
					if (HasPrefix (pKey, pNextLine, "Levels=", 7)) {
						size_t cLevels = ::atoi (pValue);
						if (cLevels > 10) {
							cLevels = 10;
							ReportWarning (_p - m_pStart, "Value of 'Levels' is too large. Reset to 10.");
						}
						m_pMap->Zoom.resize (cLevels);
						m_pMap->Levels.resize (cLevels);
						for (size_t cLevel = 0; cLevel < cLevels; ++ cLevel) {
							m_pMap->Zoom   [cLevel] = 0;
							m_pMap->Levels [cLevel] = 0;
						}
					} else if (HasPrefix (pKey, pNextLine, "Zoom", 4)) {
						const int iLevel = pKey [4] - '0';
						if (iLevel < 0 || size_t (iLevel) >= m_pMap->Zoom.size ()) {
							ReportWarning (_p - m_pStart, "Zoom%d is discarded. Check Levels definition.", iLevel);
							break;
						}

						int iZoom = ::atoi (pValue);
						if (iZoom < 0) {
							ReportWarning (_p - m_pStart, "Zoom%d has incorrect value %d. Reset to 0.", iLevel, iZoom);
							iZoom = 0;
						} else if (iZoom > 7) {
							ReportWarning (_p - m_pStart, "Zoom%d has incorrect value %d. Reset to 7.", iLevel, iZoom);
							iZoom = 7;
						}

						m_pMap->Zoom [iLevel] = iZoom;
					} else if (HasPrefix (pKey, pNextLine, "RgnLimit=", 9)) {
						m_pMap->cRgnLimit = ::atoi (pValue);
						if (m_pMap->cRgnLimit > 1024) {
							ReportWarning (_p - m_pStart, "RgnLimit is too large. Should be < 1024.");
							m_pMap->cRgnLimit = 1024;
						}
					} else if (HasPrefix (pKey, pNextLine, "LevelFill=", 10)) {
						// TODO:
						goto l_Unknown;
					} else if (HasPrefix (pKey, pNextLine, "LevelLimit=", 11)) {
						// TODO:
						goto l_Unknown;
					} else if (HasPrefix (pKey, pNextLine, "Level", 5)) {
						const int iLevel = pKey [5] - '0';
						if (iLevel < 0 || size_t (iLevel) >= m_pMap->Levels.size ()) {
							ReportWarning (_p - m_pStart, "Level%d is discarded. Check Levels definition.", iLevel);
							break;
						}

						int iLevelV = ::atoi (pValue);
						if (iLevelV < 0)
							ReportWarning (_p - m_pStart, "Level%d has incorrect value %d.", iLevel, iLevelV);
						else if (iLevelV > 24)
							ReportWarning (_p - m_pStart, "Level%d has incorrect value %d.", iLevel, iLevelV);

						m_pMap->Levels [iLevel] = iLevelV;
					} else {
l_Unknown:
						// Save all unknown key-value pairs.
						m_pMap->header_data.push_back (std::string (_p, pNextLine - _p - 1));
					}
					break;
				}

				case GT_RGN10:
				case GT_RGN20:
				case GT_RGN40:
				case GT_RGN80: {
					rgn_t & rgn = m_pMap->layers [m_ushKind].rgns.back ();
					if (HasPrefix (pKey, pNextLine, "Type=", 5)) {
						const USHORT ushType = ParseType (pValue, pNextLine);
						rgn.pTypeDesc = GetTypeDesc (ushType, m_ushKind);
						if (rgn.pTypeDesc == & g_types [0]) {
							static const char * strRgns [] = {"RGN10", "RGN20", "RGN40", "RGN80"};
							ReportWarning (_p - m_pStart, "Unknown type 0x%x, %s.", ushType, strRgns [m_ushKind]);
						}
						rgn.ushType = ushType;
					} else if (HasPrefix (pKey, pNextLine, "Label=", 6)) {
						size_t cLen = pNextLine - pValue;
						if (pValue [cLen - 1] == '\n' || pValue [cLen - 1] == '\r')
							-- cLen;
						rgn.strLabel.assign (pValue, cLen);
					} else if (HasPrefix (pKey, pNextLine, "DirIndicator=", 13)) {
						if (::atoi (pValue) != 0)
							rgn.dwFlags |= rgn_t::fDirIndicator;
						rgn.dwFlags |= rgn_t::fDirIndicatorDefined;
					} else if (HasPrefix (pKey, pNextLine, "Data", 4)) {
						const size_t cLevel = pKey [4] - '0';
						if (cLevel >= m_pMap->Levels.size ()) {
							ReportWarning (_p - m_pStart, "Incorrect level: Data%d. The element will be ignored.", cLevel);
							break;
						}
						if (cLevel == m_pMap->Levels.size () - 1) {
							ReportWarning (_p - m_pStart, "Level%d should be empty. The element will be ignored.", cLevel);
							break;
						}

						rgn.elements.push_back (rgn_element_t ());
						rgn_element_t & element = rgn.elements.back ();
						element.cLevel = cLevel;
						element.cFileLine = m_cFileLine;
						element.cFileOffset = _p - m_pStart;

						points_t & points = element.points;
						points.reserve ((pNextLine - pValue + 2)/m_cPointLen);
						ParsePoints (pValue, pNextLine, points);

						// Recalculate average point length.
						if (! points.empty ()) {
							m_cPointLen = (pNextLine - pValue + 2)/points.size ();
							if (m_cPointLen == 0)
								m_cPointLen = 22;
						}
					} else if (HasPrefix (pKey, pNextLine, "Origin", 6)) {
						const size_t cLevel = pKey [6] - '0';
						if (cLevel >= m_pMap->Levels.size ()) {
							ReportWarning (_p - m_pStart, "Incorrect level: Origin%d. The point will be ignored.", cLevel);
							break;
						}
						if (cLevel == m_pMap->Levels.size () - 1) {
							ReportWarning (_p - m_pStart, "Level%d should be empty. The point will be ignored.", cLevel);
							break;
						}

						point_t point;
						const char * p = NULL;
						if (! ParsePoint (pValue, pNextLine, point, & p))
							break;

						rgn.origins.push_back (rgn_origin_t ());
						rgn_origin_t & origin = rgn.origins.back ();
						origin.cLevel = cLevel;
						origin.point = point;
						origin.cFileLine = m_cFileLine;
						origin.cFileOffset = _p - m_pStart;
					} else {
						char strBuf [32];
						const size_t cLen = pValue - pKey - 1 > 31 ? 31 : pValue - pKey - 1;
						::strncpy (strBuf, _p, cLen);
						strBuf [cLen] = '\0';

						ReportWarning (_p - m_pStart, "Unknown key name %s. It will be ignored.", strBuf);
					}
				}
			}
			break;
		}
	}

	if (pNextLine [-1] == '\n')
		// Count lines.
		++ m_cFileLine;

	return pNextLine;
}

bool CTxtLoader::Load (const char * _pData, size_t _cFileSize, const char * _strMapFileName, map_t * _pMap) {
	if (! CheckAnsiFile (_pData, _cFileSize, false))
		return false;

	m_pMap = _pMap;
	m_pStart = _pData;
	m_pEnd = _pData + _cFileSize;

	m_file_refs.clear ();
	m_file_refs.push_back (file_ref_t ());
	m_file_refs.back ().strFileName = _strMapFileName;
	m_file_refs.back ().bAttachment = false;

	m_pComment = NULL;
	m_cPrecOrder = 0;
	m_bHeader = false;
	m_GroupType = GT_NONE_START;
	m_ushKind = 0;
	m_cFileLine = 1;
	m_cOrdinal = 0;
	m_cPointLen = 22;

	SetProgressStatus (0);

	const char * pLine = _pData;
	do {
		pLine = ParseLine (pLine);
		SetProgressStatus (fast_floor ((50.*(pLine - _pData))/_cFileSize + .5));
	} while (pLine < _pData + _cFileSize && ! g_bStop);

	if (m_GroupType != GT_NONE)
		ReportWarning (m_pEnd - m_pStart - 1, "Section is not closed with [END].");
	if (! m_bHeader)
		ReportWarning ("No [IMG ID] section found.");

	SetProgressStatus (50);

	m_pMap->cPrecOrder = m_cPrecOrder > 0 ? m_cPrecOrder : 5;

	m_pMap->file_refs.swap (m_file_refs);

	return true;
}

bool CTxtLoader::IsValid (const char * _pData, size_t _cFileSize) const {
	if (! CheckAnsiFile (_pData, _cFileSize, true))
		return false;

	// Limit the time of validation.
	if (_cFileSize > 65536)
		_cFileSize = 65536;

	size_t cFileLine = 0;
	bool bImgId = false;

	const char * pLine = _pData;
	do {
		const char * const pNextLine = FindNextLine (pLine, _pData + _cFileSize);

		switch (pLine [0]) {
			case '[':
				if (! bImgId)
					bImgId = HasPrefix (pLine + 1, pNextLine, "IMG ID]", 7);
				else
					return HasPrefix (pLine + 1, pNextLine, "END", 3);
		}

		if (pNextLine [-1] == '\n')
			// Count lines.
			++ cFileLine;

		if (cFileLine > 1000)
			return false;

		pLine = pNextLine;
	} while (pLine < _pData + _cFileSize && ! g_bStop);

	return false;
}