# include "stdafx.h"
# include "Map.h"
# include "Types.h"
# include "Save.h"
# include "Log.h"
# include "Status.h"
# include "Version.h"
# include "Globals.h"

const size_t c_cBufferSize = 65536;
static char g_pBuffer [c_cBufferSize];
static size_t g_cBufferPosition = 0;

static bool g_bOverflow;

static size_t g_cPos;
static size_t g_cLine;
static size_t g_cOrdinal;

void CTxtSaver::FlushBuffer () {
	if (g_cBufferPosition == 0)
		return;

	DWORD dwWritten = 0;
	::WriteFile (m_hFile, g_pBuffer, g_cBufferPosition, & dwWritten, NULL);
	g_bOverflow = dwWritten < g_cBufferPosition;
	if (g_bOverflow)
		ReportWarning ("Write error: 0x%x.", ::GetLastError ());
		
	g_cBufferPosition = 0;
}

void CTxtSaver::WriteS (const char * _str) {
	const size_t cLen = ::strlen (_str);
	g_cPos += cLen;

	for (const char * p = _str; * p; ++ p)
		if (* p == '\n')
			++ g_cLine;

	if (cLen > c_cBufferSize) {
		FlushBuffer ();

		DWORD dwWritten = 0;
		::WriteFile (m_hFile, _str, cLen, & dwWritten, NULL);
		g_bOverflow = dwWritten < cLen;

		return;
	}

	if (g_cBufferPosition + cLen > c_cBufferSize)
		FlushBuffer ();

	::memcpy (g_pBuffer + g_cBufferPosition, _str, cLen);
	g_cBufferPosition += cLen;
}

void CTxtSaver::WriteC (char _c) {
	char buf [2];
	buf [0] = _c;
	buf [1] = '\0';
	WriteS (buf);
}

void CTxtSaver::WriteU (unsigned _u) {
	char buf [16];
	buf [15] = '\0';
	for (size_t c = 14; c > 0; -- c) {
		buf [c] = '0' + (_u % 10);
		_u /= 10;
		if (_u == 0) {
			WriteS (buf + c);
			return;
		}
	}
	assert (false);
}

void CTxtSaver::WriteX (unsigned _x) {
	char buf [10];
	::ultoa (_x, buf, 16);
	WriteS (buf);
}

void CTxtSaver::WriteF (float _f) {
	if (_f < 0) {
		WriteS ("-");
		_f = -_f;
	}

	_f += .5/m_fPrecFactor;

	const unsigned uInt = fast_floor (_f);

	WriteU (uInt);
	WriteC ('.');
	WriteFraction (_f - uInt);
}

void CTxtSaver::WriteFraction (float _f) {
	assert (_f >= 0);
	unsigned u = fast_floor (_f*m_fPrecFactor);

	assert (m_cPrecOrder <= 5);
	char buf [7];
	for (size_t c = 0; c < m_cPrecOrder; ++ c) {
		buf [m_cPrecOrder - c - 1] = '0' + (u % 10);
		u /= 10;
	}
	buf [m_cPrecOrder] = '\0';

	WriteS (buf);
}

void CTxtSaver::Save (HANDLE _hFile, const char * _strFileName, map_t * _pMap) {
	m_hFile = _hFile;
	g_cBufferPosition = 0;
	g_bOverflow = false;

	g_cPos = 0;
	g_cLine = 1;
	g_cOrdinal = 0;

	_pMap->file_refs.push_back (file_ref_t ());
	file_ref_t & file_ref = _pMap->file_refs.back ();
	file_ref.strFileName = _strFileName;
	file_ref.bAttachment = false;

	m_cPrecOrder = _pMap->cPrecOrder;
	if (m_cPrecOrder > 5)
		// NOTE: Garmin IMG format has limited precision.
		m_cPrecOrder = 5;
	m_fPrecFactor = ::pow (10, m_cPrecOrder);

	SetProgressStatus (0);

	//
	// Save header.
	//

	WriteS (
		"; Generated by GPSMap Edit " GME_VER_S "\r\n"
		"\r\n"
		"[IMG ID]\r\n"
	);

	for (size_t cKV = 0; cKV < _pMap->header_data.size (); ++ cKV) {
		WriteS (_pMap->header_data [cKV].data ());
		WriteS ("\r\n");
	}

	WriteS ("RgnLimit=");
	WriteU (_pMap->cRgnLimit);
	WriteS ("\r\nLevels=");
	WriteC ('0' + _pMap->Levels.size ());
	WriteS ("\r\n");
	size_t cLevel;
	for (cLevel = 0; cLevel < _pMap->Levels.size (); ++ cLevel) {
		WriteS ("Level");
		WriteC ('0' + cLevel);
		WriteS ("=");
		WriteU (_pMap->Levels [cLevel]);
		WriteS ("\r\n");
	}
	for (cLevel = 0; cLevel < _pMap->Zoom.size (); ++ cLevel) {
		WriteS ("Zoom");
		WriteC ('0' + cLevel);
		WriteS ("=");
		WriteU (_pMap->Zoom [cLevel]);
		WriteS ("\r\n");
	}

	WriteS ("[END-IMG ID]\r\n\r\n");

	size_t cRgn = 0;
	const size_t cRgns = _pMap->GetRgnCount ();
	for (rgn_t * pRgn = _pMap->pFirstRgn; pRgn; pRgn = pRgn->pNextRgn, ++ cRgn) {
			rgn_t & rgn = * pRgn;
			if (rgn.IsDeleted () || (rgn.dwFlags & rgn_t::fWaypointOrTrack))
				continue;

			rgn.cFileLine = g_cLine;
			rgn.cBeginPos = g_cPos;
			rgn.cOrdinal = g_cOrdinal ++;
			rgn.pFileRef = & file_ref;

			// Save comment.
			if (! rgn.strComment.empty ())
				WriteS (rgn.strComment.c_str ());

			// Save type.
			WriteS ("[RGN");
			if (rgn.pTypeDesc == g_types) {
				WriteC ('0' + (1 << rgn.ushKind));
				WriteS ("0]\r\nType=0x");
				WriteX (rgn.ushType);
			} else {
				WriteC ('0' + (1 << rgn.pTypeDesc->ushKind));
				WriteS ("0]\r\nType=0x");
				WriteX (rgn.pTypeDesc->ushType);
			}
			WriteS ("\r\n");

			// Save label.
			if (! rgn.strLabel.empty ()) {
				WriteS ("Label=");
				WriteS (rgn.strLabel.c_str ());
				WriteS ("\r\n");
			}

			// Save elements.
			const size_t cElements = rgn.elements.size ();
			for (size_t cElement = 0; cElement < cElements; ++ cElement) {
				rgn_element_t & element = rgn.elements [cElement];
				if (element.bDeleted)
					continue;

				const points_t & points = element.points;
				const size_t cPoints = points.size ();
				if (cPoints == 0)
					continue;

				element.cFileLine = g_cLine;
				element.cFileOffset = g_cPos;

				WriteS ("Data");
				WriteC ('0' + element.cLevel);
				WriteS ("=(");

				// Determine the bounding rect.
				for (size_t cPoint = 0; cPoint < cPoints; ++ cPoint) {
					const point_t & point = points [cPoint];

					if (cPoint != 0)
						WriteS ("),(");
					WriteF (point.y);
					WriteS (",");
					WriteF (point.x);
				}

				WriteS (")\r\n");
			}

			if (rgn.dwFlags & rgn_t::fDirIndicatorDefined) {
				WriteS ("DirIndicator=");
				WriteS (rgn.dwFlags & rgn_t::fDirIndicator ? "1\r\n" : "0\r\n");
			}

			// Save origins.
			for (size_t cOrigin = 0; cOrigin < rgn.origins.size (); ++ cOrigin) {
				rgn_origin_t & origin = rgn.origins [cOrigin];
				if (origin.bDeleted)
					continue;

				origin.cFileLine = g_cLine;
				origin.cFileOffset = g_cPos;

				WriteS ("Origin");
				WriteC ('0' + origin.cLevel);
				WriteS ("=(");
				WriteF (origin.point.y);
				WriteS (",");
				WriteF (origin.point.x);
				WriteS (")\r\n");
			}

			WriteS ("[END-RGN");
			if (rgn.pTypeDesc == g_types)
				WriteC ('0' + (1 << rgn.ushKind));
			else
				WriteC ('0' + (1 << rgn.pTypeDesc->ushKind));
			WriteS ("0]\r\n\r\n");

			rgn.cEndPos = g_cPos;

			SetProgressStatus (fast_floor (100.*cRgn/cRgns + .5));
	}

	// Save references to the images.
	for (images_t::const_iterator iImage = _pMap->images.begin (); iImage != _pMap->images.end (); ++ iImage) {
		if (iImage->bOverlay || iImage->bDeleted)
			continue;

		WriteS (";@IMAGE,\"");
		WriteS (iImage->strFileName.data ());
		// NOTE: '0,0' is reserved for position correction.
		WriteS ("\",0,0\r\n");
	}

	// Save references to the attached waypoints and tracks.
	for (file_refs_t::const_iterator iFileRef = _pMap->file_refs.begin (); iFileRef != _pMap->file_refs.end (); ++ iFileRef) {
		if (iFileRef->bDeleted || ! iFileRef->bAttachment)
			continue;

		WriteS (";@FILE,\"");
		WriteS (iFileRef->strFileName.data ());
		// NOTE: '0,0' is reserved for position correction.
		WriteS ("\",0,0\r\n");
	}

	// TODO: save all unknown sections.

	FlushBuffer ();

	SetProgressStatus (100);

	if (g_bOverflow)
		ErrorMsgBox ("File has been written partialy.\nThis may be caused by disk overflow.");
}
