# ifndef __Wrappers_H_
# define __Wrappers_H_

class CByteArrayHolder {
	// Disable copy semantics.
	CByteArrayHolder (const CByteArrayHolder &);
	void operator = (const CByteArrayHolder &);

  public:
	CByteArrayHolder (BYTE * _p) : m_p (_p) {}
	~CByteArrayHolder () {delete [] m_p;}

  private:
	BYTE * const m_p;
};

class CCom {
  public:
	CCom () {::CoInitialize (NULL);}
	~CCom () {::CoUninitialize ();}
};

class CUnknownPtr {
	// Disable copy semantics.
	CUnknownPtr (const CUnknownPtr &);
	void operator = (const CUnknownPtr &);

  public:
	CUnknownPtr (IUnknown * _p) : m_p (_p) {}
	~CUnknownPtr () {if (m_p) m_p->Release ();}

  private:
	IUnknown * m_p;
};

class CHandle {
	// Disable copy semantics.
	CHandle (const CHandle &);
	void operator = (const CHandle &);

  public:
	CHandle (HANDLE _h) : m_h (_h) {}
	~CHandle () {Close ();}

	void Close () {
		if (m_h && m_h != INVALID_HANDLE_VALUE)
			::CloseHandle (m_h);
		m_h = NULL;
	}

	operator HANDLE () const {return m_h;}

	void operator = (HANDLE _h) {
		Close ();
		m_h = _h;
	}

  private:
	HANDLE m_h;
};

class CCritSec {
	// Disable copy semantics.
	CCritSec (const CCritSec &);
	void operator= (const CCritSec &);

  public:
	CCritSec () {::InitializeCriticalSection (& m_cs);}
	~CCritSec () {::DeleteCriticalSection (& m_cs);}

	void Lock () {::EnterCriticalSection (& m_cs);}
	void Unlock () {::LeaveCriticalSection (& m_cs);}

	class CLock {
	  public:
		CLock (CCritSec & _cs) : m_cs (_cs) {m_cs.Lock ();}
		~CLock () {m_cs.Unlock ();}

	  private:
		CCritSec & m_cs;
	};

  private:
	CRITICAL_SECTION m_cs;
};

class CClientDC {
	// Disable copy semantics.
	CClientDC (const CClientDC &);
	void operator = (const CClientDC &);

  public:
	CClientDC (HWND _hWnd) :
		m_hWnd (_hWnd),
		m_hDC (::GetDC (_hWnd))
	{}
	~CClientDC () {if (m_hDC) ::ReleaseDC (m_hWnd, m_hDC);}

	operator HDC () const {return m_hDC;}

  private:
	const HWND m_hWnd;
	const HDC m_hDC;
};

class CDCHolder {
	// Disable copy semantics.
	CDCHolder (const CDCHolder &);
	void operator = (const CDCHolder &);

  public:
	CDCHolder (HDC _hDC) : m_hDC (_hDC) {}
	~CDCHolder () {if (m_hDC) ::DeleteDC (m_hDC);}

	operator HDC () const {return m_hDC;}

	HDC Detach () {
		const HDC hDC = m_hDC;
		m_hDC = NULL;
		return hDC;
	}

  private:
	HDC m_hDC;
};

class CBitmapHolder {
	// Disable copy semantics.
	CBitmapHolder (const CBitmapHolder &);
	void operator = (const CBitmapHolder &);

  public:
	CBitmapHolder (HBITMAP _hBmp) : m_hBmp (_hBmp) {}
	~CBitmapHolder () {if (m_hBmp) ::DeleteObject (m_hBmp);}

	operator HBITMAP () const {return m_hBmp;}

	HBITMAP Detach () {
		const HBITMAP hBmp = m_hBmp;
		m_hBmp = NULL;
		return hBmp;
	}

  private:
	HBITMAP m_hBmp;
};

# endif // __Wrappers_H_